// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// base version: JPP spectrum emulator
// heavily modified by Ketmar // Vampire Avalon.
// now we have really powerfull disassembler
// that can be used in various code analyzers.
// disassembler also allows making of custom
// output strings (due to independent fields
// for command name and arguments)
//
// i also fixed some bugs and introduced
// the new ones.
//
// the source is not the best of i can do, but
// it compiles and works somtimes, and that's
// enough for now %-)  K8
//
// TODO: add timings?
//
{$R-,I-}
unit dspUnasmU;

interface

uses
  SysUtils, Classes{$IFNDEF STANDALONE_UNASM}, dspEmuVarsU, z80U{$ENDIF};


const
{$IFDEF STANDALONE_UNASM}
  // flag bit positions
  Z80FLAGC = 1;
  Z80FLAGN = 2;
  Z80FLAGPV = 4;
  Z80FLAG3 = 8;
  Z80FLAGH = 16;
  Z80FLAG5 = 32;
  Z80FLAGZ = 64;
  Z80FLAGS = 128;
{$ENDIF}

  unasmMemReadFlag = $01;
  unasmMemWriteFlag = $02;
  unasmPortReadFlag = $04;
  unasmPortWriteFlag = $08;

  unasmAffectA = $01;
  unasmAffectB = $02;
  unasmAffectC = $04;
  unasmAffectD = $08;
  unasmAffectE = $10;
  unasmAffectH = $20;
  unasmAffectL = $40;
  unasmAffectF = $80;
  unasmAffectHX = $100;
  unasmAffectLX = $200;
  unasmAffectHY = $400;
  unasmAffectLY = $800;
  unasmAffectSP = $1000;
  unasmAffectI = $2000;
  unasmAffectR = $4000;


type
  TUnasmInfo = record
    valid: Boolean;
    addr: Word;
    len: Integer; // command length (in bytes)
    dump: array [0..7] of Byte; // cmd dump

    cmdName: string; // command name ('LD'; 'JP'; etc.)

    argC: Integer; // # of command args (0, 1, 2 or 3)
    args: array [0..2] of string; // string representation
    argsN: array [0..2] of Integer; // numeric representation
      // args[3] & argsN[2] is used in #DD|#FD/#CB/...

    memRefArg: Integer; // number of arg that contains memRef or -1

    isBranch: Boolean; // jp, jr, djnz, call, ret, rst?
    isJump: Boolean; // jp/jr/djnz?
    isRet: Boolean; // ret?
    isCall: Boolean; // call?
    isRst: Boolean; // rst?
    isCond: Boolean; // conditional call/jump/ret?
    isDoCond: Boolean; // condition is true?
    branchDest: Integer; // ALWAYS correct (does stack reading if necessary)

    dataSize: Integer; // 0, 1 or 2 bytes
    ioFlags: Integer;
      // bit 0: reads memory (POP, LDI[R], OTIR, etc. included)
      // bit 1: writes memory (PUSH, LDI[R], INIR, etc. included)
      // bit 2: read port
      // bit 3: writes port
    memSrcAddr: Integer; // memory address for i/o
    memDestAddr: Integer; // memory address for i/o
    portAddr: Integer; // port address for i/o
      // memAddr & portAddr are ALWAYS correct!
      // (does stack or registers reading if necessary)
    affectRegs: Integer;
      // registers affected by the instruction
      // bit 0: A
      // bit 1: B
      // bit 2: C
      // bit 3: D
      // bit 4: E
      // bit 5: H
      // bit 6: L
      // bit 7: F
      // bit 8: HX
      // bit 9: LX
      // bit 10: HY
      // bit 11: LY
      // bit 12: SP
      // bit 13: I
      // bit 14: R
  end;


function ZCmdLength (addr: Word): Integer;
procedure ZCmdDecode (addr: Word; out info: TUnasmInfo);
  // does not proces labels!

function ZPeekB (addr: Word): Byte;
function ZPeekW (addr: Word): Word;


{$IFDEF STANDALONE_UNASM}
type
  TZ80UVirtualPeekB = function (addr: Word): Byte;
  TZ80UGetReg16 = function (): Word;
  TZ80UGetReg8 = function (): Byte;


var
  z80uVirtualPeekB: TZ80UVirtualPeekB = nil;
  z80uGetBC: TZ80UGetReg16 = nil;
  z80uGetDE: TZ80UGetReg16 = nil;
  z80uGetHL: TZ80UGetReg16 = nil;
  z80uGetIX: TZ80UGetReg16 = nil;
  z80uGetIY: TZ80UGetReg16 = nil;
  z80uGetSP: TZ80UGetReg16 = nil;
  z80uGetA: TZ80UGetReg8 = nil;
  z80uGetF: TZ80UGetReg8 = nil;
{$ENDIF}


implementation

{$IFDEF STANDALONE_UNASM}
function ZPeekB (addr: Word): Byte;
begin
  result := z80uVirtualPeekB(addr); //Z80PeekBInternal(emuZ80, addr);
end;

function ZPeekW (addr: Word): Word;
begin
  result := ZPeekB(addr)+ZPeekB(addr+1)*256;
end;

function ZGetBC (): Word;
begin
  if Assigned(z80uGetBC) then result := z80uGetBC() else result := 0;
end;

function ZGetDE (): Word;
begin
  if Assigned(z80uGetDE) then result := z80uGetDE() else result := 0;
end;

function ZGetHL (): Word;
begin
  if Assigned(z80uGetHL) then result := z80uGetHL() else result := 0;
end;

function ZGetIX (): Word;
begin
  if Assigned(z80uGetIX) then result := z80uGetIX() else result := 0;
end;

function ZGetIY (): Word;
begin
  if Assigned(z80uGetIY) then result := z80uGetIY() else result := 0;
end;

function ZGetSP (): Word;
begin
  if Assigned(z80uGetSP) then result := z80uGetSP() else result := 0;
end;

function ZGetA (): Word;
begin
  if Assigned(z80uGetA) then result := z80uGetA() else result := 0;
end;

function ZGetF (): Word;
begin
  if Assigned(z80uGetF) then result := z80uGetF() else result := 0;
end;
{$ELSE}
function ZPeekB (addr: Word): Byte;
begin
  result := Z80PeekBInternal(emuZ80, addr);
end;

function ZPeekW (addr: Word): Word;
begin
  result := ZPeekB(addr) + ZPeekB(addr+1)*256;
end;

function ZGetBC (): Word;
begin
  result := Z80GetBC(emuZ80);
end;

function ZGetDE (): Word;
begin
  result := emuZ80.regDE;
end;

function ZGetHL (): Word;
begin
  result := emuZ80.regHL;
end;

function ZGetIX (): Word;
begin
  result := emuZ80.regIX;
end;

function ZGetIY (): Word;
begin
  result := emuZ80.regIY;
end;

function ZGetSP (): Word;
begin
  result := emuZ80.regSP;
end;

function ZGetA (): Word;
begin
  result := emuZ80.regA;
end;

function ZGetF (): Word;
begin
  result := Z80GetF(emuZ80);
end;
{$ENDIF}

function ZCmdLength (addr: Word): Integer;
var
  index, finito: Boolean;
  l, a, p: Integer;
begin
  finito := false;
  index := false;
  l := 0;
  a := addr;
  repeat
    p := ZPeekB(a);
    if (p = $DD) or (p = $FD) then
    begin
      if ZPeekB(a+1) in [$DD, $FD] then
      begin
        result := 1; exit;
      end;
      inc(l);
      inc(a);
      index := true;
    end
    else
    begin
      if p = $CB then
      begin
        Inc(l, 2);
        if index then Inc(l);
        finito := true;
      end
      else if p = $ED then
      begin
        Inc(l, 2);
        p := ZPeekB(a+1);
        if (p >= $40) and (p < $80) and ((p mod 8) = 3) then
        begin
          (* ld pq,(nn)  ld (nn),pq *)
          Inc(l, 2);
        end;
        finito := true;
      end
      else if (p < $40) and ((p mod 8) = 6) then
      begin
        (* ld r,n *)
        Inc(l, 2);
        if (p = $36) and index then Inc(l);
        finito := true;
      end
      else if (p < $40) and ((p mod 16) = 1) then
      begin
        (* ld pq,nn *)
        Inc(l, 3);
        finito := true;
      end
      else if (p < $40) and ((p mod 8) = 0) and (p > 8) then
      begin
        (* djnz , jr cc *)
        Inc(l, 2);
        finito := true;
      end
      else if (p = $22) or (p = $2A) then
      begin
        (*ld (nn),hl  ld hl,(nn) *)
        Inc(l, 3);
        finito := true;
      end
      else if (p = $32) or (p = $3A) then
      begin
        (* ld a,(nn)  ld (nn),a *)
        Inc(l, 3);
        finito := true;
      end
      else if p < $40 then
      begin
        Inc(l);
        if index and ((p = $34) or (p = $35)) then Inc(l);
        finito := true;
      end
      else if p < $80 then
      begin
        if index and (p <> $76) and (((p mod 8) = 6) or ((p >= $70) and (p < $78))) then Inc(l);
        Inc(l);
        finito := true;
      end
      else if p < $C0 then
      begin
        if index and ((p mod 8) = 6) then Inc(l);
        Inc(l);
        finito := true;
      end
      else if ((p mod 8) = 2) or ((p mod 8) = 4) or (p = $C3) or (p = $CD) then
      begin
        (* jp cc,nn  call cc,nn  jp nn  call nn *)
        Inc(l,3);
        finito := true;
      end
      else if ((p mod 8) = 6) or (p = $D3) or (p = $DB) then
      begin
        Inc(l,2);
        finito := true;
      end
      else
      begin
        Inc(l);
        finito := true;
      end;
    end;
  until finito;

  result := l;
end;

function Int2Hex (n, len: Integer): string;
begin
  result := '#'+IntToHex(n, len);
end;

procedure ZCmdDecode (addr: Word; out info: TUnasmInfo);
var
  ix, iy: Boolean;
  f, prefOfs: Integer;
  a, b, c, p: Integer;

  function Reg16 (reg: Integer; useSP, setAffect: Boolean): string;
  begin
    case reg of
      0: result := 'BC';
      1: result := 'DE';
      2: if ix then result := 'IX' else if iy then result := 'IY' else result := 'HL';
      3: if useSP then result := 'SP' else result := 'AF';
    end;

    if not setAffect then exit;
    case reg of
      0: info.affectRegs := unasmAffectB or unasmAffectC;
      1: info.affectRegs := unasmAffectD or unasmAffectE;
      2:
        if ix then info.affectRegs := unasmAffectHX or unasmAffectLX
        else if iy then info.affectRegs := unasmAffectHY or unasmAffectLY
        else info.affectRegs := unasmAffectH or unasmAffectL;
      3: info.affectRegs := unasmAffectSP;
    end;
  end;

  function Reg8 (reg: Integer; out addr: Integer; setAffect: Boolean): string;
  var
    f: Integer;
  begin
    addr := -1;
    case reg of
      0: begin result := 'B'; if setAffect then info.affectRegs := unasmAffectB; end;
      1: begin result := 'C'; if setAffect then info.affectRegs := unasmAffectC; end;
      2: begin result := 'D'; if setAffect then info.affectRegs := unasmAffectD; end;
      3: begin result := 'E'; if setAffect then info.affectRegs := unasmAffectE; end;
      4:
        begin
          result := 'H';
          if setAffect then
          begin
            if ix and setAffect then info.affectRegs := unasmAffectHX
            else if iy and setAffect then info.affectRegs := unasmAffectHY
            else info.affectRegs := unasmAffectH;
          end;
        end;
      5:
        begin
          result := 'L';
          if setAffect then
          begin
            if ix and setAffect then info.affectRegs := unasmAffectLX
            else if iy and setAffect then info.affectRegs := unasmAffectLY
            else info.affectRegs := unasmAffectL;
          end;
        end;
      6:
        begin
          if ix or iy then
          begin
            f := info.dump[2];
            if ix then
            begin
              addr := ZGetIX();
              result := '(IX';
            end
            else
            begin
              addr := ZGetIY();
              result := '(IY';
            end;
            if f > 127 then
            begin
              f := 256-f; result := result+'-';
            end
            else if f <> 0 then result := result+'+';

            if f <> 0 then result := result+'#'+IntToHex(f, 2);
            result := result+')';
            addr := (addr+f) and $FFFF;
          end
          else
          begin
            result := '(HL)'; addr := ZGetHL();
          end;
          if setAffect then
          begin
            info.ioFlags := info.ioFlags or unasmMemWriteFlag;
            info.memDestAddr := addr;
          end
          else
          begin
            info.ioFlags := info.ioFlags or unasmMemReadFlag;
            info.memSrcAddr := addr;
          end;
        end;
      7: begin result := 'A'; if setAffect then info.affectRegs := unasmAffectA; end;
    end;

    if (reg = 4) or (reg = 5) then
    begin
      // lx or hx
      if ix or iy then result := result+'X';
    end;
  end;

  function Cond (cc: Integer): string;
  // sets info.isDoCond too
  var
    f: Byte;
  begin
    f := ZGetF();
    case cc of
      0: begin result := 'NZ'; info.isDoCond := (f and Z80FLAGZ) <> 0; end;
      1: begin result := 'Z'; info.isDoCond := (f and Z80FLAGZ) = 0; end;
      2: begin result := 'NC'; info.isDoCond := (f and Z80FLAGC) <> 0; end;
      3: begin result := 'C'; info.isDoCond := (f and Z80FLAGC) = 0; end;
      4: begin result := 'PO'; info.isDoCond := (f and Z80FLAGPV) <> 0; end;
      5: begin result := 'PE'; info.isDoCond := (f and Z80FLAGPV) = 0; end;
      6: begin result := 'P'; info.isDoCond := (f and Z80FLAGS) = 0; end;
      7: begin result := 'M'; info.isDoCond := (f and Z80FLAGS) <> 0; end;
    end;
  end;

begin
  // clear info
  info.valid := false;
  info.addr := addr;
  info.len := ZCmdLength(addr);
  for f := 0 to 7 do info.dump[f] := ZPeekB(addr+f);
  info.cmdName := '???';
  info.argC := 0;
  info.args[0] := '';
  info.args[1] := '';
  info.args[2] := '';
  info.argsN[0] := -1;
  info.argsN[1] := -1;
  info.argsN[2] := -1;
  info.memRefArg := -1;
  info.isBranch := false;
  info.isJump := false;
  info.isRet := false;
  info.isCall := false;
  info.isRst := false;
  info.isCond := false;
  info.isDoCond := true;
  info.branchDest := -1;
  info.dataSize := 0;
  info.ioFlags := 0;
  info.memSrcAddr := -1;
  info.memDestAddr := -1;
  info.portAddr := -1;
  info.affectRegs := 0;

  p := info.dump[0]; ix := (p = $DD); iy := (p = $FD);
  if ix or iy then
  begin
    p := info.dump[1];
    if (p = $DD) or (p = $FD) or (p = $ED) then
    begin
      // double prefix or ID/#ED
      info.cmdName := Int2Hex(info.dump[0], 2);
      exit;
    end;
    prefOfs := 1;
  end
  else prefOfs := 0;
  if p <> $ED then info.valid := true;

  a := p div 64; b := (p mod 64) div 8; c := p mod 8;
  case a of
    0:
      case c of
        0:
          begin
            case b of
              0: info.cmdName := 'NOP';
              1:
                begin
                  info.cmdName := 'EX';
                  info.argC := 2;
                  info.args[0] := 'AF';
                  info.args[1] := 'AF'#39;
                  info.affectRegs := unasmAffectA or unasmAffectF;
                end;
              2: info.cmdName := 'DJNZ';
              3: info.cmdName := 'JR';
              else
                info.cmdName := 'JR';
                info.argC := 1;
                info.args[0] := Cond(b-4);
            end;
            if b > 1 then
            begin
              info.isBranch := true;
              info.isJump := true;
              info.isCond := (b <> 3);
              if b = 2 then
              begin
                info.affectRegs := unasmAffectB;
                info.isDoCond := (ZGetBC() and $FF) <> 1;
              end;
              // relative jump disp
              f := info.dump[1+prefOfs];
              if f > 127 then f := f+info.addr-254 else f := f+info.addr+2;
              f := f and $FFFF;
              info.branchDest := f;
              info.argsN[info.argC] := f;
              info.args[info.argC] := Int2Hex(f, 4);
              Inc(info.argC);
            end;
          end;
        1:
          begin
            info.dataSize := 2;
            if (b mod 2) = 1 then
            begin
              // ADD HL,rr
              info.cmdName := 'ADD';
              info.argC := 2;
              info.args[0] := Reg16(2, true, true);
              info.args[1] := Reg16(b div 2, true, false);
            end
            else
            begin
              // LD rr,nn
              info.cmdName := 'LD';
              info.argC := 2;
              info.args[0] := Reg16(b div 2, true, true);
              f := 1+prefOfs;
              info.argsN[1] := info.dump[f]+256*info.dump[f+1];
              info.args[1] := Int2Hex(info.argsN[1], 4);
            end;
          end;
        2:
          begin
            info.dataSize := 1;
            info.cmdName := 'LD';
            info.argC := 2;
            case b of
              0: // LD (BC),A
                begin
                  info.args[0] := '(BC)';
                  info.args[1] := 'A';
                  info.ioFlags := unasmMemWriteFlag;
                  info.memDestAddr := ZGetBC();
                  info.affectRegs := unasmAffectB or unasmAffectC;
                end;
              1: // LD A,(BC)
                begin
                  info.args[0] := 'A';
                  info.args[1] := '(BC)';
                  info.ioFlags := unasmMemReadFlag;
                  info.memSrcAddr := ZGetBC();
                  info.affectRegs := unasmAffectA;
                end;
              2: // LD (DE),A
                begin
                  info.args[0] := '(DE)';
                  info.args[1] := 'A';
                  info.ioFlags := unasmMemWriteFlag;
                  info.memDestAddr := ZGetDE();
                  info.affectRegs := unasmAffectD or unasmAffectE;
                end;
              3: // LD A,(DE)
                begin
                  info.args[0] := 'A';
                  info.args[1] := '(DE)';
                  info.ioFlags := unasmMemReadFlag;
                  info.memSrcAddr := ZGetDE();
                  info.affectRegs := unasmAffectA;
                end;
              4: // LD (nn),HL
                begin
                  info.dataSize := 2;
                  info.ioFlags := unasmMemWriteFlag;
                  if ix or iy then info.memDestAddr := info.dump[2]+256*info.dump[3]
                  else info.memDestAddr := info.dump[1]+256*info.dump[2];
                  info.argsN[0] := info.memDestAddr;
                  info.args[0] := '('+Int2Hex(info.argsN[0], 4)+')';
                  info.memRefArg := 0;
                  info.args[1] := Reg16(2, true, false);
                end;
              5: // LD HL,(nn)
                begin
                  info.dataSize := 2;
                  info.ioFlags := unasmMemReadFlag;
                  info.args[0] := Reg16(2, true, true);
                  if ix or iy then info.memSrcAddr := info.dump[2]+256*info.dump[3]
                  else info.memSrcAddr := info.dump[1]+256*info.dump[2];
                  info.argsN[1] := info.memSrcAddr;
                  info.args[1] := '('+Int2Hex(info.argsN[1], 4)+')';
                  info.memRefArg := 1;
                end;
              6: // LD (nn),A
                begin
                  info.ioFlags := unasmMemWriteFlag;
                  if ix or iy then info.memDestAddr := info.dump[2]+256*info.dump[3]
                  else info.memDestAddr := info.dump[1]+256*info.dump[2];
                  info.argsN[0] := info.memDestAddr;
                  info.args[0] := '('+Int2Hex(info.argsN[0], 2)+')';
                  info.memRefArg := 0;
                  info.args[1] := 'A';
                end;
              7: // LD A,(nn)
                begin
                  info.ioFlags := unasmMemReadFlag;
                  info.args[0] := 'A';
                  if ix or iy then info.memSrcAddr := info.dump[2]+256*info.dump[3]
                  else info.memSrcAddr := info.dump[1]+256*info.dump[2];
                  info.argsN[1] := info.memSrcAddr;
                  info.args[1] := '('+Int2Hex(info.argsN[1], 2)+')';
                  info.memRefArg := 1;
                  info.affectRegs := unasmAffectA;
                end;
            end;
          end;
        3: // INC/DEC rr
          begin
            info.dataSize := 2;
            info.argC := 1;
            info.args[0] := Reg16(b div 2, true, true);
            if (b mod 2) = 0 then info.cmdName := 'INC' else info.cmdName := 'DEC';
          end;
        4, 5: // INC/DEC r
          begin
            info.dataSize := 1;
            info.argC := 1;
            info.args[0] := Reg8(b, f, true);
            if c = 4 then info.cmdName := 'INC' else info.cmdName := 'DEC';
            info.affectRegs := info.affectRegs or unasmAffectF;
          end;
        6: // LD r,n
          begin
            info.cmdName := 'LD';
            info.dataSize := 1;
            info.argC := 2;
            info.args[0] := Reg8(b, f, true);
            if (ix or iy) and (b = 6) then f := 3
            else if (ix or iy) then f := 2 else f := 1; //!BFIX
            info.argsN[1] := info.dump[f];
            info.args[1] := Int2Hex(info.argsN[1], 2);
          end;
        7:
          begin
            if b < 6 then info.affectRegs := unasmAffectA or unasmAffectF
            else info.affectRegs := unasmAffectF;
            case b of
              0: info.cmdName := 'RLCA';
              1: info.cmdName := 'RRCA';
              2: info.cmdName := 'RLA';
              3: info.cmdName := 'RRA';
              4: info.cmdName := 'DAA';
              5: info.cmdName := 'CPL';
              6: info.cmdName := 'SCF';
              7: info.cmdName := 'CCF';
            end;
          end;
      end;
    ///////////////////////////////////////////////
    1:
      if p = $76 then
      begin
        info.cmdName := 'HALT';
        if ix or iy then begin info.valid := false; info.cmdName := '*HALT'; end;
      end
      else
      begin
        // LD r,r
        info.cmdName := 'LD';
        info.dataSize := 1;
        info.argC := 2;
        if c = 6 then
        begin
          // LD r,(HL)
          if ix then
          begin
            ix := false;
            info.args[0] := Reg8(b, f, true);
            ix := true;
          end
          else if iy then
          begin
            iy := false;
            info.args[0] := Reg8(b, f, true);
            iy := true;
          end
          else info.args[0] := Reg8(b, f, true);
        end
        else
        begin
          info.args[0] := Reg8(b, f, true);
          if b = 6 then
          begin
            // LD (HL),r
            ix := false; iy := false;
          end;
        end;

        info.args[1] := Reg8(c, f, false);
      end;
    ///////////////////////////////////////////////
    2: // ALU A,r
      begin
        info.dataSize := 1;
        info.argC := 2;
        case b of
          0: info.cmdName := 'ADD';
          1: info.cmdName := 'ADC';
          2: info.cmdName := 'SUB';
          3: info.cmdName := 'SBC';
          4: info.cmdName := 'AND';
          5: info.cmdName := 'XOR';
          6: info.cmdName := 'OR';
          7: info.cmdName := 'CP';
        end;
        info.affectRegs := unasmAffectA or unasmAffectF;
        info.args[0] := 'A';
        info.args[1] := Reg8(c, f, false);
      end;
    ///////////////////////////////////////////////
    3:
      case c of
        0: // RET CC
          begin
            info.cmdName := 'RET';
            info.argC := 1;
            info.args[0] := Cond(b);
            info.isBranch := true;
            info.isRet := true;
            info.isCond := true;
            info.branchDest := ZPeekW(ZGetSP());
          end;
        1:
          case b of
            1: // RET
              begin
                info.cmdName := 'RET';
                info.isBranch := true;
                info.isRet := true;
                info.branchDest := ZPeekW(ZGetSP());
              end;
            3: // EXX
              begin
                info.cmdName := 'EXX';
                info.affectRegs := unasmAffectB or unasmAffectC or
                  unasmAffectD or unasmAffectE or unasmAffectH or unasmAffectL;
              end;
            5: // JP (HL)
              begin
                info.cmdName := 'JP';
                info.isBranch := true;
                info.isJump := true;
                info.argC := 1;
                info.args[0] := '('+Reg16(2, false, false)+')';
                if ix then info.branchDest := ZGetIX()
                else if iy then info.branchDest := ZGetIY()
                else info.branchDest := ZGetHL();
              end;
            7: // LD SP,HL
              begin
                info.dataSize := 2;
                info.cmdName := 'LD';
                info.argC := 2;
                info.args[0] := 'SP';
                info.args[1] := Reg16(2, false, true);
                info.affectRegs := info.affectRegs or unasmAffectSP;
              end
            else
              // POP rr
              info.dataSize := 2;
              info.cmdName := 'POP';
              info.argC := 1;
              info.args[0] := Reg16(b div 2, false, true);
              info.affectRegs := info.affectRegs or unasmAffectSP;
              info.ioFlags := unasmMemReadFlag;
              info.memSrcAddr := ZGetSP();
          end;
        2: // JP CC,nn
          begin
            info.cmdName := 'JP';
            info.isBranch := true;
            info.isJump := true;
            info.isCond := true;
            info.argC := 2;
            info.args[0] := Cond(b);
            info.branchDest := info.dump[1+prefOfs]+256*info.dump[2+prefOfs];
            info.argsN[1] := info.branchDest;
            info.args[1] := Int2Hex(info.branchDest, 4);
          end;
        3:
          case b of
            0: // JP nn
              begin
                info.cmdName := 'JP';
                info.isBranch := true;
                info.isJump := true;
                info.argC := 1;
                info.branchDest := info.dump[1+prefOfs]+256*info.dump[2+prefOfs];
                info.argsN[0] := info.branchDest;
                info.args[0] := Int2Hex(info.branchDest, 4);
              end;
            1: // #CBxx
              begin
                info.dataSize := 1;
                if ix or iy then p := info.dump[3] else p := info.dump[1];
                a := p div 64; b := (p div 8) mod 8; c := p mod 8;
                case a of
                  0:
                    begin
                      case b of
                        0: info.cmdName := 'RLC';
                        1: info.cmdName := 'RRC';
                        2: info.cmdName := 'RL';
                        3: info.cmdName := 'RR';
                        4: info.cmdName := 'SLA';
                        5: info.cmdName := 'SRA';
                        6: info.cmdName := 'SLL';
                        7: info.cmdName := 'SRL';
                      end;
                      info.argC := 1;
                      info.args[0] := Reg8(c, f, true);
                      if c = 6 then // (HL)
                      begin
                        info.ioFlags := info.ioFlags or unasmMemReadFlag;
                        info.memSrcAddr := info.memDestAddr;
                      end
                      else if ix or iy then
                      begin
                        info.argC := 2;
                        info.args[1] := info.args[0];
                        info.args[0] := Reg8(6, f, false);
                          // adds "memory read" flag
                        info.ioFlags := info.ioFlags or unasmMemWriteFlag;
                        info.memDestAddr := info.memSrcAddr;
                      end;
                      exit;
                    end;
                  1: info.cmdName := 'BIT';
                  2: info.cmdName := 'RES';
                  3: info.cmdName := 'SET';
                end;
                info.argC := 2;
                info.args[0] := chr(b+ord('0'));
                info.argsN[0] := b;
                if ix or iy  then
                begin
                  info.args[1] := Reg8(6, f, true);
                  if (c <> 6) and (a <> 1) then
                  begin
                    Inc(info.argC);
                    info.args[2] := Reg8(c, f, false);
                  end;
                  info.ioFlags := info.ioFlags or unasmMemWriteFlag;
                  info.memDestAddr := info.memSrcAddr;
                end
                else info.args[1] := Reg8(c, f, true);
              end;
            2: // OUT (n),A
              begin
                info.dataSize := 1;
                info.argC := 2;
                info.args[1] := 'A';
                info.cmdName := 'OUT';
                info.ioFlags := unasmPortWriteFlag;
                info.portAddr := info.dump[1+prefOfs]+256*ZGetA();
                info.args[0] := '('+Int2Hex(info.portAddr, 2)+')';
                info.argsN[0] := info.portAddr;
              end;
            3: // IN A,(n)
              begin
                info.dataSize := 1;
                info.affectRegs := unasmAffectA;
                info.argC := 2;
                info.args[0] := 'A';
                info.cmdName := 'IN';
                info.ioFlags := unasmPortReadFlag;
                info.portAddr := info.dump[1+prefOfs]+256*ZGetA();
                info.args[1] := '('+Int2Hex(info.portAddr, 2)+')';
                info.argsN[1] := info.portAddr;
              end;
            4: // EX (SP),HL
              begin
                info.cmdName := 'EX';
                info.dataSize := 2;
                info.ioFlags := info.ioFlags or unasmMemWriteFlag;
                info.memDestAddr := ZGetSP();
                info.argC := 2;
                info.args[0] := '(SP)';
                info.args[1] := Reg16(2, false, true);
              end;
            5: // EX DE,HL
              begin
                info.cmdName := 'EX';
                info.dataSize := 2;
                info.argC := 2;
                info.args[0] := 'DE';
                info.args[1] := Reg16(2, false, true);
              end;
            6: info.cmdName := 'DI';
            7: info.cmdName := 'EI';
          end;
        4: // CALL CC,nn
          begin
            info.cmdName := 'CALL';
            info.isBranch := true;
            info.isCall := true;
            info.isCond := true;
            info.argC := 2;
            info.args[0] := Cond(b);
            info.branchDest := info.dump[1+prefOfs]+256*info.dump[2+prefOfs];
            info.argsN[1] := info.branchDest;
            info.args[1] := Int2Hex(info.branchDest, 4);
          end;
        5:
          case b of
            1: // CALL nn
              begin
                info.cmdName := 'CALL';
                info.isBranch := true;
                info.isCall := true;
                info.argC := 1;
                info.branchDest := info.dump[1+prefOfs]+256*info.dump[2+prefOfs];
                info.argsN[0] := info.branchDest;
                info.args[0] := Int2Hex(info.branchDest, 4);
              end;
            3: ;(* dd-prefix -- the thing that should not be *)
            5: (* ed-prefix *)
              begin
                if ix or iy then p := ZPeekB(info.addr+2) else p := ZPeekB(info.addr+1);
                if p = $FB then
                begin
                  info.valid := true;
                  info.dataSize := 0;
                  info.argC := 0;
                  info.cmdName := 'XSLT';
                  info.affectRegs := info.affectRegs or unasmAffectF;
                end
                else if p = $06 then
                begin
                  info.valid := true;
                  info.dataSize := 0;
                  info.argC := 0;
                  info.cmdName := 'XLSP';
                end
                else
                begin
                  a := p div 64; b := (p div 8) mod 8; c := p mod 8;
                  case a of
                    1:
                      case c of
                        0: // IN r,(C)
                          begin
                            info.valid := true;
                            info.dataSize := 1;
                            info.argC := 2;
                            info.cmdName := 'IN';
                            info.ioFlags := unasmPortReadFlag;
                            info.portAddr := ZGetBC();
                            info.args[1] := '(C)';
                            if b = 6 then info.args[0] := 'F'
                            else info.args[0] := Reg8(b, f, true);
                            info.affectRegs := info.affectRegs or unasmAffectF;
                          end;
                        1: // OUT (C),r
                          begin
                            info.valid := true;
                            info.dataSize := 1;
                            info.argC := 2;
                            info.cmdName := 'OUT';
                            info.args[0] := '(C)';
                            info.ioFlags := unasmPortWriteFlag;
                            info.portAddr := ZGetBC();
                            if b = 6 then info.args[1] := 'F'
                            else info.args[1] := Reg8(b, f, false);
                          end;
                        2: // ADC HL,rr SBC HL,rr
                          begin
                            info.valid := true;
                            info.dataSize := 2;
                            info.argC := 2;
                            if (b mod 2) = 0 then info.cmdName := 'SBC'
                            else info.cmdName := 'ADC';
                            info.args[0] := Reg16(2, true, true);
                            info.args[1] := Reg16(b div 2, true, false);
                          end;
                        3: // LD (nn),rr / LD rr,(nn)
                          begin
                            info.valid := true;
                            info.dataSize := 2;
                            info.argC := 2;
                            info.cmdName := 'LD';
                            info.args[1] := Reg16(b div 2, true, false);
                            if (b mod 2) = 1 then
                            begin
                              info.args[0] := Reg16(b div 2, true, true);
                              info.argsN[1] := info.dump[2]+256*info.dump[3];
                              info.args[1] := '('+Int2Hex(info.argsN[1], 4)+')';
                              info.ioFlags := unasmMemReadFlag;
                              info.memSrcAddr := info.argsN[1];
                            end
                            else
                            begin
                              info.argsN[0] := info.dump[2]+256*info.dump[3];
                              info.args[0] := '('+Int2Hex(info.argsN[0], 4)+')';
                              info.args[1] := Reg16(b div 2, true, false);
                              info.ioFlags := unasmMemWriteFlag;
                              info.memDestAddr := info.argsN[1];
                            end;
                          end;
                        4:
                          if b = 0 then
                          begin
                            info.valid := true;
                            info.cmdName := 'NEG';
                            info.affectRegs := unasmAffectA or unasmAffectF;
                          end;
                        5:
                          case b of
                            0: begin info.cmdName := 'RETN'; info.valid := true; end;
                            1: begin info.cmdName := 'RETI'; info.valid := true; end;
                          end;
                        6:
                          case b of
                            0:
                              begin
                                info.valid := true;
                                info.cmdName := 'IM';
                                info.argC := 1;
                                info.argsN[0] := 0;
                                info.args[0] := '0';
                              end;
                            2:
                              begin
                                info.valid := true;
                                info.cmdName := 'IM';
                                info.argC := 1;
                                info.argsN[0] := 1;
                                info.args[0] := '1';
                              end;
                            3:
                              begin
                                info.valid := true;
                                info.cmdName := 'IM';
                                info.argC := 1;
                                info.argsN[0] := 2;
                                info.args[0] := '2';
                              end;
                          end;
                        7:
                          case b of
                            0:
                              begin
                                info.valid := true;
                                info.cmdName := 'LD';
                                info.argC := 2;
                                info.args[0] := 'I';
                                info.args[1] := 'A';
                                info.affectRegs := unasmAffectI;
                              end;
                            1:
                              begin
                                info.valid := true;
                                info.cmdName := 'LD';
                                info.argC := 2;
                                info.args[0] := 'R';
                                info.args[1] := 'A';
                                info.affectRegs := unasmAffectR;
                              end;
                            2:
                              begin
                                info.valid := true;
                                info.cmdName := 'LD';
                                info.argC := 2;
                                info.args[0] := 'A';
                                info.args[1] := 'I';
                                info.affectRegs := unasmAffectA or unasmAffectF;
                              end;
                            3:
                              begin
                                info.valid := true;
                                info.cmdName := 'LD';
                                info.argC := 2;
                                info.args[0] := 'A';
                                info.args[1] := 'R';
                                info.affectRegs := unasmAffectA or unasmAffectF;
                              end;
                            4, 5:
                              begin
                                info.valid := true;
                                if b = 4 then info.cmdName := 'RRD'
                                else info.cmdName := 'RLD';
                                info.affectRegs := unasmAffectA or unasmAffectF;
                                info.ioFlags := unasmMemReadFlag or unasmMemWriteFlag;
                                info.memSrcAddr := ZGetHL();
                              end;
                          end;
                       end;
                    2:
                      begin
                        // !!! TODO: WRITE FLAG SETTINGS!
                        info.valid := c < 4;
                        case b of
                          4:
                            case c of
                              0:
                                begin
                                  info.cmdName := 'LDI';
                                  info.ioFlags := unasmMemReadFlag or unasmMemWriteFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.memDestAddr := ZGetDE();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectD or unasmAffectE or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB or unasmAffectC;
                                end;
                              1:
                                begin
                                  info.cmdName := 'CPI';
                                  info.ioFlags := unasmMemReadFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB or unasmAffectC;
                                end;
                              2:
                                begin
                                  info.cmdName := 'INI';
                                  info.ioFlags := unasmMemWriteFlag or unasmPortReadFlag;
                                  info.memDestAddr := ZGetHL();
                                  info.portAddr := ZGetBC();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB;
                                end;
                              3:
                                begin
                                  info.cmdName := 'OUTI';
                                  info.ioFlags := unasmMemReadFlag or unasmPortWriteFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.portAddr := ZGetBC();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB;
                                end;
                            end;
                          5:
                            case c of
                              0:
                                begin
                                  info.cmdName := 'LDD';
                                  info.ioFlags := unasmMemReadFlag or unasmMemWriteFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.memDestAddr := ZGetDE();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectD or unasmAffectE or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB or unasmAffectC;
                                end;
                              1:
                                begin
                                  info.cmdName := 'CPD';
                                  info.ioFlags := unasmMemReadFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB or unasmAffectC;
                                end;
                              2:
                                begin
                                  info.cmdName := 'IND';
                                  info.ioFlags := unasmMemWriteFlag or unasmPortReadFlag;
                                  info.memDestAddr := ZGetHL();
                                  info.portAddr := ZGetBC();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB;
                                end;
                              3:
                                begin
                                  info.cmdName := 'OUTD';
                                  info.ioFlags := unasmMemReadFlag or unasmPortWriteFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.portAddr := ZGetBC();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB;
                                end;
                            end;
                          6:
                            case c of
                              0:
                                begin
                                  info.cmdName := 'LDIR';
                                  info.ioFlags := unasmMemReadFlag or unasmMemWriteFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.memDestAddr := ZGetDE();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectD or unasmAffectE or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB or unasmAffectC;
                                end;
                              1:
                                begin
                                  info.cmdName := 'CPIR';
                                  info.ioFlags := unasmMemReadFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB or unasmAffectC;
                                end;
                              2:
                                begin
                                  info.cmdName := 'INIR';
                                  info.ioFlags := unasmMemWriteFlag or unasmPortReadFlag;
                                  info.memDestAddr := ZGetHL();
                                  info.portAddr := ZGetBC();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB;
                                end;
                              3:
                                begin
                                  info.cmdName := 'OTIR';
                                  info.ioFlags := unasmMemReadFlag or unasmPortWriteFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.portAddr := ZGetBC();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB;
                                end;
                            end;
                          7:
                            case c of
                              0:
                                begin
                                  info.cmdName := 'LDDR';
                                  info.ioFlags := unasmMemReadFlag or unasmMemWriteFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.memDestAddr := ZGetDE();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectD or unasmAffectE or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB or unasmAffectC;
                                end;
                              1:
                                begin
                                  info.cmdName := 'CPDR';
                                  info.ioFlags := unasmMemReadFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB or unasmAffectC;
                                end;
                              2:
                                begin
                                  info.cmdName := 'INDR';
                                  info.ioFlags := unasmMemWriteFlag or unasmPortReadFlag;
                                  info.memDestAddr := ZGetHL();
                                  info.portAddr := ZGetBC();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB;
                                end;
                              3:
                                begin
                                  info.cmdName := 'OTDR';
                                  info.ioFlags := unasmMemReadFlag or unasmPortWriteFlag;
                                  info.memSrcAddr := ZGetHL();
                                  info.portAddr := ZGetBC();
                                  info.affectRegs := unasmAffectF or
                                    unasmAffectH or unasmAffectL or
                                    unasmAffectB;
                                end;
                            end;
                        end;
                      end;
                  end;
                end;
              end;
            7: ; (* fd-prefix -- the thing that should not be *)
            else // PUSH rr
            begin
              info.dataSize := 2;
              info.cmdName := 'PUSH';
              info.argC := 1;
              info.args[0] := Reg16(b div 2, false, false);
              info.affectRegs := info.affectRegs or unasmAffectSP;
              info.ioFlags := unasmMemWriteFlag;
              info.memSrcAddr := ZGetSP();
            end;
          end;
        6: // ALU A,n
          begin
            info.dataSize := 1;
            info.argC := 2;
            case b of
              0: info.cmdName := 'ADD';
              1: info.cmdName := 'ADC';
              2: info.cmdName := 'SUB';
              3: info.cmdName := 'SBC';
              4: info.cmdName := 'AND';
              5: info.cmdName := 'XOR';
              6: info.cmdName := 'OR';
              7: info.cmdName := 'CP';
            end;
            info.affectRegs := unasmAffectA or unasmAffectF;
            info.args[0] := 'A';
            info.argsN[1] := info.dump[1+prefOfs];
            info.args[1] := Int2Hex(info.argsN[1], 2);
          end;
        7:
          begin
            info.argC := 1;
            info.cmdName := 'RST';
            info.isBranch := true;
            info.isRst := true;
            info.branchDest := b*8;
            info.argsN[0] := b*8;
            info.args[0] := Int2Hex(b*8, 2);
          end;
      end;
  end;
end;


end.
