// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// main emulation procedures, win32 message handler
//
{$INCLUDE options.inc}
unit dspEmuU;

interface

uses
  Windows, Messages, SysUtils, Classes,
  lispCon3U,
  z80U, dspDbgU, dspEmuVarsU, dspMachineU,
  dspMachConstU, dspVScrU, dspDDrawU, dspZXDrawU, dspWinU,
  dspEmuSoundU, dspEmuMiscU, dspEmuMouseU, dspKeyNamesU
  {$IFDEF USE_HP_TIMER},hpTImerU{$ENDIF};


procedure CreateEmuWindow (fullScreen: Boolean; scale: Integer);

procedure EmuExec (); // one intr
  // emuExec can exit before intr (on certain event -- pausing, for
  // example). the next EmuExec() will continue building the
  // interrupted frame.

var
  __logoWH: HWND = 0; // 窠  誨  


implementation


{$INCLUDE btime.inc}

const
  winClassName = 'DarkSpecterMainWindowClass';


var
  waitTimeLeft: Double = 0;
  keybdOn: Boolean = true;


function WindowProc (winH: HWnd; msg, wParam: LongWord; lParam: LongInt): LongInt; stdcall;
{var
  f: Integer;}
begin
  if __logoWH <> 0 then
  begin
    if (msg = WM_KEYDOWN) or (msg = WM_SYSKEYDOWN) or
       (msg = WM_LBUTTONDOWN) or (msg = WM_RBUTTONDOWN) or
       (msg = WM_MBUTTONDOWN) then PostMessage(__logoWH, WM_CLOSE, 0, 0);
  end;
  result := 0;
  case msg of
    WM_CLOSE: begin PostQuitMessage(0); exit; end;
    WM_QUIT: PostQuitMessage(wParam);
    WM_ERASEBKGND: begin result := 1; exit; end;
    WM_KILLFOCUS:
      begin
        if config.pauseOnDefocus then ConExecute('temp_pause tan');
        EmuFixShiftsState();
      end;
    WM_SETFOCUS:
      begin
        EmuFixShiftsState();
        ConExecute('temp_pause ona');
        EmuFixShiftsState();
      end;
    WM_LBUTTONDOWN: begin emuMouseButtons := emuMouseButtons or 1; exit; end;
    WM_LBUTTONUP: begin emuMouseButtons := emuMouseButtons and (not 1); exit; end;
    WM_RBUTTONDOWN: begin emuMouseButtons := emuMouseButtons or 2; exit; end;
    WM_RBUTTONUP: begin emuMouseButtons := emuMouseButtons and (not 2); exit; end;
    WM_MBUTTONDOWN: begin emuMouseButtons := emuMouseButtons or 4; exit; end;
    WM_MBUTTONUP: begin emuMouseButtons := emuMouseButtons and (not 4); exit; end;
    WM_MOUSEMOVE:
      begin
        emuMouseX := LoWord(lParam) div emuScale;
        emuMouseY := HiWord(lParam) div emuScale;
        exit;
      end;
    WM_KEYDOWN:
      begin
        EmuFixShiftsState();
        //WriteLn('dlParam: $', IntToHex(lParam, 0));
        case wParam and $FFFF of
          VK_SHIFT, VK_CONTROL, VK_MENU: EmuFixShiftsState();
          VK_SPACE:
            if emuShiftsState = emuSSAlt then
            begin
              emuShiftsState := 0;
              WriteLn('MENU!');
              PostMessage(winH, WM_SYSCOMMAND, SC_KEYMENU, 0);
              exit;
            end;
          VK_INSERT:
            if emuShiftsState = emuSSShift then
            begin
              emuConInputLine := emuConInputLine+ConGetClipboardString(winH);
            end
            else
            if emuShiftsState = emuSSCtrl then
            begin
              ConSetClipboardString(emuConInputLine, winH);
            end;
          VK_PRIOR:
            begin
              if emuShiftsState = emuSSCtrl then Inc(conScrollDelta, config.conHeight-2)
              else Inc(conScrollDelta);
            end;
          VK_NEXT:
            begin
              if emuShiftsState = emuSSCtrl then Dec(conScrollDelta, config.conHeight-2)
              else Dec(conScrollDelta);
              if conScrollDelta < 0 then conScrollDelta := 0;
            end;
        end;

        PostMessage(winH, WM_USER+1, wParam and $FFFF, emuShiftsState);

        if emuShiftsState = emuSSCtrl or emuSSAlt or emuSSShift then
          WriteLn('t-states: ', emuZ80.tStates, '; intr at: ', emuTicksIntr,
            '; one line: ', emuTicksLine);
        exit;
      end;
    WM_KEYUP:
      begin
        //WriteLn('ulParam: $', IntToHex(lParam, 0));
        case wParam and $FFFF of
          VK_SHIFT, VK_CONTROL, VK_MENU: EmuFixShiftsState();
        end;
        exit;
      end;

    WM_PAINT:
      begin
        result := DefWindowProc(winH, msg, wParam, lParam);
        {if emuUseDDraw then DDrawUpdate() else }EmuDrawScreen();
        exit;
      end;
  end;
  result := DefWindowProc(winH, msg, wParam, lParam);
end;


function AsyncShiftsState (): Integer;
begin
  result := 0;
  if (GetAsyncKeyState(VK_MENU) and $80) <> 0 then result := result or emuSSAlt;
  if (GetAsyncKeyState(VK_CONTROL) and $80) <> 0 then result := result or emuSSCtrl;
  if (GetAsyncKeyState(VK_SHIFT) and $80) <> 0 then result := result or emuSSShift;
end;

function AltDebugActivate (): Boolean;
begin
  if emuAltDebugKey = 0 then result := false
  else
  begin
    //result := (GetAsyncKeyState(emuAltDebugKey) and $8000) <> 0;
    result := (GetKeyState(emuAltDebugKey) and $8000) <> 0;
    if not result then exit;
    //result := (AsyncShiftsState = emuAltDebugShift);
    EmuFixShiftsState();
    result := (emuShiftsState = emuAltDebugShift);
  end;
end;


procedure InitWndClass ();
var
  windowClass: Windows.TWndClass;
begin
  windowClass.style := CS_HREDRAW or CS_VREDRAW or CS_OWNDC{ or CS_DBLCLKS};
  windowClass.lpfnWndProc := @WindowProc;
  windowClass.cbClsExtra := 0;
  windowClass.cbWndExtra := 0;
  windowClass.hInstance := 0;
  windowClass.hIcon := LoadIcon(hInstance, 'MAINICON');
  windowClass.hCursor := LoadCursor(0, IDC_ARROW);
  windowClass.hbrBackground := 0;
  windowClass.lpszMenuName := '';
  windowClass.lpszClassName := winClassName;

  if Windows.RegisterClass(windowClass) = 0 then
  begin
    WriteLn('ERROR: can''t register window class.');
    Halt(1);
  end;
end;


procedure CreateEmuWindow (fullScreen: Boolean; scale: Integer);
var
  wr, wa: TRect;
  est, st: LongWord;
  dmScreenSettings: DEVMODE;
  res: Integer;
begin
  EmuSoundDeinit();

  if ddrawLocked then DDrawReleaseScr();
  DDrawDeinit();

  EmuUnlockMouse();
  EmuShowMouse();
  if scale < 1 then scale := 1 else if scale > 2 then scale := 2;

  if emuHWnd <> 0 then DestroyWindow(emuHWnd);

  if fullScreen then
  begin
    ChangeDisplaySettings(devmode(nil^), 0);
    emuFullScreen := false;
    //if not emuFullScreen then
    begin
      WriteLn('going fullscreen...');
      ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));
      with dmScreenSettings do
      begin
        dmSize := SizeOf(dmScreenSettings);
        if scale = 1 then begin dmPelsWidth := 320; dmPelsHeight := 240; end
        else begin dmPelsWidth := 640; dmPelsHeight := 480; end;
        //dmPelsWidth := 320; dmPelsHeight := 240;
        //dmPelsWidth := 640; dmPelsHeight := 480;
        //if emuMode = dmDDraw then dmBitsPerPel := 32 else dmBitsPerPel := 8;
        dmBitsPerPel := 32;
        dmFields := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
      end;

      // Try to change screen mode to fullscreen
      if ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED then
      begin
        WriteLn('unable to switch to fullscreen %-(');
      end
      else
      begin
        emuFullScreen := true;
        EmuHideMouse();
        //scale := 2;
        //scale := 1;
        WriteLn('fullscreen OK...');
      end;
    end;
  end
  else
  begin
    if emuFullScreen then
    begin
      WriteLn('going normal...');
      emuFullScreen := false;
      //scale := 2;
    end;
  end;

  if emuFullScreen then
  begin
    est := WS_EX_TOPMOST or WS_EX_APPWINDOW;
    st := WS_POPUP{ or WS_VISIBLE};
  end
  else
  begin
    est := WS_EX_APPWINDOW or WS_EX_WINDOWEDGE;
    st := WS_CAPTION or WS_MINIMIZEBOX or WS_SYSMENU{ or WS_VISIBLE};
  end;

  wr.Left := 0; wr.Top := 0;
  wr.Right := 320*scale;
  wr.Bottom := 240*scale;
  //WriteLn(wr.Right, 'x', wr.Bottom);
  AdjustWindowRect(wr, st, false);
  //WriteLn('(', wr.Left, ', ', wr.Top, ')-(', wr.Right, ', ', wr.Bottom, ')');
  Dec(wr.Right, wr.Left); Dec(wr.Bottom, wr.Top);
  emuWWidth := wr.Right; emuWHeight := wr.Bottom;

  if not emuFullScreen then
  begin
    SystemParametersInfo(SPI_GETWORKAREA, 0, @wa, 0);
    wa.Left := wa.Left+(wa.Right-wa.Left-emuWWidth) div 2;
    wa.Top := wa.Top+(wa.Bottom-wa.Top-emuWHeight) div 2;
  end
  else
  begin
    wa.Left := 0; wa.Top := 0;
  end;

  //wa.Left := 0; //wa.Top := 100; //!!!
  emuHWnd := Windows.CreateWindowEx(
    est,
    winClassName, PChar(emuName+' -- '+BUILD_DATE),
    st,
    //Integer(CW_USEDEFAULT), Integer(CW_USEDEFAULT),
    //wr.Right, wr.Bottom,
    wa.Left, wa.Top,
    emuWWidth, emuWHeight,
    0, 0, hInstance, nil);
  if emuHWnd = 0 then
  begin
    WriteLn('ERROR: can''t create window.');
    Halt(1);
  end;

  dspWinHandle := emuHWnd; // dspVScrU
  EmuSoundInit(emuHWnd);
  EmuSoundDriverPause(not emuSoundEnabled);

  emuScale := scale;
  GetCursorPos(winMousePos); emuMouseButtons := 0;
  emuMousePos.x := 0; emuMousePos.y := 0;

  if emuUseDDraw then
  begin
    res := DDrawInit();
    if res <> 0 then begin WriteLn('ddraw shit: ', res); emuUseDDraw := false; end;
  end;

  EmuRenderScreenNoInv();
  ShowWindow(emuHWnd, SW_SHOWNORMAL);
  EmuDrawScreen();

  {$IFDEF USE_HP_TIMER}
  HPTimerOn(emuHQCounter);
  HPTimerOn(emuFPSHQCounter);
  emuFPSHQCounterCounting := false;
  {$ELSE}
  emuHQCounter.Stop(); emuFPSHQCounter.Stop();
  emuHQCounter.Start();
  {$ENDIF}
end;

procedure DoInterrupt ();
var
  ta: Word;
  eiCase: Boolean;
begin
  Inc(emuIntrCount);
  emuFlashState := (emuIntrCount mod 50) > 24;

  Dec(emuTapeInfo.nextEdgeT, emuTicksIntr);
  Dec(emuZ80.tStates, emuTicksIntr);

  // if it's a maskable interrupt and it is masked, get out of here %-)
  // (and don't touch "CPU halted" flag)
  if not emuZ80.iff1 then exit;

  emuZ80.halted := false;
  if Z80PeekBInternal(emuZ80, emuZ80.orgPC) = $FB then // EI?
  begin
    // prev was EI, so don't interrupt CPU just now.
    // we must execute one more instruction in this case
    if Z80PeekBInternal(emuZ80, emuZ80.regPC) = $76 then // HALT?
      Inc(emuZ80.tStates, 4) //??? anyway, no contention is necessary here
    else Z80Execute(emuZ80, emuZ80.tStates+1); // fuck up EI:EI sequence for now

    // and check if interrupts was masked by DI (to deal with EI:DI)
    if not emuZ80.iff1 then exit;
    eiCase := true;
  end
  else eiCase := false;

  // Speccy holds INTR about 31(?) t-states, so try to emulate this
  // (it's not necessary though)
  // some sources says it's 29 t-states...
  if emuZ80.tStates > 31 then
  begin
    Write('*** INTERRUPT SKIPED');
    if eiCase then Write(' due to EI');
    WriteLn;
    //WriteLn('I: #', IntToHex(emuZ80.PeekBInternalProc(emuZ80, emuZ80.prevPC), 2));
    exit;
  end;

  emuZ80.iff2 := emuZ80.iff1;
  emuZ80.iff1 := false;

  // clear EI info
  emuZ80.lastEItState := 0;

  //!!! CHECK TIMING! (it's right, but not exact...)
  // but i don't think that exact timing is necessary here,
  // 'cause there's no screen drawing or so at this time

  // push PC (& do contention %-)
  Dec(emuZ80.regSP);
  Z80PokeB(emuZ80, emuZ80.regSP, (emuZ80.regPC and $FF00) shr 8);
  Dec(emuZ80.regSP);
  Z80PokeB(emuZ80, emuZ80.regSP, emuZ80.regPC and $FF);
  case emuZ80.im of
    0, 1:
      begin
        emuZ80.regPC := 56;
        Inc(emuZ80.tStates, 13);
      end;
    2:
      begin
        ta := (emuZ80.regI*256) or $FF;
        emuZ80.regPC := Z80PeekB(emuZ80, ta)+256*Z80PeekB(emuZ80, ta+1);
        Inc(emuZ80.tStates, 19);
      end;
  end;
  // according to Spectrum FAQ, R is increased here
  Inc(emuZ80.regRTemp);
end;

function TicksToTimeStr (ticks: Integer): string;
begin
  ticks := ticks div 3500000; // seconds
  result := IntToStr(ticks mod 60);
  if ticks mod 60 < 10 then result := '0'+result;
  ticks := ticks div 60;
  result := IntToStr(ticks mod 60)+':'+result;
  if ticks mod 60 < 10 then result := '0'+result;
  ticks := ticks mod 60;
  if ticks > 0 then result := IntToStr(ticks)+':'+result;
end;

procedure EmuExec (); // not more that one frame
var
  ks: TKeyboardState;
  f: Integer;
  sleepTime, speedConst: Double;
  ts: Integer;
  y: Integer;
  s: string;
  tc: LongWord;
  fpsET: LongWord;
begin
  if emuPaused or emuTempPaused then
  begin
    EmuRenderScreenNoInv();
    if emuInfoMsgHideTime > GetTickCount() then
      DrawTextOutline(emuInfoMsgX, emuInfoMsgY, emuInfoMsg, emuInfoMsgColor, emuInfoMsgOutColor);
    {$IFDEF USE_HP_TIMER}
    HPTimerOn(emuFPSHQCounter);
    emuFPSHQCounterCounting := false;
    {$ELSE}
    emuFPSHQCounter.Stop();
    {$ENDIF}
    emuFPSStIntrCount := emuIntrCount;
    if not emuShowKeyHelp then DrawTextOutline(4, 230, 'paused', 15, 0);
    // activate debugger?
    if emuDbgActive and AltDebugActivate() then
    begin
      ConExecute('activate_debugger');
      DoDebugger();
    end;
    EmuDrawScreen();
    EmuResetTimers();
    exit;
  end;

  if emuFrameComplete then
  begin
    // frame complete, draw screen, do time sync, fps calc, etc...

    // show previous frame
    if emuFramesToSkip < 0 then
    begin
      // draw indicators

      // FPS stuff
      {$IFDEF USE_HP_TIMER}
      if not emuFPSHQCounterCounting then
      {$ELSE}
      if not emuFPSHQCounter.Counting then
      {$ENDIF}
      begin
        {$IFDEF USE_HP_TIMER}
        HPTimerOn(emuFPSHQCounter);
        emuFPSHQCounterCounting := true;
        {$ELSE}
        emuFPSHQCounter.Start();
        {$ENDIF}
        emuFPSStIntrCount := emuIntrCount;
      end
      else
      begin
        {$IFDEF USE_HP_TIMER}
        fpsET := HPTimerLap(emuFPSHQCounter);
        if fpsET >= 1*1000000 then
        {$ELSE}
        if emuFPSHQCounter.RunElapsedTime >= 1.0 then
        {$ENDIF}
        begin // emuFPS calculation
          // get deltas
          f := emuIntrCount-emuFPSStIntrCount;
          // calc emuFPS
          {$IFDEF USE_HP_TIMER}
          emuFPS := f/(fpsET/1000000);
          {$ELSE}
          emuFPS := f/emuFPSHQCounter.RunElapsedTime;
          {$ENDIF}
          // remember new values
          //!!!if emuFPSHQCounter.RunElapsedTime >= 1.0 then
          begin
            emuFPSStIntrCount := emuIntrCount;
            {$IFDEF USE_HP_TIMER}
            HPTimerOn(emuFPSHQCounter);
            {$ELSE}
            emuFPSHQCounter.Stop(); emuFPSHQCounter.Start();
            {$ENDIF}
          end;
        end;
      end;

      // draw FPS
      //Str(emuFPS:7:2, s); DrawTextOutline(4, 230, s, 15, 0);
      tc := GetTickCount();
      if not emuMaxSpeed and (emuFPS >= 25) then
      begin
        f := trunc(emuFPS+0.5);
        Dec(f, trunc(50*emuSpeedFactor));
        if (f < -2) then
        begin
          Inc(emuPrevFPSSlows);
          if emuPrevFPSSlows > 3 then
          begin
            emuSlowIconHide := tc+500;
            emuPrevFPSSlows := 0;
          end;
        end
        else
        begin
          emuSlowIconHide := 0;
          Inc(emuPrevFPSSlows); if emuPrevFPSSlows > 0 then emuPrevFPSSlows := 0;
        end
      end
      else emuPrevFPSSlows := 0;

      if emuShowFPS or emuMaxSpeed or ((emuSlowIconHide > tc) and not emuNoTurtle) then
      begin
        f := trunc(emuFPS+0.5); Str(f:2, s); if f < 10 then s[1] := '0';
        DrawTextOutline(4, 230, 'FPS: '+s, 15, 0);
      end;

      if emuMaxSpeed then DrawTextOutline(1, 1, 'max', 15, 0)
      else if emuSpeedFactor = 2 then DrawTextOutline(1, 1, '2x', 15, 0)
      else if emuSpeedFactor <> 1 then DrawTextOutline(1, 1, 'custom speed', 15, 0);

      if emuTapeInfo.playing and (emuTapeInfo.tape.Count > 0) then
      begin
        DrawTextOutline(64, 230, 'tape: '+
          IntToStr(emuTapeInfo.tape.Position*100 div emuTapeInfo.tape.Count)+'%', 7, 0);
          //TicksToTimeStr(emuTapeInfo.tape.TotalTicks-emuTapeInfo.ticksPassed), 7, 0);
        //EmuDrawTapeCounter(0, 217);
      end;
      if emuTapeInfo.tape.Count > 0 then EmuDrawTapeCounter(0, 217);

      if emuWasDiskActivity then
      begin
        emuDiskIconHide := tc+100;
        //DrawTextOutline(320-28, 4, 'disk', 7, 0);
      end;

      if emuInfoMsgHideTime > GetTickCount() then
        DrawTextOutline(emuInfoMsgX, emuInfoMsgY, emuInfoMsg, emuInfoMsgColor, emuInfoMsgOutColor);

      EmuDrawScreen();
    end;

    // reset disk activity flag, so no message for the next frame
    // (if there's no disk activity in it %-)
    emuWasDiskActivity := false;

    if not emuMaxSpeed and (emuFramesToSkip < 0)
       {$IFNDEF USE_HP_TIMER}and emuHQCounter.Counting{$ENDIF} then
    begin
      // just sync with <emuHQCounter>

      // calculate desired ms per frame
      speedConst := 20.0/emuSpeedFactor;
      if speedConst < 5 then speedConst := 5;
      if speedConst > 500 then speedConst := 500;

      // calc time to waste
      {$IFDEF USE_HP_TIMER}
      sleepTime := speedConst-(HPTimerLap(emuHQCounter)/1000+waitTimeLeft);
      // restart frame timer
      HPTimerOn(emuHQCounter);
      {$ELSE}
      sleepTime := speedConst-(emuHQCounter.RunElapsedTime*1000+waitTimeLeft);
      // restart frame timer
      emuHQCounter.Stop(); emuHQCounter.Start();
      {$ENDIF}
      //Write(trunc(sleepTime):4);
      //Write('!', trunc(waitTimeLeft):3);
      //MessageBox(0, PChar(FloatToStr(sleepTime)), 'RUN', 0);

      if sleepTime < 0 then
      begin
        // the previous frame takes more time than we want it to take
        // just make the next frame delay shorter
        waitTimeLeft := waitTimeLeft-sleepTime; // -(-n)=+n
        // 1--3 ms slowdown doesn't matter
        if waitTimeLeft > 3 then
        begin
          // there is major slowdown, so
          // show turtle...
          // ...and skip some next frames (if we aren't
          // skiping 'em already)
          if (emuFramesToSkip < 0) and (emuFrameSkip < 0) then
          begin
            emuFramesToSkip := trunc(waitTimeLeft/speedConst+0.5);
          end;
          waitTimeLeft := 0;
        end;
        //if waitTimeLeft > 3000 then waitTimeLeft := 0;
      end
      else
      begin
        // there's some time to waste %-)
        if sleepTime > 0.7 then
        begin
          {$IFDEF USE_HP_TIMER}
          while HPTimerLap(emuHQCounter)/1000 < sleepTime do ;
          // restart frame timer
          HPTimerOn(emuHQCounter);
          {$ELSE}
          emuHQCounter.Wait(sleepTime/1000);
          // restart frame timer
          emuHQCounter.Stop(); emuHQCounter.Start();
          {$ENDIF}
        end;

        waitTimeLeft := 0; // no slowdown at the next frame
      end;
    end
    else
    begin
      // restart frame timer
      {$IFDEF USE_HP_TIMER}
      HPTimerOn(emuHQCounter);
      {$ELSE}
      emuHQCounter.Stop(); emuHQCounter.Start();
      {$ENDIF}
    end;

    // reset "frame complete flag"
    emuFrameComplete := false;
    ZXExUpdateScrInit();
  end;

  // if we can't skip frames (due to config settings)
  // don't skip 'em
  if emuFrameSkip = 0 then emuFramesToSkip := -1;

  // DI:HALT reanimation -- enable interrupts
  if emuReanimated then begin emuZ80.iff1 := true; emuZ80.iff2 := true; end;

  if emuFramesToSkip >= 0 then
  begin
    // execute at the maximum speed (w/o accurate screen rendering & sound)
    Z80Execute(emuZ80, emuTicksIntr);
    //y := 0; // render the whole screen (see the following code)
  end
  else
  begin
    // now we have to determine where we are (in other words,
    // what screen line we have to draw)

    // work out which line we're on, relative to the top of the screen
    if emuZ80.tStates < emuLineTStates[0] then y := 0
    else y := ((emuZ80.tStates-emuLineTStates[0]) div emuTicksLine)-1;
    repeat
      // execute Z80 instructions; duration is about Speccy screen line
      // drawing time
      // calc the time (t-states) of the next breakpoint
      repeat
        Inc(y);
        if y < High(emuLineTStates) then ts := emuLineTStates[y+1]
        else ts := emuTicksIntr;
        if ts > emuTicksIntr then ts := emuTicksIntr;
      until (ts > emuZ80.tStates) or (ts >= emuTicksIntr);
      // now <y> is the current screen line to update, <ts> is the
      // next breakpoint time (right after the line has to be rendered)

      // execute Z80 commands till the next breakpoint
      Z80Execute(emuZ80, ts);
      // note that we can exit ealier than <ts>!

      // render sound
      if not emuMaxSpeed then EmuSoundFillBuffer(emuZ80.tStates);

      // activate debugger?
      if emuDbgActive and AltDebugActivate() then ConExecute('activate_debugger');

      //!if emuTapeInfo.playing then EmuTapeBit(); // to do "autostop"
    until (emuZ80.tStates >= emuTicksIntr) or emuZ80.halted or emuTempPaused or emuPaused;
  end;

  {if emuTapeInfo.playing then }EmuTapeBit(); // to do "autostop"

  if emuTempPaused or emuPaused then exit; //!

  // check if we interrupted before the frame rendering ends
  if not emuZ80.halted and (emuZ80.tStates < emuTicksIntr) then
  begin
    // the current frame isn't complete, do something %-)
    // just exit for now
    exit;
  end;

  // the current frame is complete, set the flag and do the work
  emuFrameComplete := true;

  // here emuZ80.tStates can be less then emuTicksIntr due to HALTing
  // so we have to handle this situation
  if emuZ80.halted and (emuZ80.tStates < emuTicksIntr) then
  begin
    // in HALT mode CPU executes NOP's, so minimal t-state
    // step is 4 ticks
    f := emuTicksIntr-emuZ80.tStates; // t-states left
    Inc(emuZ80.tStates, (f div 4)*4);
    // "while", 'cause i'm too lazy to think %-)
    while emuZ80.tStates < emuTicksIntr do Inc(emuZ80.tStates, 4);

    // check DI:HALT
    if emuZ80.halted and not emuZ80.iff1 then
    begin
      // DI:HALT found, try to reanimate at the next step
      if not emuReanimated then WriteLn('***DI/HALT executed. reanimating...')
      else WriteLn('***DI/HALT reanimation failed %-(');
      emuReanimated := true;
      //WriteLn('I: #', IntToHex(Z80PeekBInternal(emuZ80, emuZ80.prevPC), 2));
    end
    else emuReanimated := false;
  end;

  // if the tape is playing, advance it
  {if emuTapeInfo.playing then }EmuTapeBit();

  if emuSoundEnabled then
  begin
    // finish sound rendering. some sound samples can be lost due to
    // the fact that the one frame can take more that emuTicksIntr
    // t-states.
    EmuSoundFinishFillBuffer();
    // emuAY.Update(); -- EmuSoundFinishFillBuffer does this
  end;

  if emuFramesToSkip < 0 then
  begin
    // this frame wasn't skipped, so render the part of the screen
    // that still wasn't rendered (if any)
    ZXExUpdateScr(emuTicksIntr, @emuMemory[emuScreenPage]);
    ZXWindowsDraw(); //!!
  end;

  // jump to ISR if interrupts are enabled, update "flash",
  // set new value of emuZ80.tStates & next tape edge timing
  DoInterrupt();

  // deal with autofire
  if config.autofiring then
  begin
    Dec(emuAutofireDelay);
    if emuAutofireDelay <= -config.autofireKeepPressed then emuAutofireDelay := config.autofireRate;
  end;

  // do not check keys & mouse on the "skipped" frame
  if emuFramesToSkip < 0 then
  begin
    // convert PC keys to Speccy ones
    if keybdOn then
    begin
      GetKeyboardState(ks); for f := 1 to 255 do pcKeys[f] := (ks[f] and $80) <> 0;
      PCKeys2ZX();
    end;

    // process joysticks
    PCJoyScan();
    PCJoy2ZX();

    // process kempston mouse (shitty code! %-)
    DoPCMouse();
  end;

  // draw icons (if any)
  if emuSlowIconHide > GetTickCount() then
  begin
    // show turtle
    if not emuNoTurtle then
      VDrawIcon(dspIconTooSlow, 320-dspIconDisk[0]-10-dspIconTurtle[1], 240-dspIconTurtle[1]);
  end;
  if emuDiskIconHide > GetTickCount() then
  begin
    // show disk
    VDrawIcon(dspIconDiskIO, 320-dspIconDisk[0], 240-dspIconDisk[1]);
  end;

  // should we skip some frames?
  if emuFramesToSkip >= 0 then Dec(emuFramesToSkip)
  else
  begin
    // no frames to skip, so check another things
    if emuMaxSpeed then
    begin
      if emuDbgActive then emuFramesToSkip := 32 else emuFramesToSkip := 64;
    end
    else
    begin
      if emuFrameSkip > 0 then emuFramesToSkip := emuFrameSkip;
    end;

    EmuDrawWCon();
  end;
end;

function ZXGetKeyNameByXYEx (row, col: Integer; out down: Boolean): string;
var
  f: Integer;
  b: Byte;
begin
  down := false;
  if (row < 0) or (row > 3) or (col < 0) or (col > 9) then begin result := '???'; exit; end;
  if col > 4 then begin col := 9-col; Inc(row, 4); end
  else row := 3-row;
  // now: row: emuKeybMask number; col: emuKeybMask mask bit shift
  b := not (1 shl col);
  // search for the key name
  for f := 0 to High(zxKNames) do
  begin
    if (zxKNames[f].pvptr = @emuKeybMask[row]) and (zxKNames[f].mask = b) then
    begin
      result := zxKNames[f].name;
      down := (zxKNames[f].pvptr^ and (not zxKNames[f].mask)) = 0;
      exit;
    end;
  end;

  result := '!!!'; // this should never happens! %-)
end;

function ZXGetKeyNameByXY (row, col: Integer): string;
var
  b: Boolean;
begin
  result := ZXGetKeyNameByXYEx(row, col, b);
end;

procedure CmdKeyboard (const cmd: string; const args: array of string);
var
  f, c: Integer;
  s: string;
  down: Boolean;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn(#0#1#1'enable/disable PC to ZX keyboard translation.');
    WriteLn('usage: keyboard [tan|ona|state]');
    WriteLn('  "state" shows the current zx keyboard state');
    exit;
  end;

  if ConGetArgStr(1, args) = 'state' then
  begin
    for f := 0 to 3 do
    begin
      Write(' ');
      for c := 0 to 9 do
      begin
        s := ZXGetKeyNameByXYEx(f, c, down);
        if Length(s) < 3 then s := ' '+s+' ';
        if down then s := '['+UpperCase(s)+']' else s := ' '+s+' ';
        Write(s);
      end;
      WriteLn;
    end;
    WriteLn('keyboard: ', ConYesNoName(keybdOn));
    exit;
  end;

  keybdOn := ConGetArgBoolDef(1, args, not keybdOn);
  WriteLn('keyboard: ', ConYesNoName(keybdOn));
end;

procedure CmdZXKeyDown (const cmd: string; const args: array of string);
var
  f, c, cnt: Integer;
  down: Boolean;
  s: string;
  b: Byte;
begin
  down := (cmd = 'zxkeydown');
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    if down then Write('press') else Write('release');
    WriteLn('Speccy key.');
    WriteLn('usage: ', cmd, ' key0 [key1 ...] | all');
    exit;
  end;

  cnt := ConGetArgCount(args);
  if cnt < 1 then
  begin
    WriteLn('wtf?');
    exit;
  end;

  for f := 1 to cnt do
  begin
    s := ConGetArgStr(f, args);
    if s = 'all' then
    begin
      if down then b := Byte(not $1F) else b := $FF;
      emuKeybMask[0] := b; emuKeybMask[1] := b; emuKeybMask[2] := b;
      emuKeybMask[3] := b; emuKeybMask[4] := b; emuKeybMask[5] := b;
      emuKeybMask[6] := b; emuKeybMask[7] := b;
      Write('all keys ');
      if down then WriteLn('pressed') else WriteLn('released');
      exit;
    end;

    c := High(zxKNames); while (c >= 0) and (zxKNames[c].name <> s) do Dec(c);
    if c < 0 then begin WriteLn(s, '?'); continue; end;
    if down then zxKNames[c].pvptr^ := zxKNames[c].pvptr^ and zxKNames[c].mask
    else zxKNames[c].pvptr^ := zxKNames[c].pvptr^ or (not zxKNames[c].mask);
  end;
end;


initialization
  InitWndClass();

  emuFullScreen := false;
  emuHWnd := 0;

  emuPaused := false;
  emuTempPaused := false;
  emuReanimated := false;

  RegisterConsoleCommand('keyboard', @CmdKeyboard);
  RegisterConsoleCommand('zxkeydown', @CmdZXKeyDown);
  RegisterConsoleCommand('zxkeyup', @CmdZXKeyDown);
finalization
  if emuFullScreen then ChangeDisplaySettings(devmode(nil^), 0);
  EmuUnlockMouse(); EmuShowMouse();
end.
