// command console with built-in ALISP3 engine
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
{$DEFINE USE_ALISP}
unit lispCon3U;

interface

uses
  Windows, SysUtils, Classes{$IFDEF USE_ALISP}, alisp3U{$ENDIF};


type
  TConsoleCommandHandler = procedure (const cmd: string; const args: array of string);


const
  conCommentChar = '#';
  conCmdDelimiterChar = ';';
  {$IFDEF USE_ALISP}
  conLispCommandChar = '.';
  conMuteLispCommandChar = ':';
  {$ENDIF}
  conVersion = '0.4.0';
  conFullVersion = 'lispConsole v'+conVersion+' by Ketmar <ketmar@ua.fm>';


{$IFDEF USE_ALISP}
procedure ConAddStdLispCommands ();
{$ENDIF}

procedure RegisterConsoleCommand (const name: string; handler: TConsoleCommandHandler);
{$IFDEF USE_ALISP}
procedure RegisterConsoleLispCommand (const name, lispSymName: string);
{$ENDIF}
procedure UnregisterConsoleCommand (const name: string);

function ConYesNoValue (s: string): Boolean;
function ConYesNoName (const value: Boolean): string;

// first arg: 1
function ConGetArgStr (n: Integer; const args: array of string): string;
function ConGetArgIntDef (n: Integer; const args: array of string; default: Integer): Integer;
function ConGetArgIntEx (n: Integer; const args: array of string; out res: Integer): Boolean;
function ConGetArgDisk (n: Integer; const args: array of string): Integer;
function ConGetArgBoolEx (n: Integer; const args: array of string; out res: Boolean): Boolean;
function ConGetArgBoolDef (n: Integer; const args: array of string; default: Boolean): Boolean;
function ConGetArgCount (const args: array of string): Integer;
function ConIsArgHelp (const args: array of string): Boolean;

function ConStr2Int (const s: string; out res: Integer): Boolean;
  // %bin; $hex; #hex; 0xHEX; @octal


function ConFindPartialCommand (name: string; var nextPartial: Integer): string;
  // start with nextPartial = -1
  // returns empty string if there's no matching commands

procedure ConExecute (cmd: string);
  // execute console line
procedure ConExecuteCommand (const cmd: string);
  // execute one console command. do not processes aliases. intercepts exceptions.

{$IFDEF USE_ALISP}
function ConALIsArgHelp (nn: Integer): Boolean;
{$ENDIF}

function ConIsSpecialFile (const fName: string): Boolean;
function ConGetClipboardString (wndH: HWND): string;
procedure ConSetClipboardString (const s: string; wndH: HWND);


implementation


var
  conAliasList: TStringList;
  conCommandList: TStringList;
    //Objects[] < 0: index in conCommandLisp (abs(Objects[x])-1)
  {$IFDEF USE_ALISP}
  conCommandLisp: TStringList; // global symbols!
  {$ENDIF}


function ConIsSpecialFile (const fName: string): Boolean;
var
  hd: Integer;
begin
  result := true;
  if fName = '' then exit;
  if not FileExists(fName) then
  begin
    // no such file, try to create one
    hd := Windows.CreateFile(PChar(fName), 0, FILE_SHARE_READ or FILE_SHARE_WRITE, nil,
      OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
    if hd = -1 then exit; // not a device, not a file
    CloseHandle(hd);
    if not FileExists(fName) then exit; // a device
    DeleteFile(fName);
  end;
  result := false;
end;

function ConGetClipboardString (wndH: HWND): string;
var
  data: THandle;
  pc: PChar;
begin
  result := '';
  if not OpenClipboard(wndH) then exit;
  try
    data := GetClipboardData(CF_TEXT);
    if data <> 0 then
    begin
      pc := GlobalLock(data);
      if pc <> nil then
      try
        SetString(result, pc, StrLen(pc));
      finally
        GlobalUnlock(data);
      end;
    end;
  finally
    CloseClipboard();
  end;
end;

procedure ConSetClipboardString (const s: string; wndH: HWND);
var
  data: THandle;
  pc: PChar;
begin
  if s = '' then exit;
  if not OpenClipboard(wndH) then exit;
  try
    EmptyClipboard();
    data := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, Length(s)+1);
    if data = 0 then exit;
    try
      pc := GlobalLock(data);
      if pc = nil then exit;
      try
        FillChar(pc^, Length(s)+1, 0);
        Move(s[1], pc^, Length(s));
        SetClipboardData(CF_TEXT, data);
        //SetClipboardData(CF_OEMTEXT, data);  //???
      finally
        GlobalUnlock(data);
      end;
    finally
      GlobalFree(data);
    end;
  finally
    CloseClipboard();
  end;
end;


function FindConCmd (const name: string): Integer;
// search for a command. return pos or -1
begin
  if not conCommandList.Find(name, result) then result := -1;
end;

function FindAlias (const name: string): Integer;
// search for an alias. return pos or -1
begin
  result := 0;
  while (result < conAliasList.Count) and (conAliasList.Names[result] <> name) do Inc(result);
  if result >= conAliasList.Count then result := -1;
end;


function ConFindPartialCommand (name: string; var nextPartial: Integer): string;
var
  tmp: Integer;
begin
  result := ''; if conCommandList.Count = 0 then exit;

  if (nextPartial < 0) or (nextPartial >= conCommandList.Count) then nextPartial := 0;

  tmp := nextPartial;
  while nextPartial < conCommandList.Count do
  begin
    Inc(nextPartial);
    result := conCommandList[nextPartial-1];
    if Copy(result, 1, Length(name)) = name then exit;
  end;

  nextPartial := 0;
  while nextPartial < tmp do
  begin
    Inc(nextPartial);
    result := conCommandList[nextPartial-1];
    if Copy(result, 1, Length(name)) = name then exit;
  end;

  nextPartial := -1; result := '';
end;


function ConYesNoValue (s: string): Boolean;
begin
  if s = '' then begin result := false; exit; end;
  s := LowerCase(s);
  if (s[1] = 'y') or (s[1] = 't') then begin result := true; exit; end;
  result := (s = 'on') or (s = 'si') or
    (s = '1') or (s = '-1') or
    (s = 'aie') or (s = 'aye');
end;

function ConYesNoName (const value: Boolean): string;
begin
  if value then result := 'tan' else result := 'ona';
end;


function ConStr2Int (const s: string; out res: Integer): Boolean;
// %bin; $hex; #hex; 0xHEX; @octal
var
  base, c, dig: Integer;
  ch: Char;
  neg: Boolean;
begin
  result := false; res := 0;

  c := 1; while (c <= Length(s)) and (s[c] <= ' ') do Inc(c);
  if c > Length(s) then exit;
  neg := false;
  if s[c] = '+' then Inc(c)
  else if s[c] = '-' then begin neg := true; Inc(c); end;
  if c > Length(s) then exit;

  if (c+2 <= Length(s)) and (s[c] = '0') and (s[c+1] = 'x') then
  begin
    base := 16;
    Inc(c, 2);
  end
  else
  begin
    case s[c] of
      '#', '$': begin base := 16; Inc(c); end;
      '%': begin base := 2; Inc(c); end;
      '@': begin base := 8; Inc(c); end;
      else base := 10;
    end;
  end;
  if c > Length(s) then exit;

  while (c <= Length(s)) and (s[c] > ' ') do
  begin
    ch := UpCase(s[c]); Inc(c);
    if (ch < '0') or (ch > 'F') or ((ch > '9') and (ch < 'A')) then exit;
    dig := ord(ch)-ord('0'); if dig > 9 then Dec(dig, 7);
    if dig >= base then exit;
    res := res*base+dig;
  end;
  if neg then res := -res;

  while (c <= Length(s)) and (s[c] <= ' ') do Inc(c);

  result := (c > Length(s));
end;

function ConIsArgHelp (const args: array of string): Boolean;
begin
  result := (Length(args) = 1) and (
    (args[0] = '?') or (args[0] = '/?') or (args[0] = '-?'));
end;

function ConGetArgStr (n: Integer; const args: array of string): string;
begin
  if (n < 1) or (n > Length(args)) then result := '' else result := args[n-1];
end;

function ConGetArgIntEx (n: Integer; const args: array of string; out res: Integer): Boolean;
begin
  res := 0; result := false;
  if (n >= 1) and (n <= Length(args)) then
  begin
    result := ConStr2Int(args[n-1], res);
  end;
end;

function ConGetArgIntDef (n: Integer; const args: array of string; default: Integer): Integer;
begin
  if not ConGetArgIntEx(n, args, result) then result := default;
end;

function ConGetArgDisk (n: Integer; const args: array of string): Integer;
begin
  result := -1; Dec(n);
  if (n < 0) or (n > High(args)) or (Length(args[n]) <> 2) then exit;
  if args[n][2] <> ':' then exit;
  result := ord(UpCase(args[n][1]))-ord('A');
  if (result < 0) or (result > 3) then result := -1;
end;

function ConGetArgBoolEx (n: Integer; const args: array of string; out res: Boolean): Boolean;
begin
  res := false; result := false;
  if (n >= 1) and (n <= Length(args)) then
  begin
    res := ConYesNoValue(LowerCase(args[n-1])); result := true;
  end;
end;

function ConGetArgBoolDef (n: Integer; const args: array of string; default: Boolean): Boolean;
begin
  if not ConGetArgBoolEx(n, args, result) then result := default;
end;

function ConGetArgCount (const args: array of string): Integer;
begin
  result := Length(args);
end;


procedure RegisterConsoleCommand (const name: string; handler: TConsoleCommandHandler);
var
  f: Integer;
  s: string;
begin
  s := name;
  f := 1; while (f <= Length(s)) and (s[f] <= ' ') do Inc(f);
  Delete(s, 1, f-1);
  f := Length(s); while (f > 0) and (s[f] <= ' ') do Dec(f);
  s := Copy(s, 1, f);
  if s = '' then
  begin
    WriteLn('empty console command was not added');
    exit;
  end;

  f := FindConCmd(s);
  if f <> -1 then
  begin
    UnregisterConsoleCommand(s);
    WriteLn('console command ("', s, '") was re-defined');
  end;

  conCommandList.AddObject(name, Pointer(@handler));
end;

{$IFDEF USE_ALISP}
procedure RegisterConsoleLispCommand (const name, lispSymName: string);
var
  f: Integer;
  s: string;
begin
  s := name;
  f := 1; while (f <= Length(s)) and (s[f] <= ' ') do Inc(f);
  Delete(s, 1, f-1);
  f := Length(s); while (f > 0) and (s[f] <= ' ') do Dec(f);
  s := Copy(s, 1, f);
  if s = '' then
  begin
    WriteLn('empty console command was not added');
    exit;
  end;

  f := FindConCmd(s);
  if f <> -1 then
  begin
    UnregisterConsoleCommand(s);
    WriteLn('console command ("', s, '") was re-defined');
  end;

  conCommandList.AddObject(name, Pointer(-(conCommandLisp.Add(lispSymName)+1)));
end;
{$ENDIF}

procedure UnregisterConsoleCommand (const name: string);
var
  f{$IFDEF USE_ALISP}, c, n{$ENDIF}: Integer;
  s: string;
begin
  s := name;
  f := 1; while (f <= Length(s)) and (s[f] <= ' ') do Inc(f);
  Delete(s, 1, f-1);
  f := Length(s); while (f > 0) and (s[f] <= ' ') do Dec(f);
  s := Copy(s, 1, f);

  f := FindConCmd(s);
  if f <> -1 then
  begin
    conCommandList.Delete(f);
    {$IFDEF USE_ALISP}
    c := Integer(conCommandList.Objects[f]);
    if c < 0 then
    begin
      c := abs(c)-1;
      if c < conCommandLisp.Count then
      begin
        conCommandLisp.Delete(c);
        f := conCommandList.Count;
        while f > 0 do
        begin
          Dec(f);
          n := Integer(conCommandList.Objects[f]);
          if n < 0 then
          begin
            n := abs(n)-1;
            ASSERT(n <> c);
            if (n > c) and (n <= conCommandLisp.Count) then
              conCommandList.Objects[f] := Pointer(-(n));
          end;
        end;
      end;
    end;
    {$ENDIF}
  end;
end;


var
  tokenStrFlag: Boolean;

function Token (const cmd: string; var stPos: Integer; out endPos: Integer): Boolean;
// false: no token found
var
  qch: Char;
  pos: Integer;
begin
  result := false; tokenStrFlag := false;

  pos := stPos;
  // skip blanks
  while (pos <= Length (cmd)) and (cmd[pos] <= ' ') do Inc(pos);

  stPos := pos; endPos := pos;
  if stPos > Length (cmd) then exit; // no token

  result := true; // token found

  qch := cmd[pos];
  if qch = conCommentChar then
  begin
    Inc(endPos);
    // comment -- end of line, no token found
    exit;
  end;

  if qch = conCmdDelimiterChar then
  begin
    // end marker
    Inc(endPos); exit;
  end;

  if qch in ['"', #39] then Inc(pos) else qch := ' ';

  while pos <= Length(cmd) do
  begin
    if qch = ' ' then
    begin
      if cmd[pos] in [#0..' ', conCmdDelimiterChar, conCommentChar] then break;
    end
    else if cmd[pos] = qch then break;
    Inc(pos);
  end;
  if (qch <> ' ') and (pos <= Length(cmd)) then Inc(pos);
  endPos := pos;
  tokenStrFlag := (qch <> ' ');
end;

procedure ConExecuteCommand (const cmd: string);
// do one command
var
  args: array of string;
  f, cmdI, ssP, stPos, endPos: Integer;
  {$IFDEF USE_ALISP}
  c, {cnt, }lcell: Integer;
  fv: TALFloat;
  lispMute: Boolean;
  s: string;
  {$ENDIF}
  cs: string;
begin
  args := nil;
  try
    stPos := 1;
    if not Token(cmd, stPos, endPos) then exit;
    cs := Copy(cmd, stPos, endPos-stPos);
    if (cs = '') or (cs = conCmdDelimiterChar) or (cs = conCommentChar) then exit;
    {$IFDEF USE_ALISP}
    if (cs[1] = conLispCommandChar) or (cs[1] = conMuteLispCommandChar) then
    begin
      // the rest of the line is a lisp source
      lispMute := (cs[1] = conMuteLispCommandChar);
      cs := Copy(cmd, stPos+1, Length(cmd));
      f := ALEvalStr(cs);
      if ALGetBreakFlag() = ALBREAK_THROW then begin Write('THROW: '); ALPrint(f); WriteLn; end
      else if (ALGetBreakFlag() <> ALBREAK_ERROR) and not lispMute then begin ALPrint(f); WriteLn; end;
      ALSetBreakFlag(ALBREAK_NONE);
      exit;
    end;
    {$ENDIF}
    ssP := endPos;

    cmdI := FindConCmd(cs);
    if cmdI = -1 then
    begin
      WriteLn(cs, '?');
      exit;
    end;
    if conCommandList.Objects[cmdI] = nil then exit;

    {$IFDEF USE_ALISP}
    f := Integer(conCommandList.Objects[cmdI]);
    if f < 0 then
    begin
      f := abs(f)-1;
      if f < conCommandLisp.Count then
      begin
        lcell := ALFindGlobalSymbol(conCommandLisp[f]);
        if ALIsNil(lcell) then lcell := ALFindGlobalSymbol('console-cmd-handler');
        if ALIsNil(lcell) then
        begin
          WriteLn(cs, '?!');
          exit;
        end;
        // lisp command. build arglist
        f := ALNil(); stPos := ssP; //cnt := 0;
        while Token(cmd, stPos, endPos) do
        begin
          if (stPos+1 = endPos) and
             ((cmd[stPos] = conCmdDelimiterChar) or (cmd[stPos] = conCommentChar)) then break;
          //Inc(cnt);
          s := Copy(cmd, stPos, endPos-stPos); stPos := endPos;
          if tokenStrFlag then c := ALNewStrAtom(Copy(s, 2, Length(s)-2))
          else if ConStr2Int(s, c) then c := ALNewIntAtom(c)
          else
          begin
            Val(s, fv, c);
            if c = 0 then c := ALNewFloatAtom(fv) else c := ALNewStrAtom(s);
          end;
          if ALIsNil(f) then
          begin
            f := ALCons(c, ALNil());
            ALPushCell(f);
          end
          else
          begin
            ALSetCdr(f, ALCons(c, ALNil()));
            f := ALCdr(f);
          end;
        end;
        if not ALIsNil(f) then
        begin
          f := ALPopCells(1);
          f := ALCons(ALXQuote(), f);
          f := ALCons(f, ALNil());
        end;
        // f: arglist
        //ALPrintNoQ(f); WriteLn;
        ALPushCell(f);
        //f := ALCons(ALCar(lcell), ALCons(ALCons(ALQuote(), ALNewStrAtom(cs)), ALPopCells(1)));
        f := ALCons(ALCar(lcell),
          ALCons(
            ALCons(ALQuote(), ALCons(ALNewStrAtom(cs), ALNil())), ALPopCells(1))
          );
        //ALPrintNoQ(f); WriteLn;
        ALSetBreakFlag(ALBREAK_NONE);
        f := ALEval(f);
        if ALGetBreakFlag() = ALBREAK_THROW then begin Write('THROW: '); ALPrint(f); WriteLn; end;
        ALSetBreakFlag(ALBREAK_NONE);
        exit;
      end
      else
      begin
        WriteLn(cs, '?!!');
        exit;
      end;
    end;
    {$ENDIF}

    f := 0; stPos := ssP;
    while Token(cmd, stPos, endPos) do
    begin
      if (stPos+1 = endPos) and
         ((cmd[stPos] = conCmdDelimiterChar) or (cmd[stPos] = conCommentChar)) then break;
      Inc(f); stPos := endPos;
    end;

    SetLength(args, f);
    f := 0; stPos := ssP;
    while Token(cmd, stPos, endPos) do
    begin
      if (stPos+1 = endPos) and
         ((cmd[stPos] = conCmdDelimiterChar) or (cmd[stPos] = conCommentChar)) then break;
      if f >= Length(args) then break; // just in case %-)
      if tokenStrFlag then args[f] := Copy(cmd, stPos+1, endPos-stPos-2)
      else args[f] := Copy(cmd, stPos, endPos-stPos);
      Inc(f); stPos := endPos;
    end;

    TConsoleCommandHandler(conCommandList.Objects[cmdI])(cs, args);
  except on e: Exception do
    begin
      WriteLn('***CONSOLE: EXCEPTION [', e.ClassName, '] -- "', e.Message, '"');
    end;
  end;
end;


procedure ConExecute (cmd: string);
var
  stPos, endPos, f: Integer;
  {$IFDEF USE_ALISP}
  blv: Integer;
  ch: Char;
  {$ENDIF}
  s, cs: string;
  stopAliasing: Boolean;
begin
  stopAliasing := false;
  repeat
    // get the command text length
    // skip blanks and empty commands
    stPos := 1;
    while (stPos <= Length(cmd)) and (cmd[stPos] in [#0..' ', conCmdDelimiterChar]) do Inc(stPos);
    Delete(cmd, 1, stPos-1);

    if (cmd = '') or (cmd[1] = conCommentChar) then exit;

    {$IFDEF USE_ALISP}
    if (cmd[1] = conLispCommandChar) or (cmd[1] = conMuteLispCommandChar) then
    begin
      // lisp. check for command delimiter
      f := 2; blv := 0;
      while f <= Length(cmd) do
      begin
        if cmd[f] = conCmdDelimiterChar then break
        else if (cmd[f] = conCommentChar) and (blv < 1) then break
        else if cmd[f] = '(' then Inc(blv)
        else if cmd[f] = ')' then begin if blv > 0 then Dec(blv); end
        else if (cmd[f] = '"') or (cmd[f] = '`') then
        begin
          ch := cmd[f]; Inc(f);
          while f <= Length(cmd) do
          begin
            if cmd[f] = ch then break
            else if cmd[f] = '\' then Inc(f);
            Inc(f);
          end;
        end
        else if cmd[f] = '!' then Inc(f)
        else if (f < Length(cmd)) and (cmd[f] = '/') and (cmd[f+1] = '*') then
        begin
          Inc(f, 2);
          while f < Length(cmd) do
          begin
            if (cmd[f] = '*') and (cmd[f+1] = '/') then begin Inc(f); break; end;
            Inc(f);
          end;
        end;
        Inc(f);
      end;
      s := Copy(cmd, 1, f-1);
      if (f <= Length(cmd)) and (cmd[f] = conCommentChar) then cmd := ''
      else cmd := Copy(cmd, f+1, Length(cmd));
      ConExecuteCommand(s);
      continue;
    end;
    {$ENDIF}

    stopAliasing := stopAliasing or (Length(cmd) > 4096);
    stPos := 1; s := cmd;
    while Token(cmd, stPos, endPos) do
    begin
      if (stPos+1 = endPos) and (cmd[stPos] in [conCmdDelimiterChar, conCommentChar]) then
      begin
        //Dec(endPos);
        break;
      end;
      stPos := endPos;
    end;

    s := TrimRight(Copy(cmd, 1, stPos-1)); Delete(cmd, 1, stPos-1);
    if s = '' then continue;

    // do not make too long commands -- this
    // prevents "infinite aliasing cycle"
    if not stopAliasing then
    begin
      // get first token
      f := Pos(' ', s); if f = 0 then f := Length(s)+1;
      cs := Copy(s, 1, f-1); Delete(s, 1, f);
      // check for alias
      f := FindAlias(cs);
      if f <> -1 then
      begin
        // alias found, substitute and parse the command again
        cmd := conAliasList.ValueFromIndex[f]+' '+s+' '+cmd;
        continue;
      end;
      s := cs+' '+s;
    end;

    ConExecuteCommand(s);
  until false;
end;

procedure CmdConVersion (const cmd: string; const args: array of string);
begin
  WriteLn(conFullVersion);
end;

procedure CmdCmdList (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('show list of all available commands');
    WriteLn('usage: cmd-list');
    exit;
  end;

  for f := 0 to conCommandList.Count-1 do WriteLn(conCommandList[f]);
end;

procedure CmdAliasList (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('show list of all available aliases');
    WriteLn('usage: alias-list');
    exit;
  end;

  for f := 0 to conAliasList.Count-1 do WriteLn(conAliasList[f]);
end;

procedure CmdEcho (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if Length(args) = 0 then
  begin
    WriteLn;
    WriteLn('print string');
    WriteLn('usage: echo args');
    exit;
  end;

  for f := 0 to High(args) do Write(args[f], ' ');
  WriteLn;
end;

procedure CmdClearAliases (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('reset all aliases.');
    WriteLn('usage: clear-aliases');
    exit;
  end;

  conAliasList.Clear();
end;

procedure CmdAlias (const cmd: string; const args: array of string);
var
  name, text: string;
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set/reset alias.');
    WriteLn('usage: alias name [text]');
    exit;
  end;

  name := ConGetArgStr(1, args); text := ConGetArgStr(2, args);
  if name = '' then
  begin
    WriteLn('alias what?');
    exit;
  end;

  f := FindAlias(name);
  if f <> -1 then
  begin
    if text = '' then conAliasList.Delete(f) else conAliasList[f] := name+'='+text;
  end
  else
  begin
    if text = '' then exit;
    f := FindConCmd(name);
    if f <> -1 then
    begin
      WriteLn('alias name conflict: ', name);
      exit;
    end;
    conAliasList.Append(name+'='+text);
  end;
end;

procedure CmdQAlias (const cmd: string; const args: array of string);
var
  name: string;
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('get alias value.');
    WriteLn('usage: alias? name');
    exit;
  end;

  name := ConGetArgStr(1, args);
  if name = '' then
  begin
    WriteLn('what alias?');
    exit;
  end;

  f := FindAlias(name);
  Write(name, ' is ');
  if f <> -1 then WriteLn('"', conAliasList.ValueFromIndex[f], '"')
  else WriteLn('undefined');
end;

{$IFDEF USE_ALISP}
function ConALIsArgHelp (nn: Integer): Boolean;
var
  s: string;
begin
  result := false;
  //nn := ALCdr(nn);
  if ALIsNil(nn) or not ALIsNil(ALCdr(nn)) then exit; // should be only one arg
  s := ALGetAtomStr(ALCar(nn));
  result := (s = '?') or (s = '-?') or (s = '/?');
end;

function ALCCmdList (nn: Integer): Integer;
var
  f: Integer;
begin
  if ConALIsArgHelp(nn) then
  begin
    WriteLn;
    WriteLn('return list of all available console commands');
    WriteLn('usage: (con-cmd-list)');
    result := ALTrue();
    exit;
  end;

  if conCommandList.Count = 0 then
  begin
    // no commands %-)
    result := ALCons(ALNil(), ALNil());
    exit;
  end;

  result := ALCons(ALNewStrAtom(conCommandList[0]), ALNil());
  ALPushCell(result);
  // add other commands (if any)
  for f := 1 to conCommandList.Count-1 do
  begin
    ALSetCdr(result, ALCons(ALNewStrAtom(conCommandList[f]), ALNil()));
    result := ALCdr(result);
  end;
  result := ALPopCells(1);
end;

function ALCAliasList (nn: Integer): Integer;

  function APair (s: string): Integer;
  var
    n, v: string;
  begin
    n := Copy(s, 1, Pos('=', s)-1);
    v := Copy(s, Length(n)+2, Length(s));
    result := ALCons(ALNewStrAtom(n), ALNewStrAtom(v));
  end;

var
  f: Integer;
begin
  if ConALIsArgHelp(nn) then
  begin
    WriteLn;
    WriteLn('return assoc-list with all defined aliases');
    WriteLn('usage: (con-alias-list)');
    result := ALTrue();
    exit;
  end;

  if conAliasList.Count = 0 then
  begin
    // no aliases
    result := ALCons(ALNil(), ALNil());
    exit;
  end;

  result := ALCons(APair(conAliasList[0]), ALNil());
  ALPushCell(result);
  // add other aliases (if any)
  for f := 1 to conAliasList.Count-1 do
  begin
    ALSetCdr(result, ALCons(APair(conAliasList[f]), ALNil()));
    result := ALCdr(result);
  end;
  result := ALPopCells(1);
end;

function ALCExecute (n: Integer): Integer;
var
  s: string;
begin
  result := ALTrue();
  if ConALIsArgHelp(n) then
  begin
    WriteLn;
    WriteLn('execute console command');
    WriteLn('usage: (con-execute [cmd]*)');
    exit;
  end;

  s := '';
  while not ALIsNil(n) do
  begin
    if s <> '' then s := s+' ';
    s := s+ALGetAtomStr(ALCar(n));
    n := ALCdr(n);
  end;

  ConExecute(s);
end;

function ALCVersion (n: Integer): Integer;
begin
  if ConALIsArgHelp(n) then
  begin
    WriteLn;
    WriteLn('get console version');
    WriteLn('usage: (con-version [full])');
    result := ALTrue();
    exit;
  end;

  if ALIsNil(n) then result := ALNewStrAtom(conVersion)
  else result := ALNewStrAtom(conFullVersion);
end;

function NumToStrBase (n: Int64; base: Integer = 10): string;
var
  d: Integer;
  neg: Boolean;
begin
  if base < 2 then base := 2 else if base > 36 then base := 36;
  neg := (n < 0); n := abs(n);
  result := '';
  repeat
    d := (n mod base); if d > 9 then Inc(d, 7);
    result := Char(d+ord('0'))+result;
    n := n div base;
  until n = 0;
  if neg then result := '-'+result;
end;

function ALCNumConvert (nn: Integer): Integer;
var
  num, base: Integer;
begin
  if ConALIsArgHelp(nn) then
  begin
    WriteLn;
    WriteLn('convert number to another base');
    WriteLn('usage: (num-convert number base)');
    result := ALTrue();
    exit;
  end;

  if not ALNextIntArg('NUM-CONVERT', nn, num, result) then exit;
  if not ALNextIntArg('NUM-CONVERT', nn, base, result) then exit;
  if (base < 2) or (base > 36) then begin result := ALError('invalid base', ALNewIntAtom(base)); exit; end;

  result := ALNewStrAtom(NumToStrBase(num, base));
end;

function ALCAddCommand (n: Integer): Integer;
var
  cn, sn: string;
begin
  result := ALTrue();
  if ConALIsArgHelp(n) then
  begin
    WriteLn;
    WriteLn('add console lisp-command.');
    WriteLn('usage: (con-add-cmd "cmdname" "symname")');
    WriteLn('symname should refer to global symbol');
    WriteLn('which is defined like:');
    WriteLn('(defun name (cmd-name arg-list) ... )');
    WriteLn('cmd-name: console command name');
    WriteLn('arg-list: list of console arguments.');
    WriteLn('each arg will be converted to the');
    WriteLn('corresponding type if it is not quoted,');
    WriteLn('i.e. to number, float or string.');
    exit;
  end;

  if not ALNextStrArg('CON-ADD-CMD', n, cn, result) then exit;
  if not ALNextStrArg('CON-ADD-CMD', n, sn, result) then exit;
  RegisterConsoleLispCommand(cn, sn);
end;

function ALCDeleteCommand (n: Integer): Integer;
var
  cn: string;
begin
  result := ALTrue();
  if ConALIsArgHelp(n) then
  begin
    WriteLn;
    WriteLn('delete console command.');
    WriteLn('usage: (con-delete-cmd "cmdname")');
    exit;
  end;

  if not ALNextStrArg('CON-DELETE-CMD', n, cn, result) then exit;
  UnregisterConsoleCommand(cn);
end;

procedure ConAddStdLispCommands ();
begin
  ALAddPrimitive('con-add-cmd', ALCAddCommand, false, 1, 2);
  ALAddPrimitive('con-delete-cmd', ALCDeleteCommand, false, 1, 1);
  ALAddPrimitive('con-version', ALCVersion, false, 0, 1);
  ALAddPrimitive('con-execute', ALCExecute, false, 1, -1);
  ALAddPrimitive('con-cmd-list', ALCCmdList, false, 0, 1);
  ALAddPrimitive('con-alias-list', ALCAliasList, false, 0, 1);
  ALAddPrimitive('num-convert', ALCNumConvert, false, 1, 2);
end;
{$ENDIF}


initialization
  conAliasList := TStringList.Create();

  conCommandList := TStringList.Create();
  conCommandList.Sorted := true; conCommandList.CaseSensitive := true;
  {$IFDEF USE_ALISP}
  conCommandLisp := TStringList.Create();
  ConAddStdLispCommands();
  {$ENDIF}
  RegisterConsoleCommand('con-version', @CmdConVersion);
  RegisterConsoleCommand('echo', @CmdEcho);
  RegisterConsoleCommand('cmd-list', @CmdCmdList);
  RegisterConsoleCommand('alias-list', @CmdAliasList);
  RegisterConsoleCommand('clear-aliases', @CmdClearAliases);
  RegisterConsoleCommand('alias', @CmdAlias);
  RegisterConsoleCommand('alias?', @CmdQAlias);
finalization
  {$IFDEF USE_ALISP}
  conCommandLisp := TStringList.Create();
  {$ENDIF}
  conCommandList.Free();
  conAliasList.Free();
end.
