// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
//!!! ADD LISP API!
// setting Speccy model & it's default environment
//
{.$DEFINE DUMPINFO}
unit dspMachineU;

interface

uses
  {$IFDEF DUMPINFO}Windows,{$ENDIF}
  SysUtils, Classes, lispCon3U,
  z80U, dspMachConstU, dspVScrU, dspZXDrawU, dspEmuVarsU,
  dsp48k128kU, dsp48kU, dsp128kU;


procedure ZXSetModel (model: TZXModel);
procedure ZXReset (trdos: Boolean = false);
procedure ZXSetProcs ();


implementation


procedure LoadROM (const fn: string; pageN: Integer);
var
  s: string;
  fi: TStream;
begin
  WriteLn('loading "', fn, '"');
  s := ExtractFilePath(ParamStr(0))+fn;
  if not FileExists(s) then s := fn;
  try
    fi := TFileStream.Create(s, fmOpenRead or fmShareDenyWrite);
    try
      if fi.Size <> 16384 then raise Exception.Create('invalid ROM file!');
      fi.ReadBuffer(emuMemory[pagen], 16384);
    finally
      fi.Free();
    end;
  except
    WriteLn('can''t load rom: "', s, '"');
    Halt(1);
  end;
end;


procedure ZXReset (trdos: Boolean = false);
var
  //p: Integer;
  f: Integer;
begin
  Write('Speccy reset');
  if trdos then Write(' to TR-DOS');
  WriteLn;

  if emuModel = zxm48k then emuLastOut7FFD := $10 else emuLastOut7FFD := 0;
  emuAYRegIndex := 0;
  emuAY.Reset();
  Z80Reset(emuZ80);
  //emuZ80.oneMoreRead := true;
  emuJoyMask := 0;
  for f := 0 to 7 do emuKeybMask[f] := $FF;
  emuKeysLocked := false;
  FillChar(pcKeys, SizeOf(pcKeys), 0);
  emuFramesToSkip := 0;
  emuReanimated := false;

  (*
  for p := 0 to 7 do
  begin
    for f := 0 to 16383 do
    begin
      //emuMemory[p, f] := Random(256);
      emuMemory[p, f] := 0;
    end;
  end;
  *)

  case emuModel of
    zxm48k:
      begin
        emuScreenPage := 5;
        emuPaging[0] := 9;
        emuPaging[1] := 5;
        emuPaging[2] := 2;
        emuPaging[3] := 0;
        emuPaging[4] := 8;
        emuLastOut7FFD := $10;
        emuBorderColor := 0;
      end;
    zxm128k, zxmPentagon128k:
      begin
        emuScreenPage := 5;
        emuPaging[0] := 8;
        emuPaging[1] := 5;
        emuPaging[2] := 2;
        emuPaging[3] := 0;
        emuPaging[4] := 8;
        emuLastOut7FFD := 0;
        emuBorderColor := 0;
      end;
  end;

  if trdos then
  begin
    emuPaging[0] := 10;
    emuLastOut7FFD := $10;
  end;
end;

procedure ZXSetProcs ();
begin
  // common procedures for memory i/o
  Z80PeekBInternal := PeekBIntComm;
  Z80PokeBInternal := PokeBIntComm;

  //Z80PeekBExec := DbgPeekBExecComm;
  if emuDbgActive then Z80PeekBExec := DbgPeekBExecComm
  else if config.contendedMemory then Z80PeekBExec := ContendPeekBExecComm
  else Z80PeekBExec := PeekBExecComm;

  if config.contendedMemory then Z80PeekB := ContendPeekBComm
  else Z80PeekB := PeekBComm;

  if config.contendedMemory then Z80PokeB := ContendPokeBComm
  else Z80PokeB := PokeBComm;

  case emuModel of
    zxm48k:
      begin
        Z80PortIn := PortIn48;
        Z80PortOut := PortOut48;
      end;
    zxm128k, zxmPentagon128k:
      begin
        Z80PortIn := PortIn128;
        Z80PortOut := PortOut128;
      end;
  end;
end;


function ScrWhereWeAre (time: Integer;
  out line, tsInLine, vtype, vofs: Integer): Boolean;
// result=false: the invisible area
// vtype=-2: the vertical retrace
// vtype=-1: the left border
// vtype=0: the screen
// vtype=1: the right border
// vtype=2: the horizontal retrace
// vofs: the offset from the first visible point of the current vtype
// line: from the top of the emulator screen, not the Speccy one %-)
//var
  //ofs: Integer;
begin
  result := false; vtype := -2;

  //if emuModel = zxm48k then ofs := 1 else ofs := 3;
  //ofs := 3;

  // the upper border (the invisible one)
  if time < emuLineTStates[0] then exit;

  // the lower border (the invisible one)
  if time >= emuLineTStates[borderV+192+borderV] then exit;

  // work out which line we're on, relative to the top of the screen
  line := (time-emuLineTStates[0]) div emuTicksLine;

  // work out where we are in this line
  tsInLine := (time+emuTicksLBord{-ofs}) mod emuTicksLine;

  if tsInLine < emuTicksLBord then
  begin
    // the left border
    vtype := -1;
    vofs := tsInLine-(emuTicksLBord-borderH div 2);
    result := (vofs >= 0);
    exit;
  end;

  if tsInLine < emuTicksLBord+128 then
  begin
    // the screen
    vtype := 0;
    vofs := tsInLine-emuTicksLBord;
    result := true;
    exit;
  end;

  if tsInLine < emuTicksLBord+128+emuTicksRBord then
  begin
    // the right border
    vtype := 1;
    vofs := tsInLine-emuTicksLBord-128;
    result := (vofs < borderH div 2);
    exit;
  end;

  // the horizontal retrace
  vtype := 2;
  result := false;
end;

function WhereWeAre (time: Integer): Integer;
var
  line, vtype, vofs: Integer;
begin
  ScrWhereWeAre(time, line, result, vtype, vofs);
end;

procedure SetupScrInfo (time: Integer);
var
  line, tsInLine, vtype, vofs: Integer;
  visible: Boolean;
begin
  // clear values
  emuDrawTickScrInfo[time].border := true;
  emuDrawTickScrInfo[time].zxscr := $FFFF;
  emuDrawTickScrInfo[time].zxattr := $FFFF;
  emuDrawTickScrInfo[time].vscrA := -1;
  emuDrawTickScrInfo[time].bshift := 1;

  visible := ScrWhereWeAre(time, line, tsInLine, vtype, vofs);
  if not visible then exit;

  if vtype = -1 then
  begin
    // the left border
    emuDrawTickScrInfo[time].vscrA := line*320+vofs*2;
    emuDrawTickScrInfo[time].bshift := 0;
    exit;
  end;

  if vtype = 1 then
  begin
    // the right border
    emuDrawTickScrInfo[time].vscrA := line*320+borderH+256+vofs*2;
    emuDrawTickScrInfo[time].bshift := 0;
    exit;
  end;

  // the screen or a border (the upper or the lower ones)
  emuDrawTickScrInfo[time].vscrA := line*320+borderH+vofs*2;
  emuDrawTickScrInfo[time].bshift := 0; // make the border visible

  if (line >= borderV) and (line-borderV < 192) then
  begin
    // the real screen
    emuDrawTickScrInfo[time].border := false;

    // work out which line we're on, relative to the top of the screen
    Dec(line, borderV);

    emuDrawTickScrInfo[time].zxscr := emuDrawZXLineOfs[line]+(vofs*2) div 8;
    ASSERT(emuDrawTickScrInfo[time].zxscr < 6144);
    emuDrawTickScrInfo[time].zxattr := 6144+(line div 8)*32+(vofs*2) div 8;
    ASSERT(emuDrawTickScrInfo[time].zxattr < 6912);
    emuDrawTickScrInfo[time].bshift := (vofs shl 1) and $07;
  end;
end;

function ContendDelay (time: Integer): Integer;
begin
  result := 0;
  if emuModel = zxmPentagon128k then exit; // no contented memory on Pentagon
  if emuModel = zxm48k then
  begin
    if time < 14335 then exit; // offscreen
    Dec(time, 14335);
    if time div 224 > 191 then exit; // offscreen
    if time mod 224 > 127 then exit; // borders, hretrace
    case time mod 8 of
      0: result := 6;
      1: result := 5;
      2: result := 4;
      3: result := 3;
      4: result := 2;
      5: result := 1;
    end;
  end
  else
  begin
    // +2A/+3 "read opcode" mystics are not implemented in DSP (acc. to Ph. Kendall)
    if time < 14361 then exit; // offscreen
    Dec(time, 14361);
    if time div 228 > 191 then exit; // offscreen
    if time mod 228 > 128 then exit; // borders, hretrace
    case time mod 8 of
      0: result := 1;
      //1: result := 0;
      2: result := 7;
      3: result := 6;
      4: result := 5;
      5: result := 4;
      6: result := 3;
      7: result := 2;
    end;
  end;
end;

procedure RealizeScrInfo ();
var
  f, a, x, y: Integer;
begin
  for f := 0 to emuTicksIntr do
  begin
    a := emuDrawTickScrInfo[f].vscrA;
    if a >= 0 then
    begin
      y := a div 320; x := a mod 320;
      if y >= 240 then
      begin
        //if f = emuTicksIntr then
        begin
          emuDrawTickScrInfo[f].vscrA := -1;
          emuDrawTickScrInfo[f].border := true;
          emuDrawTickScrInfo[f].bshift := 0;
        end
        //else ASSERT(y < 240);
      end
      else emuDrawTickScrInfo[f].vscrA := Integer(@(dspVScreenRam[239-y, x]));
    end;
  end;
end;

procedure ZXSetModel (model: TZXModel);
var
  f, c, tim: Integer;
begin
  if model = zxm128k then model := config.default128k
  else if model = zxm128kReal then model := zxm128k;

  for f := 0 to emuTicksIntr+32 do
  begin
    // clear values
    emuDrawTickScrInfo[f].border := true;
    emuDrawTickScrInfo[f].zxscr := $FFFF;
    emuDrawTickScrInfo[f].zxattr := $FFFF;
    emuDrawTickScrInfo[f].vscrA := -1;
    emuDrawTickScrInfo[f].bshift := 1;
  end;

  WriteLn('changing model to "', emuModelNames[model], '"');
  FillChar(emuXULAContention[0], SizeOf(emuXULAContention), 0);
  case model of
    zxm48k:
      begin
        emuModel := zxm48k;
        emuTicksIntr := 69888;
        emuTicksLBord := 24;
        emuTicksRBord := 24;
        emuTicksRetrace := 48;
        emuScrFirst := true;
        emuTicksLine := 128+emuTicksLBord+emuTicksRetrace+emuTicksRBord;
        ASSERT(emuTicksLine = 224);

        emuLineTStates[0] := 14333-24*emuTicksLine+3;
        //emuLineTStates[0] := 14336-24*emuTicksLine+3;
        for f := 1 to High(emuLineTStates) do
          emuLineTStates[f] := emuLineTStates[f-1]+emuTicksLine;

        for f := 0 to 239 do
        begin
          // screen, right border
          tim := emuLineTStates[f];
          for c := 0 to 127+16 do
          begin
            emuDrawTickScrInfo[tim].bshift := 0;
            emuDrawTickScrInfo[tim].vscrA := f*320+32+c*2;
            Inc(tim);
          end;

          // left border
          tim := emuLineTStates[f]+128+emuTicksRBord+emuTicksRetrace+emuTicksLBord-16;
          for c := 0 to 15 do
          begin
            emuDrawTickScrInfo[tim].bshift := 0;
            emuDrawTickScrInfo[tim].vscrA := (f+1)*320+c*2;
            Inc(tim);
          end;

          if (f < 24) or (f > 215) then continue;
          // screen
          tim := emuLineTStates[f];
          for c := 0 to 127 do
          begin
            emuDrawTickScrInfo[tim].border := false;
            emuDrawTickScrInfo[tim].bshift := (c shl 1) and $07;
            emuDrawTickScrInfo[tim].vscrA := f*320+c*2+32;
            emuDrawTickScrInfo[tim].zxscr := emuDrawZXLineOfs[f-24]+c div 4;
            emuDrawTickScrInfo[tim].zxattr := 6144+((f-24) div 8)*32+c div 4;
            Inc(tim);
          end;
        end;

        LoadROM(config.rom48k, 9);
        LoadROM(config.romtrdos, 10);
        emuAY.InitClock(1773000);
      end;
    zxm128k, zxm128kReal:
      begin
        emuModel := model;
        emuTicksIntr := 70908;

        emuTicksLBord := 24;
        emuTicksRBord := 24;
        emuTicksRetrace := 52;
        emuScrFirst := true;
        emuTicksLine := 128+emuTicksLBord+emuTicksRetrace+emuTicksRBord;
        ASSERT(emuTicksLine = 228);

        emuLineTStates[0] := 14362-24*emuTicksLine-emuTicksLBord+3;
        for f := 1 to High(emuLineTStates) do
          emuLineTStates[f] := emuLineTStates[f-1]+emuTicksLine;

        for f := 0 to 239 do
        begin
          // borders
          tim := emuLineTStates[f]+emuTicksLBord-16;
          for c := 0 to 159 do
          begin
            emuDrawTickScrInfo[tim].bshift := 0;
            emuDrawTickScrInfo[tim].vscrA := f*320+c*2;
            Inc(tim);
          end;

          if (f < 24) or (f > 215) then continue;
          // screen
          tim := emuLineTStates[f]+emuTicksLBord;
          for c := 0 to 127 do
          begin
            emuDrawTickScrInfo[tim].border := false;
            emuDrawTickScrInfo[tim].bshift := (c shl 1) and $07;
            emuDrawTickScrInfo[tim].vscrA := f*320+c*2+32;
            emuDrawTickScrInfo[tim].zxscr := emuDrawZXLineOfs[f-24]+c div 4;
            emuDrawTickScrInfo[tim].zxattr := 6144+((f-24) div 8)*32+c div 4;
            Inc(tim);
          end;
        end;

        LoadROM(config.rom128k_0, 8);
        LoadROM(config.rom128k_1, 9);
        LoadROM(config.romtrdos, 10);

        emuAY.InitClock(1773000);
      end;
    zxmPentagon128k:
      begin
        emuModel := zxmPentagon128k;
        emuTicksIntr := 71680;

        emuTicksLBord := 36;
        emuTicksRBord := 28;
        emuTicksRetrace := 32;
        emuScrFirst := true;
        emuTicksLine := 128+emuTicksLBord+emuTicksRetrace+emuTicksRBord;
        ASSERT(emuTicksLine = 224);

        //emuLineTStates[0] := 14336+3584-24*emuTicksLine-emuTicksLBord+3;
        emuLineTStates[0] := 17989-24*emuTicksLine-emuTicksLBord+3;
        for f := 1 to High(emuLineTStates) do
          emuLineTStates[f] := emuLineTStates[f-1]+emuTicksLine;

        for f := 0 to 239 do
        begin
          // borders
          tim := emuLineTStates[f]+emuTicksLBord-16;
          for c := 0 to 159 do
          begin
            emuDrawTickScrInfo[tim].bshift := 0;
            emuDrawTickScrInfo[tim].vscrA := f*320+c*2;
            Inc(tim);
          end;

          if (f < 24) or (f > 215) then continue;
          // screen
          tim := emuLineTStates[f]+emuTicksLBord;
          for c := 0 to 127 do
          begin
            emuDrawTickScrInfo[tim].border := false;
            emuDrawTickScrInfo[tim].bshift := (c shl 1) and $07;
            emuDrawTickScrInfo[tim].vscrA := f*320+c*2+32;
            emuDrawTickScrInfo[tim].zxscr := emuDrawZXLineOfs[f-24]+c div 4;
            emuDrawTickScrInfo[tim].zxattr := 6144+((f-24) div 8)*32+c div 4;
            Inc(tim);
          end;
        end;

        LoadROM(config.rom128k_0, 8);
        LoadROM(config.rom128k_1, 9);
        LoadROM(config.romtrdos, 10);

        emuAY.InitClock(1750000);
      end;
  end;

  emuZ80.tStatesPerFrame := emuTicksIntr; //!!!
  // calc line t-states
  if emuModel = zxm128k then
  begin
    for f := 1 to High(emuLineTStates) do
      emuLineTStates[f] := emuLineTStates[f-1]+emuTicksLine;
    for f := 0 to emuTicksIntr+32 do SetupScrInfo(f);
  end;

  // Set up the contention array
  for f := 0 to emuTicksIntr do emuXULAContention[f] := ContendDelay(f);

  (*
  WriteLn(' leftB: ', emuTicksLBord, '; scr: 128; rightB: ', emuTicksRBord,
    '; retrace: ', emuTicksRetrace);
  WriteLn('vis:');
  WriteLn(' leftB: ', emuTicksLBord-borderH div 2, '; scr: 128; rightB: ',
    borderH div 2);*)

  {$IFDEF DUMPINFO}
  AllocConsole();
  for f := 0 to emuTicksIntr-1 do
  begin
    if emuDrawTickScrInfo[f].border then
    begin
      if emuDrawTickScrInfo[f].bshift = 0 then Write('!') else Write(' ');
    end
    else Write('*');
    WriteLn(f, ': ',
      //'  clc=', WhereWeAre(f):3,
      '  addr=#', IntToHex(emuDrawTickScrInfo[f].vscrA, 8),
      '  scr=#', IntToHex(emuDrawTickScrInfo[f].zxscr+16384, 4),
      '  attr=#', IntToHex(emuDrawTickScrInfo[f].zxattr+16384, 4),
      '  shift=', emuDrawTickScrInfo[f].bshift);
    if f mod 1024 = 0 then SetConsoleTitle(PChar(IntToStr(f)));
  end;
  {$ENDIF}

  RealizeScrInfo();

  ZXSetProcs();
  ZXReset();
  ConExecute('?model');
end;


procedure CmdModel128 (const cmd: string; const args: array of string);
var
  s: string;
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('set default 128k model. usage: model128 model');
    WriteLn(' models: 128k, pentagon');
    exit;
  end;

  if Length(args) = 0 then
  begin
    WriteLn('what model?');
    exit;
  end;

  s := args[0];
  if Copy(s, 1, 3) = '128' then config.default128k := zxm128k
  else if Copy(s, 1, 4) = 'pent' then config.default128k := zxmPentagon128k
  else begin WriteLn('wtf?'); exit; end;

  WriteLn('default 128K model: "', emuModelNames[config.default128k], '"');
end;

procedure CmdReset (const cmd: string; const args: array of string);
var
  trd: Boolean;
  s: string;
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('reset speccy. usage: reset [model] [trdos]');
    WriteLn(' models:');
    WriteLn('  48k');
    WriteLn('  128k -- default 128k model');
    WriteLn('  pentagon');
    WriteLn('  !128 -- standard 128k model');
    exit;
  end;

  if Length(args) = 0 then begin ZXReset(); exit; end;

  if (Length(args) > 0) and (Copy(args[0], 1, 2) = 'tr') then
  begin
    trd := true;
    if Length(args) > 1 then s := args[1] else s := '';
  end
  else
  begin
    trd := (Length(args) > 1) and (Copy(args[1], 1, 2) = 'tr');
    s := args[0];
  end;

  if Copy(s, 1, 2) = '48' then ZXSetModel(zxm48k)
  else if Copy(s, 1, 3) = '128' then ZXSetModel(zxm128k)
  else if Copy(s, 1, 4) = '!128' then ZXSetModel(zxm128kReal)
  else if Copy(s, 1, 4) = 'pent' then ZXSetModel(zxmPentagon128k);
  ZXReset(trd);
end;

procedure CmdQModel (const cmd: string; const args: array of string);
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('print current emulating model info.');
    exit;
  end;

  WriteLn('model: "', emuModelNames[emuModel], '"');
  WriteLn('t-states per interrupt: ', emuTicksIntr);
  WriteLn('t-states per line: ', emuTicksLine);
  WriteLn('left border: ', emuTicksLBord);
  WriteLn('right border: ', emuTicksRBord);
  WriteLn('horiz. retrace: ', emuTicksRetrace);
end;

procedure CmdContention (const cmd: string; const args: array of string);
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('enable/disable contended memory emulation.');
    WriteLn('usage: contention [tan|ona]');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    config.contendedMemory := not config.contendedMemory;
  end
  else config.contendedMemory := ConYesNoValue(args[0]);
  WriteLn('contended memory: ', ConYesNoName(config.contendedMemory));
  ZXSetProcs();
end;

procedure CmdPortFF (const cmd: string; const args: array of string);
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('enable/disable port #FF emulation.');
    WriteLn('usage: portff [tan|ona]');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    config.emulatePortFF := not config.emulatePortFF;
  end
  else config.emulatePortFF := ConYesNoValue(args[0]);
  WriteLn('port #FF: ', ConYesNoName(config.emulatePortFF));
end;

(*
procedure CmdAccurateScreen (const cmd: string; const args: array of string);
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('enable/disable accurate screen update.');
    WriteLn('usage: accurate_screen [tan|ona]');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    config.accurateScreenUpdate := not config.accurateScreenUpdate;
  end
  else config.accurateScreenUpdate := ConYesNoValue(args[0]);
  WriteLn('port #FF: ', ConYesNoName(config.accurateScreenUpdate));
end;
*)

procedure CmdUseTRDOS (const cmd: string; const args: array of string);
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('enable/disable TR-DOS. usage: use_trdos [tan|ona]');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    config.useTRDOS := not config.useTRDOS;
  end
  else config.useTRDOS := ConYesNoValue(args[0]);
  WriteLn('tr-dos: ', ConYesNoName(config.useTRDOS));
end;

procedure CmdMatrixFX (const cmd: string; const args: array of string);
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('enable/disable keyboard matrix effect emulation.');
    Write('usage: matrixfx [tan|ona]');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    config.emulateMatrixEffect := not config.emulateMatrixEffect;
  end
  else config.emulateMatrixEffect := ConYesNoValue(args[0]);
  WriteLn('matrix f/x: ', ConYesNoName(config.emulateMatrixEffect));
end;

procedure CmdKMouse (const cmd: string; const args: array of string);
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('enable/disable kempston mouse emulation.');
    Write('usage: kempston_mouse [tan|ona]');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    config.useKempstonMouse := not config.useKempstonMouse;
  end
  else config.useKempstonMouse := ConYesNoValue(args[0]);
  WriteLn('kempston mouse: ', ConYesNoName(config.useKempstonMouse));
end;

procedure CmdKJoy (const cmd: string; const args: array of string);
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('enable/disable kempston joystick emulation.');
    Write('usage: kempston_joy [tan|ona]');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    config.useKempstonJoystick := not config.useKempstonJoystick;
  end
  else config.useKempstonJoystick := ConYesNoValue(args[0]);
  WriteLn('kempston joystick: ', ConYesNoName(config.useKempstonJoystick));
end;

procedure CmdPCJoy (const cmd: string; const args: array of string);
begin
  if (Length(args) > 0) and (args[0] = '?') then
  begin
    WriteLn;
    WriteLn('enable/disable PC joystick.');
    Write('usage: pc_joy [tan|ona]');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    config.usePCJoystick := not config.usePCJoystick;
  end
  else config.usePCJoystick := ConYesNoValue(args[0]);
  WriteLn('PC joystick: ', ConYesNoName(config.usePCJoystick));
  if config.usePCJoystick then ConExecute('joy-reset');
end;

procedure CmdPent512 (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable Pentagon 512K emulation.');
    WriteLn('usage: pent512 [tan|ona]');
    exit;
  end;

  config.pent512 := ConGetArgBoolDef(1, args, not config.pent512);
  WriteLn('pentagon 512K: ', ConYesNoName(config.pent512));
end;

procedure CmdDebugger (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable debugger.');
    WriteLn('usage: debugger [tan|ona]');
    exit;
  end;

  emuDbgActive := ConGetArgBoolDef(1, args, not emuDbgActive);
  ZXSetProcs();
  WriteLn('debugger: ', ConYesNoName(emuDbgActive));
end;


procedure DoContention (z80: PZ80CPU; addr: Word);
begin
  // contention
  if config.contendedMemory and (addr > 16383) and (addr <= 32767) then
    Inc(z80.tStates, emuXULAContention[z80.tStates]);
end;


begin
  Z80Contention := @DoContention;
  emuSetModel := @ZXSetModel;
  hackZXSetProcs := @ZXSetProcs;

  RegisterConsoleCommand('model128', @CmdModel128);

  RegisterConsoleCommand('reset', @CmdReset);
  RegisterConsoleCommand('?model', @CmdQModel);
  RegisterConsoleCommand('contention', @CmdContention);
  RegisterConsoleCommand('portff', @CmdPortFF);
  //*RegisterConsoleCommand('accurate_screen', @CmdAccurateScreen);
  RegisterConsoleCommand('use_trdos', @CmdUseTRDOS);
  RegisterConsoleCommand('matrixfx', @CmdMatrixFX);

  RegisterConsoleCommand('debugger', @CmdDebugger);

  RegisterConsoleCommand('kempston_mouse', @CmdKMouse);
  RegisterConsoleCommand('kempston_joy', @CmdKJoy);
  RegisterConsoleCommand('pc_joy', @CmdPCJoy);

  RegisterConsoleCommand('pent512', @CmdPent512);
end.
