// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// general port i/o: keyboard, kepston joystick & mouse, tape, AY, VG93
// the idea of PC keymapping was taken from Unreal Spectrum by SMT
// the idea of tape handling was taken from Unreal Spectrum by SMT
//
{.$DEFINE FUSE_FF}
unit dspPortGeneralU;

interface

uses
  SysUtils, z80U, dspTZXU, dspMachConstU, dspZXDrawU, dspEmuVarsU, dspEmuMiscU,
  dspEmuSoundU, dspEmuMouseU, dspLdrzU, lispCon3U;


function InPortGeneral (z80: PZ80CPU; port: Word): Byte;
procedure OutPortGeneral (z80: PZ80CPU; port: Word; b: Byte);


implementation


procedure TapeMaxSpeed (wasP: Boolean);
begin
  if config.tapeAutomax and not wasP and emuTapeInfo.playing and
     not emuMaxSpeed then ConExecute('speed max on');
end;

function InPortGeneral (z80: PZ80CPU; port: Word): Byte;
// partially taken from UnrealSpectrum
var
  p1: Byte;
  f: Integer;
  b: Byte;
  time: Integer;
  knownLoader: Integer;
  knownEdge: Integer;
  wasP: Boolean;
  {$IFDEF FUSE_FF}
  line, tstatesThroughLine, column, offset: Integer;
  {$ENDIF}
begin
  // contend port
  // the code is taken from FUSE
  // i even don't understand it's reasons %-)
  (*if config.contendedMemory then
  begin
    if ((port and $0001) <> 0) then // any even port is contended
    begin
      if (port and $C000) = $4000 then //???
      begin
        Inc(z80.tStates, emuXULAContention[z80.tStates]); Inc(z80.tStates);
        Inc(z80.tStates, emuXULAContention[z80.tStates]); Dec(z80.tStates);
      end
      else
      begin
        Inc(z80.tStates);
        Inc(z80.tStates, emuXULAContention[z80.tStates]); Dec(z80.tStates);
      end;
    end
    else
    begin
      if (port and $C000) = $4000 then //???
      begin
        Inc(z80.tStates, emuXULAContention[z80.tStates]); Inc(z80.tStates);
        Inc(z80.tStates, emuXULAContention[z80.tStates]); Inc(z80.tStates);
        Inc(z80.tStates, emuXULAContention[z80.tStates]); Inc(z80.tStates);
        Inc(z80.tStates, emuXULAContention[z80.tStates]); Dec(z80.tStates, 3);
      end;
    end;
  end;*)

  if config.contendedMemory and ((port and $0001) <> 0) then // any even port is contended
    Inc(z80.tStates, emuXULAContention[z80.tStates]);

  if (emuPaging[0] = 10) and ((port and $1F) = $1F) then
  begin
    // tr-dos stamped in, VG93 ports
    port := port and $FF;

    b := (port shr 4) and $0E;
    if b < $08 then emuDisks.DoDiskVG(b, result)
    else if b = $0E then emuDisks.DoDiskVG(8, result);
    exit;
  end;

  // AY ports
  if (port and $FF) = $FD then
  begin
    case ((port shr 8) shr 6) of
      //??? 2: result := emuAY.Registers[emuAYRegIndex];
      3: result := emuAY.Registers[emuAYRegIndex];
    end;
    exit;
  end;

  // kempstons (joystick & mouse)
  if (port and $20) = 0 then
  begin
    // SPANISH FIX: (spanish kempston: #DF, #1DF)
    //!port := (port and $FFFF) or $FA00; // A13,A15 not used in decoding
    if config.useKempstonMouse and
       ((port = $FADF) or (port = $FBDF) or (port = $FFDF)) then // mouse
    begin
      if config.autolockMouse and not emuMouseLocked then
         //!*(not emuZXWActive or (winDesktop.ActiveWindow = nil)) then
      begin
        //WriteLn('mouse autolocked');
        EmuLockMouse();
      end;
      // buttons: bit 0: left; bit 1: right; bit 2: middle
      if port = $FBDF then result := emuMousePos.x
      else if port = $FFDF then result := emuMousePos.y
      else
      begin
        result := not emuMouseButtons;
        if config.mouseSwap then
        begin
          result := result or $03;
          if (emuMouseButtons and $01) <> 0 then result := result and (not 2);
          if (emuMouseButtons and $02) <> 0 then result := result and (not 1);
        end;
      end;
    end
    else if config.useKempstonJoystick then
    begin
      result := emuJoyMask;
    end;
    exit;
  end;

  // port #FE (or so) -- keyboard, tape
  if (port and 1) = 0 then
  begin
    // keyboard
    result := $BF;
    if emuWConHeight = 0 then
    begin
      p1 := port shr 8;
      for f := 0 to 7 do
      begin
        if (p1 and (1 shl f)) = 0 then result := result and emuKeybMask[f];
      end;
    end;

    if config.tapeAutostart then
    begin
      wasP := emuTapeInfo.playing;
      if (z80.regPC = $0564) {and not emuTapeInfo.playing and (emuPaging[0] = 9)} then
      begin
        // tape autostart (& "fast-loader (utm)")
        if config.tapeFastLoader then
        begin
          EmuStartTape(true);
          ZXTapeStdFastLoader();
          //EmuTapeBit();
        end{
        else EmuStartTape(false)};
      end
      else if config.tapeFastLoader then
      begin
        if config.detectKnownLoaders then
        begin
          knownLoader := ZXTapeDetectLoader();

          if knownLoader <> ZXLDR_UNKNOWN then
          begin
            EmuStartTape(true);
            ZXTapeFastLoad(knownLoader);
          end{
          else TapeMaxSpeed(false)};
        end
        else knownLoader := ZXLDR_UNKNOWN;
        if (knownLoader = ZXLDR_UNKNOWN) and (config.detectEdgeRoutine) then
        begin
          knownEdge := ZXTapeDetectEdge();
          if knownEdge <> ZXEDGE_UNKNOWN then
          begin
            EmuStartTape(true);
            ZXTapeDoEdge(knownEdge);
          end{
          else TapeMaxSpeed(false)};
        end;
      end;
    end;

    if emuTapeInfo.playing then
    begin
      result := result or (EmuTapeBit() and $40);
    end
    else
    begin
      case config.tapeBit of
        0: result := result and (not $40);
        1: result := result or $40;
        else result := result or (emuTapeInfo.tapeBit and $40);
      end;
    end;
    exit;
  end;

  // port #FF
  result := $FF;

  if not config.emulatePortFF then exit;

  time := z80.tStates;
  {$IFNDEF FUSE_FF}
  // incorrect? but SideWise sprites doesn't blinking
  // and they're blinking with the FUSE method.
  if emuDrawTickScrInfo[time].vscrA < 0 then exit;
  if emuDrawTickScrInfo[time].border then exit;
  result := emuMemory[emuScreenPage, emuDrawTickScrInfo[time].zxattr];
  {$ELSE}
  // code taken from FUSE
  if emuModel = zxm48k then offset := 1 else offset := 3;

  // Return 0xff (idle bus) if we're in the top border
  if time < emuLineTStates[borderV] then exit;

  // Work out which line we're on, relative to the top of the screen
  line := (time-emuLineTStates[borderV]) div emuTicksLine;

  // Idle bus if we're in the lower or upper borders
  if line >= 192 then exit;

  // Work out where we are in this line
  tstatesThroughLine := time-emuLineTStates[borderV+line];

  // Idle bus if we're in the left border
  if tstatesThroughLine < emuTicksLBord-offset then exit;

  // Or the right border or retrace
  if tstatesThroughLine >= emuTicksLBord+128-offset then exit;

  column := ((tstatesThroughLine+1-emuTicksLBord) div 8)*2;

  tstatesThroughLine := tstatesThroughLine mod 8;
  if (tstatesThroughLine = 2) or (tstatesThroughLine = 3) then Inc(column);
  if (tstatesThroughLine mod 8 = 0) or (tstatesThroughLine mod 8 = 2) then
  begin
    // bitmap
    result := emuMemory[emuScreenPage, emuDrawZXLineOfs[line]+column];
    exit;
  end;

  if (tstatesThroughLine mod 8 = 1) or (tstatesThroughLine mod 8 = 3) then
  begin
    // attrs
    result := emuMemory[emuScreenPage, 6144+(line div 8)*32+column];
    exit;
  end;
  // other -- idle
  {$ENDIF}
end;

procedure OutPortGeneral (z80: PZ80CPU; port: Word; b: Byte);
var
  bt: Byte;
  t: Boolean;
  te: Integer;
begin
  // only INs are contended?
  if config.contendedMemory and ((port and $0001) <> 0) then // any even port is contended
    Inc(z80.tStates, emuXULAContention[z80.tStates]);

  if (emuPaging[0] = 10) and ((port and $1F) = $1F) then
  begin
    // VG-93
    emuWasDiskActivity := true;
    bt := (port shr 4) and $0F;
    if bt < 8 then emuDisks.DoDiskVG(bt, b)
    else if bt = $0F then emuDisks.DoDiskVG(9, b);
    exit;
  end;

  if (port and $FF) = $FD then
  begin
    // AY
    case ((port shr 8) shr 6) of
      2:
        begin
          emuAY.Registers[emuAYRegIndex] := b;
          emuAY.Update();
          if not emuMaxSpeed and (emuFramesToSkip < 0) then EmuSoundFillBuffer(emuZ80.tStates-0); // minus what? %-)
        end;
      3: emuAYRegIndex := b and $0F;
    end;
  end
  else if (port and $01) = 0 then
  begin
    //if (emuFramesToSkip < 0) and config.accurateScreenUpdate then
    if emuFramesToSkip < 0 then
    begin
      te := z80.tStates;
      if (te >= emuLineTStates[0]) and (te < emuLineTStates[240]) then
        ZXExUpdateScr(te-1, @emuMemory[emuScreenPage]);
    end;
    emuBorderColor := (b and $07);
    t := (b and $10) <> 0;
    if t <> emuSpeakerState then
    begin
      // speaker state changed
      // FIXME: fast speaker state changing produces the low volume sound
      // render the sound
      if not emuMaxSpeed then EmuSoundFillBuffer(emuZ80.tStates);
      // and change the variable
      emuSpeakerState := t;
    end;
    //MicOut((b and $08) <> 0);
  end;
end;


end.
