// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// shitty sound driver
//
unit dspFMU;

interface

uses
  Windows, dspFModU, fmodtypes, fmoderrors, dspAYEngineU, dspMachConstU;


type
  PSmallInt = ^SmallInt;
  PLongWord = ^LongWord;


const
  bufSize = (44100 div 50); // samples per frame
  mixBufCount = 16;


var
  mixBuffers: packed array [0..mixBufCount-1, 0..bufSize-1, 0..1] of SmallInt; // stereo mixing buffer
  mixBufNum: Integer; // buffer to mix into
  mixBufAdvanceCounter: Integer; //++ on each buffer advance
  sndPaused: Boolean = false;


procedure SoundFillBuffer (cs: Integer);
procedure SoundFinishFillBuffer ();

procedure SoundPause (paused: Boolean);

function SoundInit (): Boolean;
procedure SoundDeinit();


implementation

uses
  dspEmuVarsU;


var
  stream: PFSoundStream;
  channel: Integer;
  prevMixP: Integer = 0;
  prevWasMix: Boolean = false;

  soundInitOk: Boolean = false;


procedure SoundPause (paused: Boolean);
begin
  if paused = sndPaused then exit;
  sndPaused := paused;

  if soundInitOk then FSOUND_SetPaused(Channel, paused);

  Write('SND: the sound is ');
  if not paused then Write('un');
  WriteLn('paused');
end;

procedure SoundFillBufferToPos (cpos: Integer);
// fill sound from ps to cs t-states
var
  ppos, n: Integer;
  buffer: PChar;
begin
  if sndPaused then
  begin
    prevMixP := bufSize-1;
    exit;
  end;

  ppos := prevMixP;
  if cpos >= bufSize then cpos := bufSize-1;

  buffer := PChar(@mixBuffers[mixBufNum, ppos]);
  if emuSoundEnabled and not emuMaxSpeed then
  begin
    prevWasMix := true;
    if emuAYEnabled then
    begin
      while ppos <= cpos do
      begin
        emuAY.RenderSample();
        n := (emuAY.LeftChannel-24)*1000;
        if not emuSpeakerState then Inc(n, 16384) else Dec(n, 16384);
        if n < -32767 then n := 32767 else if n > 32767 then n := 32767;
        PSmallInt(buffer)^ := n;
        Inc(buffer, 2);
        n := (emuAY.RightChannel-24)*1000;
        if not emuSpeakerState then Inc(n, 16384) else Dec(n, 16384);
        if n < -32767 then n := 32767 else if n > 32767 then n := 32767;
        PSmallInt(buffer)^ := n;
        Inc(buffer, 2);
        Inc(ppos);
      end;
    end
    else
    begin
      if not emuSpeakerState then n := 32760 else n := -32760;
      while ppos <= cpos do
      begin
        PSmallInt(buffer)^ := n;
        Inc(buffer, 2);
        PSmallInt(buffer)^ := n;
        Inc(buffer, 2);
        Inc(ppos);
      end;
    end;
  end
  else
  begin
    if prevWasMix then
    begin
      n := SizeOf(mixBuffers) div 2;
      asm
        mov  edi, offset mixBuffers
        mov  ecx, n
        mov  al, emuSpeakerState
        or   al, al
        mov  eax, 32760
        jz   @@1
        neg  eax
    @@1:
        cld
        rep stosw
      end;
      prevWasMix := false;
    end;
    {while ppos <= cpos do
    begin
      PLongWord(buffer)^ := 0;
      Inc(buffer, 4);
      Inc(ppos);
    end;}
  end;
  prevMixP := ppos;
end;

procedure SoundFillBuffer (cs: Integer);
// fill sound from ps to cs t-states
var
  cpos: Integer;
begin
  cpos := trunc(bufSize/emuTicksIntr{tpb}*cs);
  SoundFillBufferToPos(cpos);
end;

procedure SoundFinishFillBuffer ();
begin
  SoundFillBufferToPos(bufSize);
  //if emuSoundEnabled then emuAY.Update();
end;


function StreamCallback (stream: PFSoundStream; buff: Pointer; len, param: Integer): ByteBool; stdcall;
begin
  //WriteLn(len, ' : ', bufSize*4);
  //if prevMixP < bufSize then SoundFillBufferToPos(bufSize);

  if emuMaxSpeed then FillChar(buff^, len, 0)
  else Move(mixBuffers[mixBufNum], buff^, len);

  mixBufNum := (mixBufNum+1) mod mixBufCount;
  prevMixP := 0;
  Inc(mixBufAdvanceCounter);

  // Must return true for the stream to continue
  result := true;
end;

function SoundInit (): Boolean;
begin
  if soundInitOk then begin result := true; exit; end;
  mixBufNum := 0; prevMixP := 0; result := false;
  soundInitOk := false;

  if FMOD_VERSION > FSOUND_GetVersion() then
  begin
    WriteLn('***ERROR in dspFMU.pas: invalid version');
    exit;
    //Halt(1);
  end;

  FSOUND_SetOutput(FSOUND_OUTPUT_DSOUND);
  //FSOUND_SetOutput(FSOUND_OUTPUT_WINMM);
  //FSOUND_SetOutput(FSOUND_OUTPUT_NOSOUND);
  FSOUND_SetDriver(0); // default
  if not FSOUND_Init(44100, 8,
    FSOUND_INIT_GLOBALFOCUS or FSOUND_INIT_ACCURATEVULEVELS) then
  begin
    WriteLn('***ERROR in dspFMU.pas: can''t initialize FMOD');
    //WriteLn(FMOD_ErrorString(FSOUND_GetError()));
    FSOUND_Close();
    //Halt(1);
    exit;
  end;

  stream := FSOUND_Stream_Create(StreamCallback,
    bufSize*2*2, {16 bit, stereo}
    FSOUND_LOOP_NORMAL or FSOUND_NORMAL or FSOUND_16BITS or FSOUND_STEREO,
    44100, 12345);
  if stream = nil then
  begin
    WriteLn('***ERROR in dspFMU.pas: can''t open sound stream');
    //WriteLn(FMOD_ErrorString(FSOUND_GetError()));
    FSOUND_Close();
    //Halt(1);
    exit;
  end;

  mixBufAdvanceCounter := 0;
  channel := FSOUND_Stream_Play(FSOUND_FREE, stream);
  if channel < 0 then
  begin
    WriteLn('***ERROR in dspFMU.pas: can''t start sound stream');
    //WriteLn(FMOD_ErrorString(FSOUND_GetError()));
    //FSOUND_Close();
    exit;
  end;

  soundInitOk := true;
  result := true;
end;

procedure SoundDeinit ();
begin
  if not soundInitOk then exit;
  FSOUND_Stream_Close(stream);
  FSOUND_Close();
  soundInitOk := false;
end;


initialization
  SoundInit();
finalization
  SoundDeinit();
end.
