// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// packing/depacking routines, v0.0.1
//
// -----------------------------------------------------------------------------
// this procedures are really slow, but i just don't care: they're not
// written to compress gigabyte files.
//
// ----------------------------------------------------------------------------
//
// block file format:
//   word: # of blocks (not including the last partially filled one)
//   word: bytes in the last block
//     this field can be 0 if the last block is empty
// then full blocks:
//   16 bytes: unpacked data md5
//   word: packed data size
//     if this field is 0: unpacked 65536 bytes follows,
//     else packed bytes follows
// then last block (if not empty):
//   16 bytes: md5
//   byte: packed flag (<>0: block is packed)
//     if packed:
//       word: packed data size
//     data
//
// ----------------------------------------------------------------------------
//
// ark file format:
//   db  'ARK0'     ; signature
//   dw  fileCount
//   dw  dirCount
//   dd  dirOffset  ; in the unpacked file
// then xblocks follows
//
// directory format:
//   dirCount directories
//   fileCount files
//
// file entry:
//   name (byte-counted string)
//   dw  directory_index
//   dd  fileSize
//
{.$DEFINE XPACK_CRYPT_BLOCKZ}
{$DEFINE XPACK_INCLUDE_BLOCK_FUNCZ}
{$DEFINE XPACK_INTERNAL_MD5}
{.$DEFINE XPACK_INCLUDE_ARK_FUNCZ}


{$DESCRIPTION 'XPACK v0.0.1 by Ketmar // Vampire Avalon'}

{$MESSAGE ''}
{$MESSAGE 'compiling XPACK v0.0.1'}

{$IFDEF XPACK_INCLUDE_ARK_FUNCZ}
  {$DEFINE XPACK_INCLUDE_BLOCK_FUNCZ}
{$ENDIF}


{$IFDEF XPACK_INCLUDE_BLOCK_FUNCZ}
  {$IFDEF XPACK_CRYPT_BLOCKZ}
    {$MESSAGE 'XPACK: crypted block i/o included'}
  {$ELSE}
    {$MESSAGE 'XPACK: block i/o included'}
  {$ENDIF}
  {$IFDEF XPACK_INTERNAL_MD5}
    {$MESSAGE 'XPACK: using internal MD5'}
  {$ENDIF}
  {$IFDEF XPACK_INCLUDE_ARK_FUNCZ}
    {$MESSAGE 'XPACK: ARK i/o included'}
  {$ENDIF}
{$ENDIF}
{$MESSAGE ''}

{.$DEFINE DEBUG}

{$ASSERTIONS ON}
{$ALIGN 8}
{$BOOLEVAL OFF}
{$EXTENDEDSYNTAX ON}
{$IMPORTEDDATA OFF} // was ON
{$LONGSTRINGS ON}
{$IOCHECKS ON}
{$WRITEABLECONST OFF}
{$MINENUMSIZE 1}
{$OPENSTRINGS ON}
{$OVERFLOWCHECKS OFF}
{$RANGECHECKS OFF}
{$REALCOMPATIBILITY OFF}
{$SAFEDIVIDE OFF}
{$TYPEDADDRESS OFF}
{$TYPEINFO OFF}
{$VARSTRINGCHECKS OFF}
{$E-,F-,K-,N+,S-} //???

{$IFDEF DEBUG}
  {$DEBUGINFO ON}
  {$LOCALSYMBOLS ON}
  {$DEFINITIONINFO ON}
  {$HINTS ON}
  {$WARNINGS ON}
  {$OPTIMIZATION OFF}
  {$STACKFRAMES ON}
{$ELSE}
  {$DEBUGINFO OFF}
  {$LOCALSYMBOLS OFF}
  {$DEFINITIONINFO OFF}
  {$HINTS ON}
  {$WARNINGS OFF}
  {$OPTIMIZATION ON}
  {$STACKFRAMES OFF}
{$ENDIF}
unit xpackU;

interface

{$IFDEF XPACK_INCLUDE_BLOCK_FUNCZ}
uses
  SysUtils, Classes{$IFNDEF XPACK_INTERNAL_MD5}md5U{$ENDIF};
{$ENDIF}


function XPackBuffer (const data; dataSize: Integer;
  out packBuf; packBufSize: Integer): Integer;
  // result < 0: out of packBuf; else: size of the compressed data
function XUnpackBuffer (const data; dataSize: Integer;
  out dest; destBufSize: Integer): Integer;
  // result < 0: out of dest; else: size of the decompressed data
  // ACHTUNG! there's no "end-of-compressed-stream" flag, so you
  //          can't determine packed data size.


{$IFDEF XPACK_INCLUDE_BLOCK_FUNCZ}
type
  TXArkBlockFile = class
  protected
    fStream: TStream;
    fCount: Integer;
    fOfs: TList; // offset to each block

    fLastBlockData: packed array [0..65536] of Byte;
      // last buffer
    fLastBlockUsed: Integer;
      // bytes used in the last buffer. can be 0 if
      // the last buffer is empty
    fChanged: Boolean;

    fCachedBlockData: packed array [0..65536] of Byte;
    fCachedBlock: Integer;
    fCurBlock: Integer;
    fCurOfs: Integer;

    function GetSize (): Integer;
    function GetPosition (): Integer;
    procedure SetPosition (value: Integer);

  public
    constructor Create (pStream: TStream; testAll: Boolean = false);
    constructor CreateNew (pStream: TStream);
    destructor Destroy (); override;

    function ReadBlock (index: Integer; out buf): Integer;
      // returns unpacked block size (1-65536)
      // raises an exception if there's no such block

    procedure ReadBuffer (out buf; count: Integer);
    procedure WriteBuffer (const buf; count: Integer);

    procedure Flush ();

    property Count: Integer read fCount;
    property Size: Integer read GetSize;
    property Position: Integer read GetPosition write SetPosition;
  end;
{$ENDIF}

{$IFDEF XPACK_INCLUDE_ARK_FUNCZ}
type
  EXArkError = class (Exception);

  TXArkFile = class;

  TXArkFileStream = class (TStream)
  protected
    fArk: TXArkFile;
    fStream: TStream;
    fSize: Integer;
    fStart: Integer;
    fPos: Integer;

    constructor Create (srcArk: TXArkFile; srcS: TStream; pStart, pSize: Integer);

  public
    destructor Destroy (); override;

    function Read (var buffer; count: Longint): Longint; override;
    function Write (const buffer; count: Longint): Longint; override;
    function Seek (offset: Longint; origin: Word): Longint; override;
  end;

  PXArkDirEntry = ^TXArkDirEntry;
  TXArkDirEntry = record
    name: ShortString;
    dir: Word; // index of path dir
    size: Integer;
    ofs: Integer; // in the block file
  end;

  TXArkFile = class
  protected
    fStream: TStream;
    fFileName: string; // if not empty -- Free() fStream on Destroy()ing
    fDirs: TStringList;
    fFiles: TList;
    fReadMode: Boolean;
    fBlockF: TXArkBlockFile;

    fOpenFileCount: Integer;

    procedure ClearDir ();
    procedure LoadDir (checkFile: Boolean);
    procedure SaveDir ();

    function GetFileCount (): Integer;
    function GetFilePath (index: Integer): string;
    function GetFileName (index: Integer): string;
    function GetFileSize (index: Integer): Integer;
    function GetArkFileSize (): Integer;

  public
    constructor Create ();
    destructor Destroy (); override;

    procedure Close ();

    procedure OldFromStream (srcS: TStream; checkFile: Boolean = false);
    procedure OldFromFile (const fName: string; checkFile: Boolean = false);

    procedure NewToStream (dstS: TStream);
    procedure NewToFile (const fName: string);

    procedure UpdateToStream (srcS: TStream);
    procedure UpdateToFile (const fName: string);

    procedure NewFile (const fName: string);
    procedure NewClose ();
    procedure NewWrite (const buf; len: Integer);
    procedure AppendStream (const fName: string; srcS: TStream; len: Integer);
      // len <= 0: whole srcS
    procedure AppendFile (const fName, srcFName: string);

    function IsAppending (): Boolean;

    function OpenFile (const fName: string): TStream;
      // returns nil if no such file
    function OldFile (const fName: string): TStream;
      // raises an exception if no such file

    property ReadMode: Boolean read fReadMode;
    property FileCount: Integer read GetFileCount;
    property FilePath [index: Integer]: string read GetFilePath;
    property FileName [index: Integer]: string read GetFileName;
    property FileSize [index: Integer]: Integer read GetFileSize;

    property AppendMode: Boolean read IsAppending;

    property ArkFileSize: Integer read GetArkFileSize;
  end;
{$ENDIF}


implementation


function XPackBuffer (const data; dataSize: Integer; out packBuf; packBufSize: Integer): Integer;
type
  PByte = ^Byte;
  PWord = ^Word;
var
  sp: PChar;
  inP, outP: Integer;
  curB: Integer;
  curBPos: Integer;
  lastEmitChars: Integer; // pos

  // position tree for 2-chars matches
  last: array of Integer; // last occurance
  tree: array of Integer; // next occurances

  procedure BuildTree ();
  // build search tree (dictionary is static -- lucky me %-)
  var
    f: Integer;
    ch: Char;
    n: Word;
    bufP: PChar;
  begin
    SetLength(last, 65536);
    SetLength(tree, dataSize);
    for f := 0 to High(last) do last[f] := -1;
    for f := 0 to High(tree) do tree[f] := -1;
    bufP := @data; Dec(bufP);
    for f := 0 to dataSize-2 do
    begin
      Inc(bufP);
      if f+2 < dataSize then
      begin
        // 3 repeating chars will be compressed as "fill", so
        // don't include such duplets in the tree
        ch := bufP^;
        if (bufP[1] = ch) and (bufP[2] = ch) then continue;
      end;

      n := PWord(bufP)^;
      if last[n] = -1 then last[n] := f else tree[f] := last[n];
      last[n] := f;
    end;
  end;

  procedure EmitBit (val: Byte);
  begin
    if curB > 255 then
    begin
      if curBPos < packBufSize then
        PChar(@packBuf)[curBPos] := Char(curB and $FF);
      curB := 1; curBPos := outP; Inc(outP);
    end;
    curB := (curB shl 1); if val <> 0 then curB := curB or 1;
  end;

  procedure EmitChar (val: Char);
  begin
    if outP < packBufSize then PChar(@packBuf)[outP] := val;
    Inc(outP);
  end;

  procedure EmitByte (val: Byte);
  begin
    EmitChar(Char(val));
  end;

  procedure EmitWord (val: Word);
  begin
    EmitByte(val and $FF);
    EmitByte((val shr 8) and $FF);
  end;

  procedure EmitLen (len: Integer);
  begin
    ASSERT(len > 0);
    ASSERT(len < 65537);
    if len = 1 then begin EmitBit(0); exit; end;

    EmitBit(1);
    if len < 6 then
    begin
      EmitBit(0);
      Dec(len, 2); EmitBit(len and $02); EmitBit(len and $01);
      exit;
    end;

    EmitBit(1);
    if len < 22 then
    begin
      EmitBit(0);
      Dec(len, 6);
      EmitBit(len and $08); EmitBit(len and $04);
      EmitBit(len and $02); EmitBit(len and $01);
      exit;
    end;

    EmitBit(1);
    if len < 278 then
    begin
      EmitBit(0);
      EmitByte(len-22);
    end
    else
    begin
      EmitBit(1);
      EmitWord(len and $FFFF);
    end;
  end;

  procedure EmitChars ();
  // if there was some uncompressed chars -- write 'em
  var
    len: Integer;
  begin
    len := inP-lastEmitChars;
    ASSERT(len >= 0);
    if len = 0 then exit;

    EmitLen(len);
    EmitBit(0); // "char stream" flag
    while len > 0 do
    begin
      EmitChar(sp[lastEmitChars]); Inc(lastEmitChars);
      Dec(len);
    end;
  end;

  procedure EmitFill (len: Integer);
  // emit "fill" control code
  var
    ch: Char;
    short: Boolean;
  begin
    EmitChars();
    ch := sp[inP]; Inc(inP, len); lastEmitChars := inP;

    short := (ch = #$00) or (ch = #$FF);
    if short then EmitLen(len) else EmitLen(len-1);

    EmitBit(1); // "copy/fill" flag
    EmitBit(0); // "fill" flag

    // special cases
    if short then
    begin
      EmitBit(0); // "fill #00/#FF" flag
      EmitBit(Byte(ch)); // "#00/#FF" flag
      exit;
    end;

    // normal case
    ASSERT(len > 1);

    EmitBit(1); // "fill char" flag
    // output fill byte
    EmitChar(ch);
  end;

  procedure EmitCopy (pos, len: Integer);
  // emit "copy" control code
  var
    ofs: Integer;
  begin
    EmitChars();

    ofs := inP-pos;
    Inc(inP, len); lastEmitChars := inP;

    EmitLen(len-2);
    EmitBit(1); // "copy/fill" flag
    EmitBit(1); // "copy" flag
    EmitLen(ofs);
  end;

  function RepLen (): Integer;
  // returns # of repeat bytes from the current pos
  begin
    result := 1;
    while (result < 65535) and (inP+result < dataSize) and (sp[inP+result] = sp[inP]) do Inc(result);
  end;

  function FindLongestMatch (out len: Integer): Integer;
  // returns match pos or -1
  // shitty code... %-(
  var
    c, lastMatchPos: Integer;
  begin
    lastMatchPos := -1; // no matches yet
    len := 2; // 1-char and 2-chars matches are useless
    repeat
      Inc(len);
      result := last[PWord(@sp[inP])^];
      while result <> -1 do
      begin
        if result < inP then
        begin
          c := 2; // first two chars already equals %-), check others
          while (c < len) and (result+c < inP+c) and (sp[result+c] = sp[inP+c]) do Inc(c);
          if c = len then
          begin
            // the match was found
            lastMatchPos := result;
            // try to increase len
            while (inP+len < dataSize) and (result+len < inP+len) and
                  (sp[result+len] = sp[inP+len]) do Inc(len);
            if inP+len = dataSize then exit;
            break;
          end;
        end;
        // next possible match position
        result := tree[result];
      end;
      // exit if no matches was found at this iteration
    until result = -1;
    result := lastMatchPos; Dec(len);
  end;

var
  repC, matchPos, matchLen: Integer;
  ch: Char;
begin
  result := -1;
  if (packBufSize < 1) or (dataSize < 1) or (packBufSize > 65536) then exit;

  BuildTree();

  sp := @data; inP := 0;
  curB := 1; curBPos := 0; outP := 1;
  lastEmitChars := 0;
  while (inP < dataSize) and (outP <= packBufSize) do
  begin
    repC := RepLen();
    ch := sp[inP];
    // output "fill" command?
    // bytes #00 and #FF is a special cases
    if (ch = #$00) or (ch = #$FF) or (repC >= 3) then
    begin
      EmitFill(repC); // inP was corrected by EmitFill
      continue;
    end;

    // search match
    if (inP < 2) or (inP >= dataSize-2) then
    begin
      // no suitable matches can be found at this position,
      // advance char and continue
      Inc(inP);
      continue;
    end;

    matchPos := FindLongestMatch(matchLen);
    if matchPos = -1 then Inc(inP) // no match, advance char
    else EmitCopy(matchPos, matchLen); // inP was corrected by EmitCopy
  end;
  // emit uncomressed chars (if any)
  EmitChars();

  if curB <> 1 then
  begin
    // control byte is not empty, so there's some bytes in the output buffer
    while curB < 255 do curB := curB shl 1;
    if curBPos < packBufSize then PChar(@packBuf)[curBPos] := Char(curB and $FF);
  end;

  if (inP < dataSize) or (outP > packBufSize) then result := -1 else result := outP;
end;

function XUnpackBuffer (const data; dataSize: Integer; out dest; destBufSize: Integer): Integer;
type
  PByte = ^Byte;
var
  sp, dp: PChar;
  curB, bCount: Integer;
  curAddr: Integer;
  len: Integer;
  srcA, ofs: Integer;
  fillB: Byte;

  procedure EmitByte (ch: Integer);
  begin
    if destBufSize > 0 then dp^ := Char(ch);
    Inc(dp); Dec(destBufSize); Inc(curAddr);
  end;

  function GetBit (): Byte;
  begin
    if bCount <= 0 then
    begin
      curB := Byte(sp^); Inc(sp); Dec(dataSize);
      bCount := 8;
    end;
    if (curB and $80) <> 0 then result := 1 else result := 0;
    curB := (curB shl 1) and $FF;
    Dec(bCount);
  end;

  function GetByte (): Integer;
  begin
    result := Byte(sp^); Inc(sp); Dec(dataSize);
  end;

  function GetWord (): Integer;
  // result := GetByte() or (GetByte() shl 8);
  // the previous code gives the bytes swapped when
  // optimization is off. delphi bug? or debugger lies?
  begin
    result := GetByte();
    result := result or (GetByte() shl 8);
  end;

  function GetLen (): Integer;
  begin
    if GetBit() = 0 then begin result := 1; exit; end;

    if GetBit() = 0 then
    begin
      result := ((GetBit() shl 1) or GetBit())+2;
      exit;
    end;

    if GetBit() = 0 then
    begin
      result := ((GetBit() shl 3) or (GetBit() shl 2) or (GetBit() shl 1) or
        GetBit())+6;
      exit;
    end;

    if GetBit() = 0 then
    begin
      result := GetByte()+22;
    end
    else
    begin
      result := GetWord(); if result = 0 then result := 65536;
    end;
  end;

  procedure DoCopy (addr, len: Integer);
  begin
    while len > 0 do
    begin
      EmitByte(Byte(PChar(@dest)[addr])); Inc(addr); Dec(len);
    end;
  end;

  procedure DoStream (len: Integer);
  begin
    while len > 0 do
    begin
      EmitByte(GetByte()); Dec(len);
    end;
  end;

  procedure DoFill (value: Integer; len: Integer);
  begin
    while len > 0 do
    begin
      EmitByte(value); Dec(len);
    end;
  end;

begin
  result := -1;
  if (dataSize < 1) or (destBufSize < 1) or (dataSize > 65536) then exit;

  sp := @data; dp := @dest; curAddr := 0;
  // get "flag byte"
  curB := Byte(sp^); Inc(sp); Dec(dataSize); bCount := 8;
  while destBufSize > 0 do
  begin
    len := GetLen(); // len is always here %-)
    if GetBit() = 0 then
    begin
      DoStream(len);
    end
    else
    begin
      // copy/fill
      if GetBit() = 0 then
      begin
        // fill
        if GetBit() = 0 then
        begin
          // fill #00/#FF
          fillB := GetBit(); if fillB <> 0 then fillB := $FF;
        end
        else
        begin
          // fill char/long fill
          fillB := GetByte();
          Inc(len);
        end;
        DoFill(fillB, len);
      end
      else
      begin
        // copy
        Inc(len, 2);
        ofs := GetLen();
        srcA := curAddr-ofs;
        if srcA < 0 then begin dataSize := 1; break; end;
        DoCopy(srcA, len);
      end;
    end;
    if dataSize < 0 then begin dataSize := 1; break; end;
    if destBufSize < 0 then begin dataSize := 1; break; end;
  end;

  if (dataSize <> 0) or (destBufSize <> 0) then result := -1 else result := curAddr;
end;


{$IFDEF XPACK_INCLUDE_BLOCK_FUNCZ}
{$IFDEF XPACK_INTERNAL_MD5}
// This code implements the MD5 message-digest algorithm.
// The algorithm is due to Ron Rivest.  This code was
// written by Colin Plumb in 1993, no copyright is claimed.
// This code is in the public domain; do with it what you wish.
//
// Equivalent code is available from RSA Data Security, Inc.
// This code has been tested against that, and is equivalent,
// except that you don't need to include two pages of legalese
// with every copy.
//
// To compute the message digest of a chunk of bytes, declare an
// MD5Context structure, pass it to MD5Init, call MD5Update as
// needed on buffers full of bytes, and then call MD5Final, which
// will fill a supplied 16-byte array with the digest.
//
// Changed so as no longer to depend on Colin Plumb's `usual.h' header
// definitions; now uses stuff from dpkg's config.h.
//  - Ian Jackson <ian@chiark.greenend.org.uk>.
// Still in the public domain.
//
// MD5.pas: Translated from C to Delphi by Jordan Russell on 2004-03-16.
// Still in the public domain. The original C code was taken from dpkg.
// Slighlty modified by Ketmar // Vampire Avalon. no global code/algo changes,
// though. Still in the public domain.
//
// $jrsoftware: issrc/Projects/MD5.pas,v 1.2 2004/03/16 17:58:14 jr Exp $
// (based on revision 1.15 from local 'md5' repository)
//
// slightly modified by Ketmar // Vampire Avalon:
//   * added a couple of "packed" keywords
//   * no more than 2GB of data can be processed
// Still in the public domain
// *** MD5 HASHING CODE IS *NOT* COVERED BY AVALONIAN LICENSES! ***
//

type
  TMD5Word = LongWord;
  TMD5Buf = array [0..3] of TMD5Word;
  TMD5In = array [0..15] of TMD5Word;
  TMD5Context = record
    buf: TMD5Buf;
    count: Integer;
    in_: TMD5In;
  end;

  TMD5Digest = packed array [0..15] of Byte;


const
  hexDigits: array [0..15] of Char = '0123456789abcdef';


function MD5Equal (const dig0, dig1: TMD5Digest): Boolean;
var
  f: Integer;
begin
  result := false;
  for f := Low(TMD5Digest) to High(TMD5Digest) do if dig0[f] <> dig1[f] then exit;
  result := true;
end;


function MD5ToStr (const md5: TMD5Digest): string;
var
  buf: packed array [0..31] of Char;
  p: PChar;
  f: Integer;
begin
  p := @buf;
  for f := 0 to 15 do
  begin
    p^ := hexDigits[md5[f] shr 4]; Inc(p);
    p^ := hexDigits[md5[f] and 15]; Inc(p);
  end;
  SetString(result, buf, 32);
end;

function MD5FromStr (const s: string; out md5: TMD5Digest): Boolean;
var
  c: Char;
  b: Byte;
  f: Integer;
begin
  result := false;
  if Length(s) <> 32 then exit;
  for f := 0 to 15 do
  begin
    c := UpCase(s[f*2+1]); if not (c in ['0'..'9','A'..'F']) then exit;
    b := ord(c)-ord('0'); if b > 9 then Dec(b, 7);
    md5[f] := (b shl 4);
    c := UpCase(s[f*2+2]); if not (c in ['0'..'9','A'..'F']) then exit;
    b := ord(c)-ord('0'); if b > 9 then Dec(b, 7);
    md5[f] := md5[f] or b;
  end;
  result := true;
end;


procedure MD5Transform (var buf: TMD5Buf; const in_: TMD5In);
// The core of the MD5 algorithm, this alters an existing MD5 hash to
// reflect the addition of 16 longwords of new data.  MD5Update blocks
// the data and converts bytes into longwords for this routine.
//
// JR: The macros have been unrolled...
var
  a, b, c, d: TMD5Word;
begin
  a := buf[0]; b := buf[1]; c := buf[2]; d := buf[3];

  // JR: The stuff below was generated using GenTransformCode.dpr
  Inc(a, in_[0]+$D76AA478+(d xor (b and (c xor d))));  a := ((a shl 7) or (a shr (32-7)))+b;
  Inc(d, in_[1]+$E8C7B756+(c xor (a and (b xor c))));  d := ((d shl 12) or (d shr (32-12)))+a;
  Inc(c, in_[2]+$242070DB+(b xor (d and (a xor b))));  c := ((c shl 17) or (c shr (32-17)))+d;
  Inc(b, in_[3]+$C1BDCEEE+(a xor (c and (d xor a))));  b := ((b shl 22) or (b shr (32-22)))+c;
  Inc(a, in_[4]+$F57C0FAF+(d xor (b and (c xor d))));  a := ((a shl 7) or (a shr (32-7)))+b;
  Inc(d, in_[5]+$4787C62A+(c xor (a and (b xor c))));  d := ((d shl 12) or (d shr (32-12)))+a;
  Inc(c, in_[6]+$A8304613+(b xor (d and (a xor b))));  c := ((c shl 17) or (c shr (32-17)))+d;
  Inc(b, in_[7]+$FD469501+(a xor (c and (d xor a))));  b := ((b shl 22) or (b shr (32-22)))+c;
  Inc(a, in_[8]+$698098D8+(d xor (b and (c xor d))));  a := ((a shl 7) or (a shr (32-7)))+b;
  Inc(d, in_[9]+$8B44F7AF+(c xor (a and (b xor c))));  d := ((d shl 12) or (d shr (32-12)))+a;
  Inc(c, in_[10]+$FFFF5BB1+(b xor (d and (a xor b)))); c := ((c shl 17) or (c shr (32-17)))+d;
  Inc(b, in_[11]+$895CD7BE+(a xor (c and (d xor a)))); b := ((b shl 22) or (b shr (32-22)))+c;
  Inc(a, in_[12]+$6B901122+(d xor (b and (c xor d)))); a := ((a shl 7) or (a shr (32-7)))+b;
  Inc(d, in_[13]+$FD987193+(c xor (a and (b xor c)))); d := ((d shl 12) or (d shr (32-12)))+a;
  Inc(c, in_[14]+$A679438E+(b xor (d and (a xor b)))); c := ((c shl 17) or (c shr (32-17)))+d;
  Inc(b, in_[15]+$49B40821+(a xor (c and (d xor a)))); b := ((b shl 22) or (b shr (32-22)))+c;

  Inc(a, in_[1]+$F61E2562+(c xor (d and (b xor c))));  a := ((a shl 5) or (a shr (32-5)))+b;
  Inc(d, in_[6]+$C040B340+(b xor (c and (a xor b))));  d := ((d shl 9) or (d shr (32-9)))+a;
  Inc(c, in_[11]+$265E5A51+(a xor (b and (d xor a)))); c := ((c shl 14) or (c shr (32-14)))+d;
  Inc(b, in_[0]+$E9B6C7AA+(d xor (a and (c xor d))));  b := ((b shl 20) or (b shr (32-20)))+c;
  Inc(a, in_[5]+$D62F105D+(c xor (d and (b xor c))));  a := ((a shl 5) or (a shr (32-5)))+b;
  Inc(d, in_[10]+$02441453+(b xor (c and (a xor b)))); d := ((d shl 9) or (d shr (32-9)))+a;
  Inc(c, in_[15]+$D8A1E681+(a xor (b and (d xor a)))); c := ((c shl 14) or (c shr (32-14)))+d;
  Inc(b, in_[4]+$E7D3FBC8+(d xor (a and (c xor d))));  b := ((b shl 20) or (b shr (32-20)))+c;
  Inc(a, in_[9]+$21E1CDE6+(c xor (d and (b xor c))));  a := ((a shl 5) or (a shr (32-5)))+b;
  Inc(d, in_[14]+$C33707D6+(b xor (c and (a xor b)))); d := ((d shl 9) or (d shr (32-9)))+a;
  Inc(c, in_[3]+$F4D50D87+(a xor (b and (d xor a))));  c := ((c shl 14) or (c shr (32-14)))+d;
  Inc(b, in_[8]+$455A14ED+(d xor (a and (c xor d))));  b := ((b shl 20) or (b shr (32-20)))+c;
  Inc(a, in_[13]+$A9E3E905+(c xor (d and (b xor c)))); a := ((a shl 5) or (a shr (32-5)))+b;
  Inc(d, in_[2]+$FCEFA3F8+(b xor (c and (a xor b))));  d := ((d shl 9) or (d shr (32-9)))+a;
  Inc(c, in_[7]+$676F02D9+(a xor (b and (d xor a))));  c := ((c shl 14) or (c shr (32-14)))+d;
  Inc(b, in_[12]+$8D2A4C8A+(d xor (a and (c xor d)))); b := ((b shl 20) or (b shr (32-20)))+c;

  Inc(a, in_[5]+$FFFA3942+(b xor c xor d));  a := ((a shl 4) or (a shr (32-4)))+b;
  Inc(d, in_[8]+$8771F681+(a xor b xor c));  d := ((d shl 11) or (d shr (32-11)))+a;
  Inc(c, in_[11]+$6D9D6122+(d xor a xor b)); c := ((c shl 16) or (c shr (32-16)))+d;
  Inc(b, in_[14]+$FDE5380C+(c xor d xor a)); b := ((b shl 23) or (b shr (32-23)))+c;
  Inc(a, in_[1]+$A4BEEA44+(b xor c xor d));  a := ((a shl 4) or (a shr (32-4)))+b;
  Inc(d, in_[4]+$4BDECFA9+(a xor b xor c));  d := ((d shl 11) or (d shr (32-11)))+a;
  Inc(c, in_[7]+$F6BB4B60+(d xor a xor b));  c := ((c shl 16) or (c shr (32-16)))+d;
  Inc(b, in_[10]+$BEBFBC70+(c xor d xor a)); b := ((b shl 23) or (b shr (32-23)))+c;
  Inc(a, in_[13]+$289B7EC6+(b xor c xor d)); a := ((a shl 4) or (a shr (32-4)))+b;
  Inc(d, in_[0]+$EAA127FA+(a xor b xor c));  d := ((d shl 11) or (d shr (32-11)))+a;
  Inc(c, in_[3]+$D4EF3085+(d xor a xor b));  c := ((c shl 16) or (c shr (32-16)))+d;
  Inc(b, in_[6]+$04881D05+(c xor d xor a));  b := ((b shl 23) or (b shr (32-23)))+c;
  Inc(a, in_[9]+$D9D4D039+(b xor c xor d));  a := ((a shl 4) or (a shr (32-4)))+b;
  Inc(d, in_[12]+$E6DB99E5+(a xor b xor c)); d := ((d shl 11) or (d shr (32-11)))+a;
  Inc(c, in_[15]+$1FA27CF8+(d xor a xor b)); c := ((c shl 16) or (c shr (32-16)))+d;
  Inc(b, in_[2]+$C4AC5665+(c xor d xor a));  b := ((b shl 23) or (b shr (32-23)))+c;

  Inc(a, in_[0]+$F4292244+(c xor (b or (not d))));  a := ((a shl 6) or (a shr (32-6)))+b;
  Inc(d, in_[7]+$432AFF97+(b xor (a or (not c))));  d := ((d shl 10) or (d shr (32-10)))+a;
  Inc(c, in_[14]+$AB9423A7+(a xor (d or (not b)))); c := ((c shl 15) or (c shr (32-15)))+d;
  Inc(b, in_[5]+$FC93A039+(d xor (c or (not a))));  b := ((b shl 21) or (b shr (32-21)))+c;
  Inc(a, in_[12]+$655B59C3+(c xor (b or (not d)))); a := ((a shl 6) or (a shr (32-6)))+b;
  Inc(d, in_[3]+$8F0CCC92+(b xor (a or (not c))));  d := ((d shl 10) or (d shr (32-10)))+a;
  Inc(c, in_[10]+$FFEFF47D+(a xor (d or (not b)))); c := ((c shl 15) or (c shr (32-15)))+d;
  Inc(b, in_[1]+$85845DD1+(d xor (c or (not a))));  b := ((b shl 21) or (b shr (32-21)))+c;
  Inc(a, in_[8]+$6FA87E4F+(c xor (b or (not d))));  a := ((a shl 6) or (a shr (32-6)))+b;
  Inc(d, in_[15]+$FE2CE6E0+(b xor (a or (not c)))); d := ((d shl 10) or (d shr (32-10)))+a;
  Inc(c, in_[6]+$A3014314+(a xor (d or (not b))));  c := ((c shl 15) or (c shr (32-15)))+d;
  Inc(b, in_[13]+$4E0811A1+(d xor (c or (not a)))); b := ((b shl 21) or (b shr (32-21)))+c;
  Inc(a, in_[4]+$F7537E82+(c xor (b or (not d))));  a := ((a shl 6) or (a shr (32-6)))+b;
  Inc(d, in_[11]+$BD3AF235+(b xor (a or (not c)))); d := ((d shl 10) or (d shr (32-10)))+a;
  Inc(c, in_[2]+$2AD7D2BB+(a xor (d or (not b))));  c := ((c shl 15) or (c shr (32-15)))+d;
  Inc(b, in_[9]+$EB86D391+(d xor (c or (not a))));  b := ((b shl 21) or (b shr (32-21)))+c;

  Inc(buf[0], a); Inc(buf[1], b); Inc(buf[2], c); Inc(buf[3], d);
end;


procedure MD5Init (var ctx: TMD5Context);
// Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
// initialization constants.
begin
  ctx.buf[0] := TMD5Word($67452301);
  ctx.buf[1] := TMD5Word($EFCDAB89);
  ctx.buf[2] := TMD5Word($98BADCFE);
  ctx.buf[3] := TMD5Word($10325476);
  ctx.count := 0;
end;

procedure MD5Update (var ctx: TMD5Context; const buffer; len: Integer);
// Update context to reflect the concatenation of another buffer full
// of bytes.
var
  buf: ^Byte;
  t: Integer;
begin
  if len < 1 then exit;

  buf := @buffer;

  // Update byte count
  t := ctx.count; Inc(ctx.count, len);

  t := 64-(t and $3F); // Space available in ctx.in (at least 1)
  if t > len then
  begin
    Move(buf^, Pointer(Integer(@ctx.in_)+64-t)^, len);
    exit;
  end;

  // First chunk is an odd size
  Move(buf^, Pointer(Integer(@ctx.in_)+64-t)^, t);
  MD5Transform(ctx.buf, ctx.in_);
  Inc(buf, t); Dec(len, t);

  // Process data in 64-byte chunks
  while len >= 64 do
  begin
    Move(buf^, ctx.in_, 64);
    MD5Transform(ctx.buf, ctx.in_);
    Inc(buf, 64); Dec(len, 64);
  end;

  // Handle any remaining bytes of data
  if len > 0 then Move(buf^, ctx.in_, len);
end;

function MD5Final (var ctx: TMD5Context): TMD5Digest;
// Final wrapup - pad to 64-byte boundary with the bit pattern
// 1 0* (64-bit count of bits processed, MSB-first)
var
  count: Integer;
  p: ^Byte;
begin
  count := ctx.count and $3F; // Number of bytes in ctx.in
  p := @ctx.in_; Inc(p, count);

  // Set the first char of padding to 0x80. There is always room.
  p^ := $80;
  Inc(p);

  // Bytes of padding needed to make 56 bytes (-8..55)
  count := 56-1-count;

  if count < 0 then
  begin
    // Padding forces an extra block
    FillChar(p^, count+8, 0);
    MD5Transform(ctx.buf, ctx.in_);
    p := @ctx.in_;
    count := 56;
  end;
  FillChar(p^, count, 0);

  // Append length in bits and transform
  ctx.in_[14] := LongWord(ctx.count) shl 3;
  ctx.in_[15] := LongWord(ctx.count) shr 29;
  MD5Transform(ctx.buf, ctx.in_);

  Move(ctx.buf, result, 16);
  FillChar(ctx, SizeOf(ctx), 0);  // In case it's sensitive
end;

// New functions by JR:
function MD5Buf (const buffer; len: Integer): TMD5Digest;
var
  context: TMD5Context;
begin
  MD5Init(context);
  if len > 0 then MD5Update(context, buffer, len);
  result := MD5Final(context);
end;
{$ENDIF}

{$IFDEF XPACK_CRYPT_BLOCKZ}
function RandIntSeeded (var seed: Integer): Integer;
var
  k1, ix: Integer;
begin
  ix := seed;

  k1 := ix div 127773;
  ix := 16807*(ix-k1*127773)-k1*2836;
  if ix < 0 then Inc(ix, 2147483647);
  seed := ix;

  result := ix;
end;

function RandByteSeeded (var seed: Integer): Byte;
begin
  result := (RandIntSeeded(seed) and 255);
end;

procedure XorCopyBuf (const src; len: Integer; out dest; seed: Integer);
var
  s, d: PChar;
begin
  s := @src; d := @dest;
  while len > 0 do
  begin
    d^ := Char(Byte(s^) xor RandByteSeeded(seed));
    Inc(s); Inc(d); Dec(len);
  end;
end;

procedure XorBuf (const src; len: Integer; seed: Integer);
var
  s, d: PChar;
begin
  s := @src; d := @src;
  while len > 0 do
  begin
    d^ := Char(Byte(s^) xor RandByteSeeded(seed));
    Inc(s); Inc(d); Dec(len);
  end;
end;
{$ENDIF}


{ TXArkBlockFile }
constructor TXArkBlockFile.CreateNew (pStream: TStream);
// create new block file
var
  w: Word;
begin
  fCachedBlock := -1;
  fCurBlock := 0;
  fCurOfs := 0;

  fStream := pStream;

  fCount := 1;
  // write header
  w := 0; fStream.WriteBuffer(w, 2); // # of blocks
  w := 0; fStream.WriteBuffer(w, 2); // bytes in the last block
  fLastBlockUsed := 0;
  fChanged := false;

  fOfs := TList.Create();
  fOfs.Add(Pointer(fStream.Position));
end;

constructor TXArkBlockFile.Create (pStream: TStream; testAll: Boolean = false);
var
  w: Word;
  f, c, usz: Integer;
  md5, cmd5: TMD5Digest;
  ctx5: TMD5Context;
  buf: Pointer;
  b: Byte;
begin
  fCachedBlock := -1;
  fCurBlock := 0;
  fCurOfs := 0;

  fStream := pStream;
  fChanged := false;
  // read header
  fStream.ReadBuffer(w, 2); fCount := Integer(w)+1;
  fStream.ReadBuffer(w, 2); fLastBlockUsed := w;

  GetMem(buf, 65536);
  try
    fOfs := TList.Create();
    // scan blocks
    try
      for f := 2 to fCount do  // fCount-1 full blocks
      begin
        fOfs.Add(Pointer(fStream.Position));
        fStream.ReadBuffer(md5, SizeOf(md5)); // skip md5
        fStream.ReadBuffer(w, 2); // size of the packed data
        c := w; if c = 0 then c := 65536; // size 0: unpacked block
        if testAll then
        begin
          // unpack & test the block
          if c = 65536 then
          begin
            fStream.ReadBuffer(fLastBlockData[0], 65536);
            {$IFDEF XPACK_CRYPT_BLOCKZ}
            XorBuf(fLastBlockData[0], 65536, 666);
            {$ENDIF}
          end
          else
          begin
            fStream.ReadBuffer(buf^, c);
            {$IFDEF XPACK_CRYPT_BLOCKZ}
            XorBuf(buf^, c, 666);
            {$ENDIF}
            usz := XUnpackBuffer(buf^, c, fLastBlockData[0], 65536);
            if usz <> 65536 then
              raise EReadError.Create('BlockFile: block #'+IntToHex(f-2, 4)+' is corrupted');
          end;
          MD5Init(ctx5);
          MD5Update(ctx5, fLastBlockData[0], 65536);
          cmd5 := MD5Final(ctx5);

          if not MD5Equal(md5, cmd5) then
            raise EReadError.Create('BlockFile: block #'+IntToHex(f-2, 4)+' is corrupted (md5)');

          if fCachedBlock = -1 then
          begin
            // cache first block
            fCachedBlock := f-2;
            Move(fLastBlockData[0], fCachedBlockData[0], 65536);
          end;
        end
        else fStream.Seek(c, soFromCurrent); // skip block data
      end;

      // add last partially full block
      fOfs.Add(Pointer(fStream.Position));
      if fLastBlockUsed > 0 then
      begin
        // there is some data, so cache last block
        fStream.ReadBuffer(md5, SizeOf(md5));
        fStream.ReadBuffer(b, 1); // "packed" flag
        if b = 0 then
        begin
           // not packed
          fStream.ReadBuffer(fLastBlockData[0], fLastBlockUsed);
          {$IFDEF XPACK_CRYPT_BLOCKZ}
          XorBuf(fLastBlockData[0], fLastBlockUsed, 666);
          {$ENDIF}
        end
        else
        begin
          // packed
          fStream.ReadBuffer(w, 2); c := w; // size of the packed data
          ASSERT(w <> 0); // this should never fail!
          fStream.ReadBuffer(buf^, c);
          {$IFDEF XPACK_CRYPT_BLOCKZ}
          XorBuf(buf^, c, 666);
          {$ENDIF}
          usz := XUnpackBuffer(buf^, c, fLastBlockData[0], fLastBlockUsed);
          if usz <> fLastBlockUsed then
            raise EReadError.Create('BlockFile: the last block is corrupted');
        end;

        // check md5
        MD5Init(ctx5);
        MD5Update(ctx5, fLastBlockData[0], fLastBlockUsed);
        cmd5 := MD5Final(ctx5);

        if not MD5Equal(md5, cmd5) then raise EReadError.Create('BlockFile: the last block is corrupted (md5)');
      end;
    except
      fOfs.Free();
      raise;
    end;
  finally
    ReallocMem(buf, 0);
  end;
end;

destructor TXArkBlockFile.Destroy ();
begin
  if fChanged then Flush();
  fOfs.Free();
end;

function TXArkBlockFile.GetPosition (): Integer;
begin
  result := fCurBlock*65536+fCurOfs;
end;

procedure TXArkBlockFile.SetPosition (value: Integer);
var
  sz: Integer;
begin
  sz := (fCount-1)*65536+fLastBlockUsed;
  if value < 0 then value := 0
  else if value > sz then value := sz;
  fCurBlock := value div 65536; fCurOfs := value mod 65536;
end;

function TXArkBlockFile.GetSize (): Integer;
begin
  result := (fCount-1)*65536+fLastBlockUsed;
end;

function TXArkBlockFile.ReadBlock (index: Integer; out buf): Integer;
// returns unpacked block size (1-65536)
// raises an exception if there's no such block
var
  rbuf: Pointer;
  w: Word;
  md5, cmd5: TMD5Digest;
  ctx5: TMD5Context;
  c: Integer;
begin
  if (index < 0) or (index >= fCount) then
    raise EReadError.Create('BlockFile: block #'+IntToHex(index, 4)+' not found');
  if index = fCount-1 then
  begin
    result := fLastBlockUsed;
    if result <> 0 then Move(fLastBlockData[0], buf, result);
    exit;
  end;

  result := 65536;
  fStream.Position := Integer(fOfs[index]);
  GetMem(rbuf, 65536);
  try
    fStream.ReadBuffer(md5, SizeOf(md5));
    fStream.ReadBuffer(w, 2); c := w;
    if c = 0 then
    begin
      fStream.ReadBuffer(buf, 65536);
      {$IFDEF XPACK_CRYPT_BLOCKZ}
      XorBuf(buf, 65536, 666);
      {$ENDIF}
    end
    else
    begin
      fStream.ReadBuffer(rbuf^, c);
      {$IFDEF XPACK_CRYPT_BLOCKZ}
      XorBuf(rbuf^, c, 666);
      {$ENDIF}
      c := XUnpackBuffer(rbuf^, c, buf, 65536);
      if c <> 65536 then
        raise EReadError.Create('BlockFile: block '+IntToHex(index, 4)+' is corrupted');
    end;

    MD5Init(ctx5);
    MD5Update(ctx5, buf, 65536);
    cmd5 := MD5Final(ctx5);

    if not MD5Equal(md5, cmd5) then
      raise EReadError.Create('BlockFile: block #'+IntToHex(index, 4)+' is corrupted (md5)');
  finally
    ReallocMem(rbuf, 0);
  end;
end;

procedure TXArkBlockFile.Flush ();
var
  buf: Pointer;
  c: Integer;
  w: Word;
  md5: TMD5Digest;
  ctx5: TMD5Context;
begin
  if not fChanged then exit;

  ASSERT(fLastBlockUsed > 0);
  ASSERT(fLastBlockUsed <= 65535);

  fStream.Position := Integer(fOfs[0])-2;
  w := fLastBlockUsed; fStream.WriteBuffer(w, 2);

  fStream.Position := Integer(fOfs[fCount-1]);
  if fLastBlockUsed > 0 then
  begin
    MD5Init(ctx5);
    MD5Update(ctx5, fLastBlockData[0], fLastBlockUsed);
    md5 := MD5Final(ctx5);

    GetMem(buf, 65536);
    try
      fStream.WriteBuffer(md5, SizeOf(md5));

      c := XPackBuffer(fLastBlockData[0], fLastBlockUsed, buf^, 65535);
      if (c = -1) or (c >= fLastBlockUsed-2) then
      begin
        // the block is uncompressible
        w := 0; fStream.WriteBuffer(w, 1); // "uncompressed" flag
        {$IFDEF XPACK_CRYPT_BLOCKZ}
        XorBuf(fLastBlockData[0], fLastBlockUsed, 666);
        {$ENDIF}
        fStream.WriteBuffer(fLastBlockData[0], fLastBlockUsed);
        {$IFDEF XPACK_CRYPT_BLOCKZ}
        XorBuf(fLastBlockData[0], fLastBlockUsed, 666);
        {$ENDIF}
      end
      else
      begin
        w := 1; fStream.WriteBuffer(w, 1); // "compressed" flag
        w := c; fStream.WriteBuffer(w, 2);
        {$IFDEF XPACK_CRYPT_BLOCKZ}
        XorBuf(buf^, c, 666);
        {$ENDIF}
        fStream.WriteBuffer(buf^, c);
      end;
    finally
      ReallocMem(buf, 0);
    end;
  end;

  fStream.Size := fStream.Position;
  fChanged := false;
end;

procedure TXArkBlockFile.WriteBuffer (const buf; count: Integer);
var
  p: PChar;
  wrt, c: Integer;
  md5: TMD5Digest;
  ctx5: TMD5Context;
  wbuf: Pointer;
  w: Word;
begin
  p := @buf;
  while count > 0 do
  begin
    wrt := count;
    if fLastBlockUsed+wrt > 65536 then wrt := 65536-fLastBlockUsed;
    Move(p^, fLastBlockData[fLastBlockUsed], wrt);
    Inc(fLastBlockUsed, wrt); Inc(p, wrt); Dec(count, wrt);
    fChanged := true;
    ASSERT(fLastBlockUsed <= 65536);
    if fLastBlockUsed = 65536 then
    begin
      // write full block
      if fCount = 65535 then raise EWriteError.Create('BlockFile: too much blocks');
      fStream.Position := Integer(fOfs[fCount-1]);

      MD5Init(ctx5);
      MD5Update(ctx5, fLastBlockData[0], 65536);
      md5 := MD5Final(ctx5);

      GetMem(wbuf, 65536);
      try
        c := XPackBuffer(fLastBlockData[0], 65536, wbuf^, 65535);
        if c = -1 then
        begin
          // the block is uncompressible
          w := 0; Move(fLastBlockData[0], wbuf^, 65536); c := 65536;
        end
        else w := c;

        fStream.WriteBuffer(md5, SizeOf(md5));
        fStream.WriteBuffer(w, 2);
        {$IFDEF XPACK_CRYPT_BLOCKZ}
        XorBuf(wbuf^, c, 666);
        {$ENDIF}
        fStream.WriteBuffer(wbuf^, c);
      finally
        ReallocMem(wbuf, 0);
      end;

      Inc(fCount);
      fOfs.Add(Pointer(fStream.Position));
      fStream.Position := Integer(fOfs[0])-4;
      w := fCount-1; fStream.WriteBuffer(w, 2); // block count
      w := 0; fStream.WriteBuffer(w, 2); // last block data size
      fLastBlockUsed := 0;
      fChanged := false;
    end;
  end;
end;

procedure TXArkBlockFile.ReadBuffer (out buf; count: Integer);
var
  dp, sp: PChar;
  blkSz: Integer;
  rd: Integer;
begin
  dp := @buf;
  while count > 0 do
  begin
    if GetPosition() >= GetSize() then
      raise EWriteError.Create('BlockFile: read error (out of data)');

    if fCurBlock = fCount-1 then
    begin
      sp := @(fLastBlockData[0]); blkSz := fLastBlockUsed;
    end
    else
    begin
      if fCurBlock <> fCachedBlock then
      begin
        try
          fCachedBlock := fCurBlock;
          ReadBlock(fCachedBlock, fCachedBlockData[0]);
        except
          fCachedBlock := -1;
          raise;
        end;
      end;
      sp := @(fCachedBlockData[0]);
      blkSz := 65536;
    end;

    rd := count;
    if fCurOfs+rd > blkSz then rd := blkSz-fCurOfs;
    Move(sp[fCurOfs], dp^, rd);
    Inc(fCurOfs, rd); Inc(dp, rd); Dec(count, rd);
    ASSERT(fCurOfs <= 65536);
    if fCurOfs = 65536 then begin fCurOfs := 0; Inc(fCurBlock); end;
  end;
end;
{$ENDIF}


{$IFDEF XPACK_INCLUDE_ARK_FUNCZ}
{ TXArkFileStream }
constructor TXArkFileStream.Create (srcArk: TXArkFile; srcS: TStream; pStart, pSize: Integer);
begin
  inherited Create();
  fStream := srcS; fArk := srcArk;
  fStart := pStart; fSize := pSize;
  fPos := 0;
  if fArk <> nil then Inc(fArk.fOpenFileCount);
end;

destructor TXArkFileStream.Destroy ();
begin
  fStream.Free();
  if fArk <> nil then Dec(fArk.fOpenFileCount);
  inherited;
end;

function TXArkFileStream.Seek (offset: Longint; origin: Word): Longint;
begin
  case origin of
    soFromCurrent: Inc(offset, fPos);
    soFromEnd: offset := fSize-offset;
  end;
  if offset < 0 then offset := 0 else if offset > fSize then offset := fSize;
  fPos := offset;
  result := fPos;
end;

function TXArkFileStream.Read (var buffer; count: Longint): Longint;
var
  len: Integer;
begin
  if count < 0 then raise EReadError.Create('invalid buffer size');
  len := fSize-fPos;
  if count > len then raise EReadError.Create('out of data');

  if fStream <> nil then
  begin
    fStream.Position := fStart+fPos;
    fStream.ReadBuffer(buffer, count);
  end
  else
  begin
    fArk.fBlockF.Position := fStart+fPos;
    fArk.fBlockF.ReadBuffer(buffer, count);
  end;

  Inc(fPos, count);
  result := count;
end;

function TXArkFileStream.Write (const buffer; count: Longint): Longint;
begin
  raise EWriteError.Create('writing to read-only file');
end;


{ TXArkFile }
constructor TXArkFile.Create ();
begin
  fStream := nil;
  fFileName := '';
  fDirs := TStringList.Create();
  fFiles := TList.Create();
  fReadMode := true;
  fBlockF := nil;
  fOpenFileCount := 0;
end;

destructor TXArkFile.Destroy ();
begin
  if fStream <> nil then Close();
  fFiles.Free();
  fDirs.Free();
end;

procedure TXArkFile.Close ();
begin
  if fStream = nil then raise EXArkError.Create('ARK file is not opened');
  if fOpenFileCount <> 0 then raise EXArkError.Create('ARK has some opened files');
  if not fReadMode then SaveDir();
  FreeAndNil(fBlockF);
  if fFileName <> '' then FreeAndNil(fStream) else fStream := nil;
  fFileName := '';
  ClearDir();
end;

function TXArkFile.GetArkFileSize (): Integer;
begin
  result := fStream.Size;
end;

procedure TXArkFile.ClearDir ();
var
  f: Integer;
begin
  for f := 0 to fFiles.Count-1 do Dispose(PXArkDirEntry(fFiles[f]));
  fFiles.Clear();
  fDirs.Clear();
end;

procedure TXArkFile.SaveDir ();
var
  f: Integer;
  ss: ShortString;
  s: string;
  fi: PXArkDirEntry;
  sign: packed array [0..3] of Char;
begin
  fStream.Position := 0;
  sign := 'ARK0';
  fStream.WriteBuffer(sign[0], 4);
  f := fFiles.Count; ASSERT(f < 65536);
  fStream.WriteBuffer(f, 2);
  f := fDirs.Count; ASSERT(f < 65536);
  fStream.WriteBuffer(f, 2);
  fi := PXArkDirEntry(fFiles[fFiles.Count-1]);
  f := fi.ofs+fi.size; fStream.WriteBuffer(f, 4);

  fBlockF.Position := f;
  for f := 0 to fDirs.Count-1 do
  begin
    s := fDirs[f];
    ASSERT(Length(s) < 256);
    ss := s;
    fBlockF.WriteBuffer(ss[0], Length(ss)+1);
  end;

  for f := 0 to fFiles.Count-1 do
  begin
    fi := PXArkDirEntry(fFiles[f]);
    fBlockF.WriteBuffer(fi.name[0], Length(fi.name)+1);
    fBlockF.WriteBuffer(fi.dir, 2);
    fBlockF.WriteBuffer(fi.size, 4);
  end;
end;

procedure TXArkFile.LoadDir (checkFile: Boolean);
var
  f, fc, dc, dofs, pofs: Integer;
  w: Word;
  fi: PXArkDirEntry;
  sign: packed array [0..3] of Char;
  ss: ShortString;
begin
  ClearDir();
  fStream.Position := 0;
  fStream.ReadBuffer(sign[0], 4);
  if sign <> 'ARK0' then raise EXArkError.Create('invalid ARK signature');
  fStream.ReadBuffer(w, 2); fc := w;
  fStream.ReadBuffer(w, 2); dc := w;
  fStream.ReadBuffer(dofs, 4);
  fBlockF := TXArkBlockFile.Create(fStream, checkFile);
  try
    if (dofs < 0) or (dofs > fBlockF.Size) then raise EXArkError.Create('invalid ARK file (0)');
    fBlockF.Position := dofs;

    for f := 1 to dc do
    begin
      fBlockF.ReadBuffer(ss[0], 1);
      fBlockF.ReadBuffer(ss[1], Length(ss));
      fDirs.Append(ss);
    end;

    pofs := 0;
    for f := 1 to fc do
    begin
      New(fi);
      try
        fBlockF.ReadBuffer(fi.name[0], 1);
        fBlockF.ReadBuffer(fi.name[1], Length(fi.name));
        fBlockF.ReadBuffer(fi.dir, 2);
        fBlockF.ReadBuffer(fi.size, 4);
        if fi.Size < 0 then raise EXArkError.Create('invalid ARK file (1)');
        fi.ofs := pofs; Inc(pofs, fi.size);
        fFiles.Add(fi);
      except
        Dispose(fi);
        raise;
      end;
    end;
  except
    fBlockF.Free();
    ClearDir();
    raise;
  end;
end;

procedure TXArkFile.OldFromStream (srcS: TStream; checkFile: Boolean = false);
begin
  if fStream <> nil then raise EXArkError.Create('ARK already opened');
  fReadMode := true;
  fStream := srcS;
  LoadDir(checkFile);
end;

procedure TXArkFile.UpdateToStream (srcS: TStream);
begin
  OldFromStream(srcS);
  fReadMode := false;
end;

procedure TXArkFile.NewToStream (dstS: TStream);
var
  buf: packed array [0..11] of Byte;
begin
  if fStream <> nil then raise EXArkError.Create('ARK already opened');
  ClearDir();
  dstS.Position := 0;
  dstS.WriteBuffer(buf[0], 12);
  fStream := dstS;
  try
    fReadMode := false;
    fBlockF := TXArkBlockFile.CreateNew(fStream);
  except
    fReadMode := true;
    raise;
  end;
end;

procedure TXArkFile.OldFromFile (const fName: string; checkFile: Boolean = false);
var
  fs: TFileStream;
begin
  fs := TFileStream.Create(fName, fmOpenRead or fmShareDenyWrite);
  try
    OldFromStream(fs, checkfile);
    fFileName := fName;
  except
    fs.Free();
    fFileName := '';
    fReadMode := true;
    raise;
  end;
end;

procedure TXArkFile.NewToFile (const fName: string);
var
  fs: TFileStream;
begin
  try
    fs := TFileStream.Create(fName, fmCreate);// or fmShareDenyWrite);
  except
    fs := TFileStream.Create(fName, fmOpenReadWrite or fmShareDenyWrite);
  end;
  try
    NewToStream(fs);
    fFileName := fName;
  except
    fs.Free();
    fFileName := '';
    fReadMode := true;
    raise;
  end;
end;

procedure TXArkFile.UpdateToFile (const fName: string);
var
  fs: TFileStream;
begin
  fs := TFileStream.Create(fName, fmOpenReadWrite or fmShareDenyWrite);
  try
    UpdateToStream(fs);
    fFileName := fName;
  except
    fs.Free();
    fFileName := '';
    fReadMode := true;
    raise;
  end;
end;

function TXArkFile.IsAppending (): Boolean;
begin
  result := (not fReadMode) and (fOpenFileCount > 0);
end;

procedure TXArkFile.NewFile (const fName: string);
var
  f: Integer;
  path, name: string;
  fi, fi1: PXArkDirEntry;
  dind: Integer;
begin
  if fReadMode then raise EXArkError.Create('appending to read-only archive');
  if fOpenFileCount <> 0 then raise EXArkError.Create('previous appending file is not closed');
  path := StringReplace(fName, '/', '\', [rfReplaceAll]);
  name := ExtractFileName(path);
  path := StringReplace(ExtractFilePath(path), '\', '/', [rfReplaceAll]);

  if (Length(path) > 0) and (path[1] = '/') then Delete(path, 1, 1);
  if (path <> '') and (path[Length(path)] = '/') then Delete(path, Length(path), 1);

  f := fDirs.Count-1;
  while (f >= 0) and (ANSICompareText(fDirs[f], path) <> 0) do Dec(f);
  if f < 0 then
  begin
    if fDirs.Count = 65535 then raise EXArkError.Create('too many directories');
    f := fDirs.Add(path);
    dind := f;
  end
  else
  begin
    dind := f;
    f := fFiles.Count-1;
    while f >= 0 do
    begin
      fi := PXArkDirEntry(fFiles[f]);
      if (fi.dir = dind) and (ANSICompareText(fi.name, name) = 0) then break;
      Dec(f);
    end;
    if f >= 0 then raise EXArkError.Create('duplicate file');
  end;

  New(fi);
  try
    fi.name := name;
    fi.dir := dind;
    fi.size := 0;
    if fFiles.Count = 0 then fi.ofs := 0
    else
    begin
      fi1 := PXArkDirEntry(fFiles[fFiles.Count-1]);
      fi.ofs := fi1.ofs+fi1.size;
    end;
    fFiles.Add(fi);
  except
    Dispose(fi);
    raise;
  end;
  Inc(fOpenFileCount);
end;

procedure TXArkFile.NewClose ();
begin
  if fReadMode then raise EXArkError.Create('appending to read-only archive');
  if fOpenFileCount = 0 then raise EXArkError.Create('appending file is not opened');
  Dec(fOpenFileCount);
end;

procedure TXArkFile.NewWrite (const buf; len: Integer);
var
  fi: PXArkDirEntry;
begin
  if fReadMode then raise EXArkError.Create('appending to read-only archive');
  if fOpenFileCount = 0 then raise EXArkError.Create('appending file is not opened');
  if len < 0 then raise EXArkError.Create('invalid writing buffer size');
  if len = 0 then exit;

  fi := PXArkDirEntry(fFiles[fFiles.Count-1]);
  fBlockF.Position := fi.ofs+fi.size;
  fBlockF.WriteBuffer(buf, len);
  Inc(fi.size, len);
end;

procedure TXArkFile.AppendStream (const fName: string; srcS: TStream; len: Integer);
// len <= 0: whole srcS
var
  buf: packed array [0..4095] of Byte;
  rd: Integer;
begin
  if len < 1 then begin srcS.Position := 0; len := srcS.Size; end;
  NewFile(fName);
  try
    while len > 0 do
    begin
      rd := SizeOf(buf); if rd > len then rd := len;
      srcS.ReadBuffer(buf[0], rd);
      NewWrite(buf[0], rd);
      Dec(len, rd);
    end;
  finally
    NewClose();
  end;
end;

procedure TXArkFile.AppendFile (const fName, srcFName: string);
var
  fi: TFileStream;
begin
  fi := TFileStream.Create(srcFName, fmOpenRead or fmShareDenyWrite);
  try
    AppendStream(fName, fi, 0);
  finally
    fi.Free();
  end;
end;

function TXArkFile.OpenFile (const fName: string): TStream;
var
  f: Integer;
  path, name: string;
  fi: PXArkDirEntry;
  dind: Integer;
begin
  result := nil;
  if fBlockF = nil then raise EWriteError.Create('archive is not opened');
  if not fReadMode then raise EXArkError.Create('reading from write-only archive');
  path := StringReplace(fName, '/', '\', [rfReplaceAll]);
  name := ExtractFileName(path);
  path := StringReplace(ExtractFilePath(path), '\', '/', [rfReplaceAll]);

  if (Length(path) > 0) and (path[1] = '/') then Delete(path, 1, 1);
  if (path <> '') and (path[Length(path)] = '/') then Delete(path, Length(path), 1);

  fi := nil;
  f := fDirs.Count-1;
  while (f >= 0) and (ANSICompareText(fDirs[f], path) <> 0) do Dec(f);
  if f >= 0 then
  begin
    dind := f;
    f := fFiles.Count-1;
    while f >= 0 do
    begin
      fi := PXArkDirEntry(fFiles[f]);
      if (fi.dir = dind) and (ANSICompareText(fi.name, name) = 0) then break;
      Dec(f);
    end;
  end;
  if f < 0 then exit;

  result := TXArkFileStream.Create(self, nil, fi.ofs, fi.size);
end;

function TXArkFile.OldFile (const fName: string): TStream;
begin
  result := OpenFile(fName);
  if result = nil then raise EXArkError.Create('file not found');
end;

function TXArkFile.GetFileCount (): Integer;
begin
  result := fFiles.Count;
end;

function TXArkFile.GetFilePath (index: Integer): string;
begin
  if (index < 0) or (index >= fFiles.Count) then result := ''
  else result := fDirs[PXArkDirEntry(fFiles[index]).dir];
end;

function TXArkFile.GetFileName (index: Integer): string;
begin
  if (index < 0) or (index >= fFiles.Count) then result := ''
  else result := PXArkDirEntry(fFiles[index]).name;
end;

function TXArkFile.GetFileSize (index: Integer): Integer;
begin
  if (index < 0) or (index >= fFiles.Count) then result := 0
  else result := PXArkDirEntry(fFiles[index]).size;
end;
{$ENDIF}


end.
